package algo

import core._
import lift.arithmetic
import lift.arithmetic.simplifier.SimplifyIfThenElse
import lift.arithmetic.{ArithExpr, StartFromRange, ceil}

import scala.annotation.tailrec

final case class AlgoLambda private (param: ParamDef, body: Expr, t: FunTypeT) extends LambdaT {
  override def build(newChildren: Seq[ShirIR]): AlgoLambda = AlgoLambda(newChildren.head.asInstanceOf[ParamDef], newChildren(1).asInstanceOf[Expr], newChildren(2).asInstanceOf[FunTypeT])
}

object AlgoLambda {
  def apply(param: ParamDef, body: Expr): AlgoLambda = AlgoLambda(param, body, AlgoFunType(param.t, body.t))

  @tailrec
  def apply(params: Seq[ParamDef], body: Expr): AlgoLambda = {
    if (params.isEmpty) {
      throw MalformedExprException("Lambda expression must have at least one parameter")
    } else if (params.size == 1) {
      AlgoLambda(params.head, body)
    } else {
      AlgoLambda(params.tail, AlgoLambda(params.head, body))
    }
  }
}

/**
  * root trait of all expression in the algo package / in the AlgoIR
  */
sealed trait AlgoExpr extends BuiltinExpr with AlgoIR

/*
 ***********************************************************************************************************************
 value/input generators
 ***********************************************************************************************************************
 */

case class InputExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): InputExpr = InputExpr(newInnerIR, newTypes)
}
object Input {
  def apply(identifier: TextTypeT, elementType: AlgoDataTypeT, dimensions: Seq[ArithTypeT]): InputExpr =
    InputExpr(Value(SeqType(dimensions, elementType)), Seq(identifier, elementType) ++ dimensions)
  def unapply(expr: InputExpr): Some[(TextTypeT, AlgoDataTypeT, Seq[ArithTypeT], Type)] =
    Some((expr.types.head.asInstanceOf[TextTypeT], expr.types(1).asInstanceOf[AlgoDataTypeT], expr.types.tail.tail.map(_.asInstanceOf[ArithTypeT]), expr.t))

  def apply(identifier: String, elementType: AlgoDataTypeT, numElements: ArithTypeT): InputExpr =
    apply(TextType(identifier), elementType, Seq(numElements))
  def apply(identifier: String, numInnerElements: ArithTypeT, numOuterElements: ArithTypeT, elementType: AlgoDataTypeT): InputExpr =
    apply(TextType(identifier), elementType, Seq(numInnerElements, numOuterElements))
  def apply(identifier: String, elementType: AlgoDataTypeT, numInnerElements: ArithTypeT, numOuterElements: ArithTypeT): InputExpr =
    apply(TextType(identifier), elementType, Seq(numInnerElements, numOuterElements))
}

case class StoreResultExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): StoreResultExpr = StoreResultExpr(newInnerIR)
}
object StoreResult {
  def apply(input: Expr): StoreResultExpr = StoreResultExpr({
    val dtv = AlgoDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, AlgoFunType(dtv, dtv))), input.t), input)
  })
  def unapply(expr: StoreResultExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class BufferExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): BufferExpr = BufferExpr(newInnerIR)
}
object Buffer {
  def apply(input: Expr): BufferExpr = BufferExpr({
    val dtv = AlgoDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, AlgoFunType(dtv, dtv))), input.t), input)
  })
  def unapply(expr: BufferExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class ConstantIntegerExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ConstantIntegerExpr =
    ConstantInteger(newTypes.head.asInstanceOf[ArithTypeT], Some(newTypes(1)))
}
object ConstantInteger {
  def apply(value: ArithTypeT, valueType: Option[Type] = None): ConstantIntegerExpr = {
    val requiredBits = value.ae.evalInt.toBinaryString.length
    valueType match {
      case Some(tv: TypeVarT) => ConstantIntegerExpr(Value(tv), Seq(value, tv))
      case Some(t: IntTypeT) if t.width.ae.evalInt >= requiredBits => ConstantIntegerExpr(Value(t), Seq(value, t))
      case Some(_: IntTypeT) => throw MalformedTypeException("constant value requires higher precision type")
      case Some(t) => throw MalformedTypeException("constant value requires int data type; found type: " + t)
      case None =>
        val intType = IntType(requiredBits)
        ConstantIntegerExpr(Value(intType), Seq(value, intType))
    }
  }
  def unapply(expr: ConstantIntegerExpr): Some[(ArithTypeT, Type)] =
    Some(expr.types.head.asInstanceOf[ArithTypeT], expr.t)
}

case class CounterIntegerExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): CounterIntegerExpr = CounterIntegerExpr(newInnerIR, newTypes)
}
object CounterInteger {
  def apply(start: ArithTypeT, increment: ArithTypeT, numElements: ArithTypeT): CounterIntegerExpr =
    apply(start, increment, 0, Seq(numElements), Seq(0))
  def apply(start: ArithTypeT, increment: ArithTypeT, dimensions: Seq[ArithTypeT], valueType: Option[Type]): CounterIntegerExpr =
    apply(start, increment, 0, dimensions, Seq.fill(dimensions.size)(0), valueType)
  def apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], repetitions: Seq[ArithTypeT]): CounterIntegerExpr =
    apply(start, increment, loop, dimensions, repetitions, None)

  /**
    * @param start initial value of the counter
    * @param increment step size to increment counter
    * @param loop counter is just counting up once (if value is 0), counter is looping infinitely (otherwise)
    * @param dimensions sequence of arithmetic values with the innermost dimension first and the outermost dimension last
    * @param repetitions if value at position N is 1, the Nth dimension will repeat the values instead of incrementing them
    *                    this is similar to a combination of the Counter and Repeat primitives
    * @param valueType data type for the counter's output (e.g. 8-bit int)
    */
  def apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], repetitions: Seq[ArithTypeT], valueType: Option[Type]): CounterIntegerExpr = {
    val requiredBits = (start.ae + dimensions.fold(ArithType(1))(_.ae * _.ae).ae * increment.ae).evalInt.toBinaryString.length
    valueType match {
      case Some(t: IntTypeT) if t.width.ae.evalInt >= requiredBits => _apply(start, increment, loop, dimensions, repetitions, t)
      case Some(_: IntTypeT) => throw MalformedTypeException("counter requires higher precision type")
      case Some(_) => throw MalformedTypeException("counter requires int data type")
      case None => _apply(start, increment, loop, dimensions, repetitions, IntType(requiredBits))
    }
  }
  private def _apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], repetitions: Seq[ArithTypeT], valueType: AlgoDataTypeT): CounterIntegerExpr = {
    assert(dimensions.length == repetitions.length)
    CounterIntegerExpr(Value(SeqType(dimensions, valueType)),
      Seq(start, increment, loop) ++ dimensions ++ repetitions)
  }

  def unapply(expr: CounterIntegerExpr): Some[(ArithTypeT, ArithTypeT, ArithTypeT, Seq[ArithTypeT], Seq[ArithTypeT], SeqTypeT)] = {
    val seqs = expr.types.tail.tail.tail.map(_.asInstanceOf[ArithTypeT])
    val (dimensions, repetitions) = seqs.splitAt(seqs.length / 2)
    Some(
      expr.types.head.asInstanceOf[ArithTypeT],
      expr.types(1).asInstanceOf[ArithTypeT],
      expr.types(2).asInstanceOf[ArithTypeT],
      dimensions,
      repetitions,
      expr.t.asInstanceOf[SeqTypeT]
    )
  }
}

case class TruncIntegerExpr private[algo](innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TruncIntegerExpr = TruncIntegerExpr(newInnerIR, newTypes)
}
object TruncInteger extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, targetBitWidth: ArithTypeT): TruncIntegerExpr = TruncIntegerExpr({
    val dtv = AlgoDataTypeVar()
    val outt = IntType(targetBitWidth)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, AlgoFunType(dtv, outt))), input.t), input)
  }, Seq(targetBitWidth))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): TruncIntegerExpr =
    apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: TruncIntegerExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class RepeatExpr private[algo](innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): RepeatExpr = RepeatExpr(newInnerIR, newTypes)
}
object Repeat extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, len: ArithTypeT): RepeatExpr = RepeatExpr({
    val dtv = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val outseqt = SeqType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv, lentv), AlgoFunType(dtv, outseqt))), Seq(input.t, len)), input)
  }, Seq(len))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): RepeatExpr =
    apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: RepeatExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class SinkExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SinkExpr = SinkExpr(newInnerIR)
}
object Sink {
  def apply(input: Expr): SinkExpr = SinkExpr({
    val dtv = AlgoDataTypeVar()
    val fixedOutType = IntType(1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, AlgoFunType(dtv, fixedOutType))), input.t), input)
  })
  def unapply(expr: SinkExpr): Some[(Expr, Type)] =
    Some(expr.args.head, expr.t)
}

/*
 ***********************************************************************************************************************
 simple operations
 ***********************************************************************************************************************
 */

case class IdExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): IdExpr = IdExpr(newInnerIR)
}
object Id extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): IdExpr = IdExpr({
    val dtv = AlgoDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, AlgoFunType(dtv, dtv))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): IdExpr = apply(inputs.head)
  def unapply(expr: IdExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class SignedExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SignedExpr = SignedExpr(newInnerIR)
}
object Signed extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): SignedExpr = SignedExpr({
    val width = ArithTypeVar()
    val itv = IntTypeVar(width)
    val outt = SignedIntType(width + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(itv, AlgoFunType(itv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SignedExpr = apply(inputs.head)
  def unapply(expr: SignedExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class UnSignedExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): UnSignedExpr = UnSignedExpr(newInnerIR)
}
object Unsigned extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): UnSignedExpr = UnSignedExpr({
    val width = ArithTypeVar()
    val itv = SignedIntTypeVar(width)
    val outt = IntType(width - 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(itv, AlgoFunType(itv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): UnSignedExpr = apply(inputs.head)
  def unapply(expr: UnSignedExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class AddExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): AddExpr = AddExpr(newInnerIR)
}
object Add extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): AddExpr = AddExpr({
    val ttv = TupleTypeVar(AbstractScalarTypeVar(), AbstractScalarTypeVar())
    val outt =
      ConditionalTypeVar(
        ttv,
        {
          case TupleType(Seq(it1: IntTypeT, it2: IntTypeT)) =>
            IntType(
              SimplifyIfThenElse(
                it1.width.ae.gt(it2.width.ae),
                it1.width.ae,
                SimplifyIfThenElse(
                  it1.width.ae.eq(it2.width.ae),
                  it1.width.ae + 1,
                  it2.width.ae
                )
              )
            )
          case TupleType(Seq(ft1: FloatTypeT, ft2: FloatTypeT)) if ft1 == ft2 => ft1
        }
      )
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, AlgoFunType(ttv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddExpr = apply(inputs.head)
  def unapply(expr: AddExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object Inc extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): AddExpr = Add(Tuple(input, ConstantInteger(1)))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddExpr = apply(inputs.head)
}

object Add2 extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): AddExpr = Add(Tuple(input1, input2))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddExpr = apply(inputs.head, inputs(1))
}

case class SubExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SubExpr = SubExpr(newInnerIR)
}
object Sub extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): SubExpr = SubExpr({
    val itv1 = SignedIntTypeVar()
    val itv2 = SignedIntTypeVar()
    val ttv = TupleTypeVar(itv1, itv2)
    val outit = SignedIntType(
      SimplifyIfThenElse(
        itv1.width.ae.gt(itv2.width.ae),
        itv1.width.ae,
        itv2.width.ae
      )
    )
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, AlgoFunType(ttv, outit))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SubExpr = apply(inputs.head)
  def unapply(expr: SubExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class MulExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MulExpr = MulExpr(newInnerIR)
}
object Mul extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): MulExpr = MulExpr({
    val ttv = TupleTypeVar(AbstractScalarTypeVar(), AbstractScalarTypeVar())
    val outt = ConditionalTypeVar(
      ttv,
      {
        case TupleType(Seq(it1: IntTypeT, it2: IntTypeT)) => IntType(it1.width.ae + it2.width.ae)
        case TupleType(Seq(ft1: FloatTypeT, ft2: FloatTypeT)) if ft1 == ft2 => ft1
      }
    )
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, AlgoFunType(ttv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MulExpr = apply(inputs.head)
  def unapply(expr: MulExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class MaxExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MaxExpr = MaxExpr(newInnerIR)
}
object Max extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): MaxExpr = MaxExpr({
    val elementtv = AbstractScalarTypeVar()
    val inttv = TupleTypeVar(elementtv, elementtv)
    val outt = inttv.children(0)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, AlgoFunType(inttv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxExpr = apply(inputs.head)
  def unapply(expr: MaxExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object Max2 extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): MaxExpr = Max(Tuple(input1, input2))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxExpr = apply(inputs.head, inputs(1))
}

object ReLU extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): MaxExpr = Max2(input, ConstantInteger(0, Some(input.t)))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxExpr = apply(inputs.head)
}

case class SelectExpr private[algo](innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SelectExpr = SelectExpr(newInnerIR, newTypes)
}
object Select extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, selection: ArithTypeT): SelectExpr = SelectExpr({
    // TODO allow more than 2 elements in a tuple
    val inttv = TupleTypeVar(AlgoDataTypeVar(), AlgoDataTypeVar())
    val outt = inttv.children(selection.ae.evalInt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, AlgoFunType(inttv, outt))), input.t), input)
  }, Seq(selection))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SelectExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SelectExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class Select3Expr private[algo](innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Select3Expr = Select3Expr(newInnerIR, newTypes)
}
object Select3 extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, selection: ArithTypeT): Select3Expr = Select3Expr({
    // TODO allow more than 2 elements in a tuple
    val inttv = TupleTypeVar(AlgoDataTypeVar(), AlgoDataTypeVar(), AlgoDataTypeVar())
    val outt = inttv.children(selection.ae.evalInt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, AlgoFunType(inttv, outt))), input.t), input)
  }, Seq(selection))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Select3Expr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: Select3Expr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

/*
 ***********************************************************************************************************************
 complex operations (map, reduce, slide, ...)
 ***********************************************************************************************************************
 */

case class MapExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MapExpr = MapExpr(newInnerIR)
}
object Map extends BuiltinExprFun {
  override def args: Int = 2
  def apply(function: Expr, input: Expr): MapExpr = MapExpr({
    val indtv = AlgoDataTypeVar()
    val outdtv = AlgoDataTypeVar()
    val ftv = FunTypeVar(indtv, outdtv)
    val lentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(indtv, lentv)
    val outseqt = SeqType(outdtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(ftv, inseqtv), AlgoFunType(Seq(ftv, inseqtv), outseqt))), Seq(function.t, input.t)), Seq(function, input))
  })
  def apply(nestingLevels: Int, function: Expr, input: Expr): Expr = {
    if (nestingLevels <= 0) {
      FunctionCall(function, input)
    } else if (nestingLevels == 1) {
      apply(function, input)
    } else {
      val inputtc = TypeChecker.checkIfUnknown(input, input.t)
      val et = inputtc.t match {
        case st: SeqTypeT => st.et
        case _ => throw MalformedExprException(s"Cannot create $nestingLevels times nested Maps for input of type ${inputtc.t}")
      }
      Map(
        {
          val p = ParamDef(et)
          AlgoLambda(p, apply(nestingLevels - 1, function, ParamUse(p)))
        },
        input
      )
    }
  }

  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MapExpr = apply(inputs.head, inputs(1))
  def unapply(expr: MapExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class FoldExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): FoldExpr = FoldExpr(newInnerIR)
}
object Fold extends BuiltinExprFun {
  override def args: Int = 2
  def apply(function: Expr, input: Expr): FoldExpr = FoldExpr({
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case SeqType(_: IntTypeT, _) =>
        //    val dtv = AbstractScalarTypeVar()
        val dtv = IntTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val inseqtv = SeqTypeVar(dtv, lentv)
        var outBitWidth: ArithTypeT = dtv.width.ae + ceil(arithmetic.Log(2, lentv))
        function.visit{
          case Max(_, _) => outBitWidth = dtv.width.ae
          case _ =>
        }
        val outt = IntType(outBitWidth)
        //    ConditionalTypeVar(
        //      dtv,
        //      {
        //        case it: IntTypeT => IntType(it.width.ae + ceil(arithmetic.Log(2, lentv)))
        //         TODO "... + Log(2 ..." if function is ADD; "... * Log(2 ..." if function is MUL
        //        case ft: FloatTypeT => ft
        //      }
        //    )
        val ftv = FunTypeVar(Seq(outt, dtv), AbstractScalarTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, inseqtv),
            AlgoFunType(
              Seq(ftv, inseqtv),
              outt
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
      case SeqType(SeqType(_: IntTypeT, _), _) =>
        val innerLenTv = ArithTypeVar(StartFromRange(1))
        val itv = IntTypeVar()
        val dtv = SeqTypeVar(itv, innerLenTv)
        val outerLenTv = ArithTypeVar(StartFromRange(1))
        val instv = SeqTypeVar(dtv, outerLenTv)
        var outBitWidth: ArithTypeT = itv.width.ae + ceil(arithmetic.Log(2, outerLenTv))
        function.visit{
          case Max(_, _) => outBitWidth = itv.width.ae
          case _ =>
        }
        val outt = SeqType(IntType(outBitWidth), innerLenTv)
        val ftv = FunTypeVar(Seq(outt, dtv), SeqTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, instv),
            AlgoFunType(
              Seq(ftv, instv),
              outt
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
      case SeqType(AlgoDataTypeVar(_), _) =>
        val dtv = IntTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val inseqtv = SeqTypeVar(dtv, lentv)
        var outBitWidth: ArithTypeT = dtv.width.ae + ceil(arithmetic.Log(2, lentv))
        function.visit{
          case Max(_, _) => outBitWidth = dtv.width.ae
          case _ =>
        }
        val outt = IntType(outBitWidth)
        val ftv = FunTypeVar(Seq(outt, dtv), AbstractScalarTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, inseqtv),
            AlgoFunType(
              Seq(ftv, inseqtv),
              outt
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
    }
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): FoldExpr = apply(inputs.head, inputs(1))
  def unapply(expr: FoldExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class ReduceExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ReduceExpr = ReduceExpr(newInnerIR)
}
object Reduce extends BuiltinExprFun {
  override def args: Int = 3
  def apply(function: Expr, initialValue: Expr, input: Expr): ReduceExpr = ReduceExpr({
    val initialValuetc = TypeChecker.checkIfUnknown(initialValue, initialValue.t)
    initialValuetc.t match {
      case _: IntTypeT =>
        val dtv = IntTypeVar()
        val initialValueTv = IntTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val inseqtv = SeqTypeVar(dtv, lentv)
        var outBitWidth: ArithTypeT = dtv.width.ae + ceil(arithmetic.Log(2, lentv))
        function.visit{
          case Max(_, _) => outBitWidth = dtv.width.ae
          case _ =>
        }
        val outt = IntType(outBitWidth)
        // TODO "... + Log(2 ..." if function is ADD; "... * Log(2 ..." if function is MUL
        val ftv = FunTypeVar(Seq(outt, dtv), IntTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, initialValueTv, inseqtv),
            AlgoFunType(
              Seq(ftv, initialValueTv, inseqtv),
              outt
            )
          )),
          Seq(function.t, initialValue.t, input.t)),
          Seq(function, initialValue, input)
        )
      case _ =>
        val dtv = AlgoDataTypeVar()
        val acctv = AlgoDataTypeVar()
        val ftv = FunTypeVar(Seq(acctv, dtv), acctv)
        val inseqtv = SeqTypeVar(dtv)
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, acctv, inseqtv),
            AlgoFunType(
              Seq(ftv, acctv, inseqtv),
              acctv
            )
          )),
          Seq(function.t, initialValue.t, input.t)),
          Seq(function, initialValue, input)
        )
    }
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ReduceExpr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: ReduceExpr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

object DotProduct extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): BuiltinExpr =
    Fold(
      Add2.asFunction(),
      Map(
        Mul.asFunction(),
        Zip(Tuple(input1, input2))
      )
    )
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head, inputs(1))
}

object VAdd extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): BuiltinExpr =
    Map(
      Add.asFunction(),
      Zip(Tuple(input1, input2))
    )
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head, inputs(1))
}

object MVMul extends BuiltinExprFun {
  override def args: Int = 2
  def apply(matrix: Expr, vector: Expr): BuiltinExpr =
    Map(
      DotProduct.asFunction(Seq(Some(vector), None)),
      matrix
    )
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head, inputs(1))
}

object MMMul extends BuiltinExprFun {
  override def args: Int = 2
  def apply(matrix1: Expr, matrix2: Expr): BuiltinExpr =
    Map(
      MVMul.asFunction(Seq(Some(matrix2), None)),
      matrix1
    )
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head, inputs(1))
}

case class SlideExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SlideExpr = SlideExpr(newInnerIR, newTypes)
}
object Slide extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, windowWidth: ArithTypeT): SlideExpr = SlideExpr({
    val dtv = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(dtv, lentv)
    val outseqt = SeqType(SeqType(dtv, windowWidth), lentv - windowWidth.ae + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
  }, Seq(windowWidth))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SlideExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SlideExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

object JoinAll extends BuiltinExprFun {
  override def args: Int = 1
  @tailrec
  def apply(input: Expr): BuiltinExpr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case st: SeqTypeT if
        (st.et match {
          case _: SeqTypeT => true
          case _ => false
        }) =>
        JoinAll(Join(inputtc))
      case _ =>
        inputtc.asInstanceOf[BuiltinExpr]
    }
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head)
}

object SplitAll extends BuiltinExprFun {
  override def args: Int = 1
  @tailrec
  def apply(input: Expr, dims: Seq[ArithTypeT]): BuiltinExpr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    if(dims.nonEmpty) {
      SplitAll(Split(inputtc, dims.last), dims.init)
    } else {
      inputtc.asInstanceOf[BuiltinExpr]
    }
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head, types.asInstanceOf[Seq[ArithTypeT]])
}

case class IfelseExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): IfelseExpr = IfelseExpr(newInnerIR)
}
object Ifelse extends BuiltinExprFun {
  override def args: Int = 3

  def apply(function1: Expr, function2: Expr, input: Expr): IfelseExpr = IfelseExpr({
    val seltv = IntTypeVar()
    val intv = AlgoDataTypeVar()
    val intupletv = TupleTypeVar(intv, seltv)
    val outtv = AlgoDataTypeVar()
    val f1tv = FunTypeVar(intv, outtv)
    val f2tv = FunTypeVar(intv, outtv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(f1tv, f2tv, intupletv), AlgoFunType(Seq(f1tv, f2tv, intupletv), outtv))), Seq(function1.t, function2.t, input.t)), Seq(function1, function2, input))
  })

  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): IfelseExpr = apply(inputs.head, inputs(1), inputs(2))

  def unapply(expr: IfelseExpr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

/*
 ***********************************************************************************************************************
 data type conversions / rearrangement
 ***********************************************************************************************************************
 */

case class TupleExpr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TupleExpr = TupleExpr(newInnerIR)
}
object Tuple extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): TupleExpr = TupleExpr({
    val dtv1 = AlgoDataTypeVar()
    val dtv2 = AlgoDataTypeVar()
    val outtt = TupleType(dtv1, dtv2)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv1, dtv2), AlgoFunType(Seq(dtv1, dtv2), outtt))), Seq(input1.t, input2.t)), Seq(input1, input2))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): TupleExpr = apply(inputs.head, inputs(1))
  def unapply(expr: TupleExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class ZipExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ZipExpr = ZipExpr(newInnerIR)
}
object Zip extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): ZipExpr = ZipExpr({
    val dtv1 = AlgoDataTypeVar()
    val dtv2 = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv1 = SeqTypeVar(dtv1, lentv)
    val inseqtv2 = SeqTypeVar(dtv2, lentv)
    val inttv = TupleTypeVar(inseqtv1, inseqtv2)
    val outseqt = SeqType(TupleType(dtv1, dtv2), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, AlgoFunType(inttv, outseqt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ZipExpr = apply(inputs.head)
  def unapply(expr: ZipExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class Tuple3Expr private[algo](innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Tuple3Expr = Tuple3Expr(newInnerIR)
}
object Tuple3 extends BuiltinExprFun {
  override def args: Int = 3
  def apply(input1: Expr, input2: Expr, input3: Expr): Tuple3Expr = Tuple3Expr({
    val dtv1 = AlgoDataTypeVar()
    val dtv2 = AlgoDataTypeVar()
    val dtv3 = AlgoDataTypeVar()
    val outtt = TupleType(dtv1, dtv2, dtv3)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv1, dtv2, dtv3), AlgoFunType(Seq(dtv1, dtv2, dtv3), outtt))), Seq(input1.t, input2.t, input3.t)), Seq(input1, input2, input3))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Tuple3Expr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: Tuple3Expr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

case class Zip3Expr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Zip3Expr = Zip3Expr(newInnerIR)
}
object Zip3 extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): Zip3Expr = Zip3Expr({
    val dtv1 = AlgoDataTypeVar()
    val dtv2 = AlgoDataTypeVar()
    val dtv3 = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv1 = SeqTypeVar(dtv1, lentv)
    val inseqtv2 = SeqTypeVar(dtv2, lentv)
    val inseqtv3 = SeqTypeVar(dtv3, lentv)
    val inttv = TupleTypeVar(inseqtv1, inseqtv2, inseqtv3)
    val outseqt = SeqType(TupleType(dtv1, dtv2, dtv3), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, AlgoFunType(inttv, outseqt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Zip3Expr = apply(inputs.head)
  def unapply(expr: Zip3Expr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class SplitExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SplitExpr = SplitExpr(newInnerIR, newTypes)
}
object Split extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, chunkSize: ArithTypeT): SplitExpr = apply(input, chunkSize, innerSize = true)
  def apply(input: Expr, size: ArithTypeT, innerSize: Boolean): SplitExpr = {
    val dtv = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(dtv, lentv)

    val (chunkSize: ArithExpr, numberOfChunks: ArithExpr) =
      if (innerSize)
        (size.ae, lentv.ae / size.ae)
      else
        (lentv.ae / size.ae, size.ae)

    SplitExpr({
      val outseqt = SeqType(SeqType(dtv, chunkSize), numberOfChunks)
      FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
    }, Seq(chunkSize))
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SplitExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SplitExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class JoinExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): JoinExpr = JoinExpr(newInnerIR)
}
object Join extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): JoinExpr = JoinExpr({
    val dtv = AlgoDataTypeVar()
    val innerlentv = ArithTypeVar()
    val outerlentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(SeqTypeVar(dtv, innerlentv), outerlentv)
    val outseqt = SeqType(dtv, innerlentv.ae * outerlentv.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): JoinExpr = apply(inputs.head)
  def unapply(expr: JoinExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class DropExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): DropExpr = DropExpr(newInnerIR, newTypes)
}
object Drop extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, firstElements: ArithTypeT, lastElements: ArithTypeT): DropExpr = DropExpr({
    val dtv = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(dtv, lentv)
    val outseqt = SeqType(dtv, lentv.ae - firstElements.ae - lastElements.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
  }, Seq(firstElements, lastElements))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): DropExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT], types(1).asInstanceOf[ArithTypeT])
  def unapply(expr: DropExpr): Some[(Expr, ArithTypeT, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.types(1).asInstanceOf[ArithTypeT], expr.t))
}

case class ConcatExpr private[algo] (innerIR: Expr) extends AlgoExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ConcatExpr = ConcatExpr(newInnerIR)
}
object Concat extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): ConcatExpr = ConcatExpr({
    val dtv = AlgoDataTypeVar()
    val lentv1 = ArithTypeVar()
    val lentv2 = ArithTypeVar()
    val inseqtv1 = SeqTypeVar(dtv, lentv1)
    val inseqtv2 = SeqTypeVar(dtv, lentv2)
    val outseqt = SeqType(dtv, lentv1 + lentv2)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(inseqtv1, inseqtv2), AlgoFunType(Seq(inseqtv1, inseqtv2), outseqt))), Seq(input1.t, input2.t)), Seq(input1, input2))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ConcatExpr = apply(inputs.head, inputs(1))
  def unapply(expr: ConcatExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class PermuteExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): PermuteExpr = PermuteExpr(newInnerIR, newTypes)
}
object Permute extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, permuteFun: ArithLambdaType): PermuteExpr = PermuteExpr({
    val dtv = AlgoDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv = SeqTypeVar(dtv, lentv)
    val outseqt = SeqType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
  }, Seq(permuteFun))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): PermuteExpr = apply(inputs.head, types.head.asInstanceOf[ArithLambdaType])
  def unapply(expr: PermuteExpr): Some[(Expr, ArithLambdaType, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithLambdaType], expr.t))
}

// Dummy primitive
case class TransposeNDExpr private[algo] (innerIR: Expr, types: Seq[Type]) extends AlgoExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TransposeNDExpr = TransposeNDExpr(newInnerIR, newTypes)
}
object TransposeND extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, dimOrder: Seq[ArithTypeT]): TransposeNDExpr = TransposeNDExpr({
    def genReorderedSeqType(dimOrder: Seq[ArithTypeT], lentvSeq: Seq[ArithTypeT], dtv: AlgoDataTypeVar): SeqType = {
      assert(dimOrder.nonEmpty, "dimOrder must have at least one element")
      if (dimOrder.length <= 1) {
        SeqType(dtv, lentvSeq(dimOrder.last.ae.evalInt))
      } else {
        SeqType(genReorderedSeqType(dimOrder.init, lentvSeq, dtv), lentvSeq(dimOrder.last.ae.evalInt))
      }
    }

    val dtv = AlgoDataTypeVar()
    val lentvSeq = Seq.fill(dimOrder.length)(ArithTypeVar())
    val inseqtv = SeqTypeVar(lentvSeq, dtv)
    val outseqt = genReorderedSeqType(dimOrder, lentvSeq, dtv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, AlgoFunType(inseqtv, outseqt))), input.t), input)
  }, dimOrder)
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): TransposeNDExpr = apply(inputs.head, types.asInstanceOf[Seq[ArithTypeT]])
  def unapply(expr: TransposeNDExpr): Some[(Expr, Seq[ArithTypeT], Type)] =
    Some((expr.args.head, expr.types.asInstanceOf[Seq[ArithTypeT]], expr.t))
}

object Transpose extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): BuiltinExpr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val height = inputtc.t.asInstanceOf[SeqType].len.ae.evalInt
    val width = inputtc.t.asInstanceOf[SeqType].et.asInstanceOf[SeqType].len.ae.evalInt
    val v = ArithTypeVar()
    Split(Permute(Join(inputtc), ArithLambdaType(v, v/width + (v%width) * height)), height)

  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head)
}