package backend.hdl.arch

import algo.{SeqTypeT, StoreResult}
import backend.hdl._
import backend.hdl.arch.mem._
import backend.hdl.specs.FPGA
import core._
import core.compile.{CompilerPass, CompilerPhase}

object ArchCompiler extends CompilerPass {

  override def phaseBefore: CompilerPhase = CompilerPhase.first()

  override def run(expr: Expr): Expr =
    TypeChecker.check(
      _generate(
        TypeChecker.checkAssert(
          expr match {
            case StoreResult(_, _) => expr
            // add parent 'StoreResult' expression, if not available and host ram is accessed
            case _ if hasHostRamInput(expr) => StoreResult(expr)
            case _ => expr
          }
        )
      )
    )

  private def hasHostRamInput(expr: Expr): Boolean = {
    expr.visit({
      case algo.Input(_, _, _, _) => return true
      case _ =>
    })
    false
  }

  private def _generate(expr: Expr): Expr = expr match {
    case archExpr: ArchExpr => archExpr
    case memExpr: MemExpr => memExpr
    case algoExpr: algo.AlgoExpr => algoExpr match {
      case algo.Input(identifier, _, desiredDimensions: Seq[ArithTypeT], t) =>
        val desiredDataType = _generate(t).asInstanceOf[UnorderedStreamTypeT]

        val desiredRowType = UnorderedStreamType(desiredDataType.leafType, desiredDimensions.head)
        val cacheLinesPerRow = ArchConversion.calcLength(desiredRowType, FPGA.current.cachelineType)
        val dimensions = cacheLinesPerRow +: desiredDimensions.tail // replace number of elements per row by number of cachelines required to fit an entire row

        val mem = HostRam(identifier, TextType(t + ""), desiredDataType.leafType, dimensions)
        val mapRead = TypeChecker.check(
          // TODO buffer here tends to get better performance (because larger elements are stored)
          Read(mem)
        )

        Marker(
          ArchConversion(
            Marker(
              mapRead,
              TextType("input_" + identifier.s)
            ),
            mapRead.t,
            desiredDataType
          ),
          TextType("map_fusion_guard")
        )
      case algo.StoreResult(input, _) =>
        val archExpr = TypeChecker.check(_generate(input))
        // wrap into a stream if data type is not a stream (writing back data requires a stream type)
        val archExprWrapped = TypeChecker.check(archExpr.t match {
          //        case _: OrderedStreamTypeT => OrderedStreamToUnorderedStream(archExpr)
          case _: UnorderedStreamTypeT => archExpr
          case _ => Repeat(archExpr, 1)
        })

        val givenType = archExprWrapped.t.asInstanceOf[UnorderedStreamTypeT]

        val givenRowType = UnorderedStreamType(givenType.leafType, givenType.dimensions.head)
        val cacheLinesPerRow = ArchConversion.calcLength(givenRowType, FPGA.current.cachelineType)

        // replace number of elements per row by number of cachelines required to fit an entire row
        val typeForMemory = OrderedStreamType(givenType.dimensions.tail, UnorderedStreamType(FPGA.current.cachelineType, cacheLinesPerRow))

        val convertedInputData =
          IndexStream(
            ArchConversion(archExprWrapped, archExprWrapped.t, typeForMemory)
          )
        val mem = HostRam(TextType(MemoryImage.RESULT_VAR_NAME), TextType(input.t + ""), givenType.leafType, typeForMemory.dimensions)

        Write(mem, convertedInputData)
      case algo.Buffer(input, _) => BufferStreamInBlockRam(_generate(input))
      case algo.ConstantInteger(value, valueType) => ConstantValue(value, Some(_generate(valueType)))
      case algo.CounterInteger(start, increment, loop, dimensions, repetitions, valueType) => CounterInteger(start, increment, loop, dimensions, repetitions, Some(_generate(valueType.leafType)))
      case algo.TruncInteger(input, targetBitWidth, _) => TruncInteger(_generate(input), targetBitWidth)
      case algo.Repeat(input, len, _) => Repeat(_generate(input), len)
      case algo.Sink(input, _) => Sink(_generate(input))
      case algo.Id(input, _) => Id(_generate(input))
      case algo.Signed(input, _) => Signed(_generate(input))
      case algo.Unsigned(input, _) => Unsigned(_generate(input))
      case algo.Add(input, _: algo.IntTypeT) => AddInt(_generate(input))
      case algo.Add(input, _: algo.FloatTypeT) => AddFloat(_generate(input))
      case algo.Add(_, _) => throw MalformedExprException("Unknown Type for Add")
      case algo.Sub(input, _) => SubInt(_generate(input))
      case algo.Mul(input, _: algo.IntTypeT) => MulInt(_generate(input))
      case algo.Mul(input, _: algo.FloatTypeT) => MulFloat(_generate(input))
      case algo.Mul(_, _) => throw MalformedExprException("Unknown Type for Mul")
      case algo.Max(input, _: algo.IntTypeT) => MaxInt(_generate(input))
      case algo.Max(_, _) => throw MalformedExprException("Unknown Type for Max")
      case algo.Select(input, selection, _) => Select(_generate(input), selection)
      case algo.Select3(input, selection, _) => Select3(_generate(input), selection)
      case algo.Map(f, input, _) => MapOrderedStream(_generate(f), _generate(input))
      case algo.Fold(f, input, _) => FoldOrderedStream(_generate(f), _generate(input))
      case algo.Reduce(f, initialValue, input, _) => ReduceOrderedStream(_generate(f), _generate(initialValue), _generate(input))
      case algo.Slide(input, windowWidth, _) => {
        val innerType = input.t.asInstanceOf[SeqTypeT].et
        if (innerType.isInstanceOf[SeqTypeT] ) {
          val slideSplit = SlideOrderedStream(MapOrderedStream(OrderedStreamToVector.asFunction(), _generate(input)), windowWidth)
          val slideInnerToStm = MapOrderedStream(VectorToOrderedStream.asFunction(), slideSplit)
          MapOrderedStream(MapOrderedStream.asFunction(Seq(Some(VectorToOrderedStream.asFunction()), None)), slideInnerToStm)
        } else {
          MapOrderedStream(VectorToOrderedStream.asFunction(), SlideOrderedStream(_generate(input), windowWidth))
        }
      }
      case algo.Tuple(input1, input2, _) => Tuple(_generate(input1), _generate(input2))
      case algo.Zip(input, _) => ZipOrderedStream(_generate(input))
      case algo.Tuple3(input1, input2, input3, _) => Tuple3(_generate(input1), _generate(input2), _generate(input3))
      case algo.Zip3(input, _) => Zip3OrderedStream(_generate(input))
      case algo.Split(input, chunkSize, _) => SplitOrderedStream(_generate(input), chunkSize)
      case algo.Join(input, _) => JoinStream(_generate(input))
      case algo.Drop(input, firstElements, lastElements, _) => DropOrderedStream(_generate(input), firstElements, lastElements)
      case algo.Concat(input1, input2, _) => ConcatOrderedStream(_generate(input1), _generate(input2))
      case algo.Permute(input, arith, _) => {
        val innerType = input.t.asInstanceOf[SeqTypeT].et
        if(innerType.isInstanceOf[SeqTypeT]) {
          val inputInnerVec = MapOrderedStream(OrderedStreamToVector.asFunction(), _generate(input))
          val permuted = VectorToOrderedStream(PermuteVector(OrderedStreamToVector(inputInnerVec), arith))
          MapOrderedStream(VectorToOrderedStream.asFunction(), permuted)
        } else {
          VectorToOrderedStream(PermuteVector(OrderedStreamToVector(_generate(input)), arith))
        }
      }
      case algo.TransposeND(input, dimOrder, _) => TransposeNDOrderedStream(_generate(input), dimOrder)
      case algo.Ifelse(f1, f2, input, _) => Ifelse(_generate(f1), _generate(f2), _generate(input))
    }
    case algo.AlgoLambda(param, body, _) => ArchLambda(_generate(param).asInstanceOf[ParamDef], _generate(body))
    case coreExpr: CoreExpr => coreExpr.build(coreExpr.children.map({
      case e: Expr => _generate(e)
      case t: Type => _generate(t)
    }))
    case other => throw MalformedExprException("cannot generate Arch expression from " + other)
  }

  private def _generate(t: Type): Type = t match {
    case hwDataType: HWDataTypeT => hwDataType
    case algoType: algo.AlgoDataTypeT => _generateHWDataType(algoType)
    case algo.AlgoFunType(inType, outType) => HWFunType(_generate(inType), _generate(outType))
    case coreType: CoreType => coreType.build(coreType.children.map(_generate))
    case other => other //throw new MalformedExprException("cannot generate HW type from " + other)
  }

  private def _generateHWDataType(t: algo.AlgoDataTypeT): HWDataTypeT = t match {
    case _: TypeVarT => throw MalformedExprException("cannot generate IR due to remaining type var in expression: " + t)
      // TODO generate vector / or stream type here
    case algo.SeqType(et, len) => OrderedStreamType(_generateHWDataType(et).asInstanceOf[NonArrayTypeT], len)
    case algo.NamedTupleType(namedTypes) => NamedTupleType(namedTypes.map(p => (p._1, _generateHWDataType(p._2))))//asInstanceOf[BasicDataTypeT])))
    case algo.SignedIntType(width) => SignedIntType(width)
    case algo.IntType(width) => IntType(width)
    case algo.FloatHalfType() => FloatHalfType()
    case algo.FloatSingleType() => FloatSingleType()
    case algo.FloatDoubleType() => FloatDoubleType()
    case algo.AlgoDataType() | algo.AbstractScalarType() | algo.FloatType() | algo.AbstractCollectionType() =>
      throw MalformedExprException("Cannot generate Arch type from abstract type")
  }
}
