package backend.hdl.arch

import backend.hdl._
import core._
import lift.arithmetic
import lift.arithmetic.simplifier.SimplifyIfThenElse
import lift.arithmetic.{ArithExpr, StartFromRange, ceil}

import scala.annotation.tailrec

final case class ArchLambda private (param: ParamDef, body: Expr, t: FunTypeT) extends LambdaT {
  override def build(newChildren: Seq[ShirIR]): ArchLambda = ArchLambda(newChildren.head.asInstanceOf[ParamDef], newChildren(1).asInstanceOf[Expr], newChildren(2).asInstanceOf[FunTypeT])
}

object ArchLambda {
  private[arch] def apply(param: ParamDef, body: Expr, syncType: CommunicationTypeT = SyncCommunication, schedulingStrategy: SchedulingStrategyTypeT = DefaultScheduling()): ArchLambda = ArchLambda(param, body, HWFunType(param.t, body.t, syncType, schedulingStrategy))
  def unapply(expr: Expr): Option[(ParamDef, Expr, Type)] = expr match {
    case l: LambdaT => Some(l.param, l.body, expr.t)
    case _ => None
  }
}

object ArchLambdas {
  @tailrec
  private[arch] def apply(params: Seq[ParamDef], body: Expr, syncType: CommunicationTypeT = SyncCommunication, schedulingStrategy: SchedulingStrategyTypeT = DefaultScheduling()): ArchLambda = {
    if (params.isEmpty) {
      throw MalformedExprException("Lambda expression must have at least one parameter")
    } else if (params.size == 1) {
      ArchLambda(params.head, body, syncType, schedulingStrategy)
    } else {
      ArchLambdas(params.tail, ArchLambda(params.head, body, syncType, schedulingStrategy), syncType, schedulingStrategy)
    }
  }
}

object ArchLet {
  private[arch] def apply(param: ParamDef, body: Expr, arg: Expr, syncType: CommunicationTypeT = SyncCommunication, schedulingStrategy: SchedulingStrategyTypeT = DefaultScheduling(), t: Type = UnknownType): Expr =
    FunctionCall(ArchLambda(param, body, HWFunType(param.t, body.t, syncType, schedulingStrategy)), arg, t, applyBetaReduction = false)

  def unapply(expr: FunctionCall): Option[(ParamDef, Expr, Expr, CommunicationTypeT, SchedulingStrategyTypeT, Type)] = expr.f match {
    case ArchLambda(param, body, funt: HWFunTypeT) => Some(param, body, expr.arg, funt.comType, funt.schedulingStrategy, expr.t)
    case l: LambdaT => Some(l.param, l.body, expr.arg, SyncCommunication, DefaultScheduling(), expr.t)
    case _ => None
  }
}

/**
  * root trait of all expression in the backend.hdl.arch package
  */
sealed trait ArchExpr extends BuiltinExpr with ArchIR

/*
 ***********************************************************************************************************************
 value/input generators
 ***********************************************************************************************************************
 */

case class ConstantValueExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ConstantValueExpr =
    ConstantValue(newTypes.head.asInstanceOf[ArithTypeT], Some(newTypes(1)))
}
object ConstantValue {
  private[arch] def apply(value: ArithTypeT, valueType: Option[Type] = None): ConstantValueExpr = {
    val requiredBits = if(value.ae.evalInt < 0)
      math.abs(value.ae.evalInt).toBinaryString.length + 1
    else
      value.ae.evalInt.toBinaryString.length
    valueType match {
      case Some(tv: TypeVarT) => ConstantValueExpr(Value(tv), Seq(value, tv))
      case Some(t: HWDataTypeT) if t.bitWidth.ae.evalInt >= requiredBits => ConstantValueExpr(Value(t), Seq(value, t))
      case Some(_: HWDataTypeT) => throw MalformedTypeException("constant value requires higher precision type")
      case Some(t) => throw MalformedTypeException("constant value requires hw data type; found type: " + t)
      case Some(t: SignedIntTypeT) => throw MalformedTypeException("WaHaHa")
      case None =>
        val intType = IntType(requiredBits)
        ConstantValueExpr(Value(intType), Seq(value, intType))
    }
  }
  def unapply(expr: ConstantValueExpr): Some[(ArithTypeT, Type)] =
    Some(expr.types.head.asInstanceOf[ArithTypeT], expr.t)
}

case class CounterIntegerExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): CounterIntegerExpr = CounterIntegerExpr(newInnerIR, newTypes)
}
object CounterInteger {
  def apply(start: ArithTypeT, increment: ArithTypeT, dimensions: Seq[ArithTypeT]): CounterIntegerExpr =
    apply(start, increment, 0, dimensions, Seq.fill(dimensions.size)(0), None)
  def apply(start: ArithTypeT, increment: ArithTypeT, dimensions: Seq[ArithTypeT], valueType: Option[Type]): CounterIntegerExpr =
    apply(start, increment, 0, dimensions, Seq.fill(dimensions.size)(0), valueType)
  def apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], valueType: Option[Type]): CounterIntegerExpr =
    apply(start, increment, 0, dimensions, Seq.fill(dimensions.size)(0), valueType)

  /**
    * @param start initial value of the counter
    * @param increment step size to increment counter
    * @param loop counter is just counting up once (if value is 0), counter is looping infinitely (otherwise)
    * @param dimensions sequence of arithmetic values with the innermost dimension first and the outermost dimension last
    * @param repetitions if value at position N is 1, the Nth dimension will repeat the values instead of incrementing them
    *                    this is similar to a combination of the Counter and Repeat primitives
    * @param valueType data type for the counter's output (e.g. 8-bit int)
    */
  private[arch] def apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], repetitions: Seq[ArithTypeT], valueType: Option[Type]): CounterIntegerExpr = {
    // TODO one bit is wasted
    val requiredBits = (start.ae + dimensions.zip(repetitions).filter(_._2.ae.evalInt == 0).map(_._1).fold(ArithType(1))(_.ae * _.ae).ae * increment.ae).evalInt.toBinaryString.length
    valueType match {
      case Some(t: IntTypeT) if t.bitWidth.ae.evalInt >= requiredBits => _apply(start, increment, loop, dimensions, repetitions, t)
      case Some(_: IntTypeT) => throw MalformedTypeException("counter requires higher precision type")
      case Some(_) => throw MalformedTypeException("counter requires int data type")
      case None => _apply(start, increment, loop, dimensions, repetitions, IntType(requiredBits))
    }
  }
  private def _apply(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT], repetitions: Seq[ArithTypeT], valueType: BasicDataTypeT): CounterIntegerExpr = {
    assert(dimensions.length == repetitions.length)
    CounterIntegerExpr(Value(OrderedStreamType(dimensions, valueType)),
      Seq(start, increment, loop) ++ dimensions ++ repetitions
    )
  }

  def unapply(expr: CounterIntegerExpr): Some[(ArithTypeT, ArithTypeT, ArithTypeT, Seq[ArithTypeT], Seq[ArithTypeT], Type)] = {
    val seqs = expr.types.tail.tail.tail.map(_.asInstanceOf[ArithTypeT])
    val (dimensions, repetitions) = seqs.splitAt(seqs.length / 2)
    Some(
      expr.types.head.asInstanceOf[ArithTypeT],
      expr.types(1).asInstanceOf[ArithTypeT],
      expr.types(2).asInstanceOf[ArithTypeT],
      dimensions,
      repetitions,
      expr.t
    )
  }
}

object TruncInteger extends BuiltinExprFun {
  override def args: Int = 1

  private[arch] def apply(input: Expr, bitWidthDelta: ArithTypeT, reduceBitWidth: Boolean): Expr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    TypeFunctionCall(
      {
        val tv = IntTypeVar()
        TypeLambda(
          tv,
          {
            if (reduceBitWidth) {
              Conversion(
                DropVector(
                  Conversion(inputtc, VectorType(LogicType(), tv.bitWidth)),
                  0, bitWidthDelta
                ),
                IntType(tv.bitWidth.ae - bitWidthDelta.ae)
              )
            } else {
              Conversion(ConcatVector(Tuple(
                Conversion(inputtc, VectorType(LogicType(), tv.bitWidth)),
                VectorGenerator(ConstantValue(0, Some(LogicType())), bitWidthDelta)
              )), IntType(tv.bitWidth.ae + bitWidthDelta.ae))
            }
          }
        )
      },
      inputtc.t
    )
  }

  private[arch] def apply(input: Expr, targetBitWidth: ArithTypeT): Expr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case it: IntTypeT =>
        if (it.bitWidth.ae.evalInt >= targetBitWidth.ae.evalInt) {
          val bitsToDrop = it.bitWidth.ae - targetBitWidth.ae
          apply(inputtc, bitsToDrop, reduceBitWidth = true)
        } else {
          val bitsToAppend = targetBitWidth.ae - it.bitWidth.ae
          apply(inputtc, bitsToAppend, reduceBitWidth = false)
        }
      case _ => throw MalformedTypeException("Cannot trunc non-integer types.")
    }
  }

  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Expr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
}

case class RepeatExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): RepeatExpr = RepeatExpr(newInnerIR, newTypes)
}
object Repeat extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, dimensions: Seq[ArithTypeT]): Expr = dimensions.size match {
    case 0 => input
    case _ => Repeat(apply(input, dimensions.tail), dimensions.head)
  }
  private[arch] def apply(input: Expr, len: ArithTypeT): RepeatExpr = RepeatExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val outst = OrderedStreamType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv, lentv), HWFunType(dtv, outst))), Seq(input.t, len)), input)
  }, Seq(len))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): RepeatExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: RepeatExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class RepeatHiddenExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): RepeatHiddenExpr = RepeatHiddenExpr(newInnerIR, newTypes)
}
object RepeatHidden extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, dimensions: Seq[ArithTypeT]): Expr = dimensions.size match {
    case 0 => input
    case _ => RepeatHidden(apply(input, dimensions.tail), dimensions.head)
  }
  def apply(input: Expr, len: ArithTypeT): RepeatHiddenExpr = RepeatHiddenExpr({
    val stv = NonArrayTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(stv, HWFunType(stv, stv))), input.t), input)
  }, Seq(len))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): RepeatHiddenExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: RepeatHiddenExpr): Some[(Expr, ArithTypeT, Type)] =
    Some(expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t)
}

case class VectorGeneratorExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): VectorGeneratorExpr = VectorGeneratorExpr(newInnerIR, newTypes)
}
object VectorGenerator extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, len: ArithTypeT): VectorGeneratorExpr = VectorGeneratorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val outvt = VectorType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv, lentv), HWFunType(dtv, outvt))), Seq(input.t,len)), input)
  }, Seq(len))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): VectorGeneratorExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: VectorGeneratorExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class SinkExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SinkExpr = SinkExpr(newInnerIR)
}
object Sink {
  def apply(input: Expr): SinkExpr = SinkExpr({
    val dtv = HWDataTypeVar()
    val fixedOutType = UnitType()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, HWFunType(dtv, fixedOutType))), input.t), input)
  })
  def unapply(expr: SinkExpr): Some[(Expr, Type)] =
    Some(expr.args.head, expr.t)
}

case class RegisteredExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): RegisteredExpr = RegisteredExpr(newInnerIR)
}
object Registered extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): RegisteredExpr = RegisteredExpr({
    val dtv = BasicDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, HWFunType(dtv, dtv))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): RegisteredExpr = apply(inputs.head)
  def unapply(expr: RegisteredExpr): Some[(Expr, Type)] =
    Some(expr.args.head, expr.t)
}

case class DelayExpr private[arch](innerIR: Expr, cycles: ArithTypeT) extends ArchExpr {
  override def types: Seq[Type] = Seq(cycles)
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): DelayExpr = DelayExpr(newInnerIR, newTypes.head.asInstanceOf[ArithTypeT])
}
object Delay extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, cycles: ArithTypeT): DelayExpr = DelayExpr({
    val dtv = HWDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, HWFunType(dtv, dtv))), input.t), input)
  }, cycles)
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): DelayExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: DelayExpr): Some[(Expr, Type, Type)] =
    Some(expr.args.head, expr.cycles, expr.t)
}

case class TestBenchExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TestBenchExpr = TestBenchExpr(newInnerIR)
}
object TestBench {
  def apply(input: Expr): TestBenchExpr = TestBenchExpr({
    val dtv = HWDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, HWFunType(dtv, dtv))), input.t), input)
  })
  def unapply(expr: TestBenchExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

/*
 ***********************************************************************************************************************
 simple operations
 ***********************************************************************************************************************
 */

case class IdExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): IdExpr = IdExpr(newInnerIR)
}
object Id extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): IdExpr = IdExpr({
    val dtv = HWDataTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(dtv, HWFunType(dtv, dtv))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): IdExpr = apply(inputs.head)
  def unapply(expr: IdExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class SignedExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SignedExpr = SignedExpr(newInnerIR)
}
object Signed extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): SignedExpr = SignedExpr({
    val width = ArithTypeVar()
    val itv = IntTypeVar(width)
    val outt = SignedIntType(width + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(itv, HWFunType(itv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SignedExpr = apply(inputs.head)
  def unapply(expr: SignedExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class UnSignedExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): UnSignedExpr = UnSignedExpr(newInnerIR)
}
object Unsigned extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): UnSignedExpr = UnSignedExpr({
    val width = ArithTypeVar()
    val itv = SignedIntTypeVar(width)
    val outt = IntType(width - 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(itv, HWFunType(itv, outt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): UnSignedExpr = apply(inputs.head)
  def unapply(expr: UnSignedExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class AddIntExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): AddIntExpr = AddIntExpr(newInnerIR)
}
object AddInt extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): AddIntExpr = AddIntExpr({
    val itv1 = IntTypeVar()
    val itv2 = IntTypeVar()
    val ttv = TupleTypeVar(itv1, itv2)
    val outit = IntType(
      SimplifyIfThenElse(
        itv1.bitWidth.ae.gt(itv2.bitWidth.ae),
        itv1.bitWidth.ae,
        SimplifyIfThenElse(
          itv1.bitWidth.ae.eq(itv2.bitWidth.ae),
          itv1.bitWidth.ae + 1,
          itv2.bitWidth.ae
        )
      )
    )
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, HWFunType(ttv, outit))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddIntExpr = apply(inputs.head)
  def unapply(expr: AddIntExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class AddFloatExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): AddFloatExpr = AddFloatExpr(newInnerIR)
}
object AddFloat extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): AddFloatExpr = AddFloatExpr({
    val floatTv = FloatTypeVar()
    val ttv = TupleTypeVar(floatTv, floatTv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, HWFunType(ttv, floatTv))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddFloatExpr = apply(inputs.head)
  def unapply(expr: AddFloatExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object AddInt2 extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): AddIntExpr = AddInt(Tuple(input1, input2))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddIntExpr = apply(inputs.head, inputs(1))
}

object AddFloat2 extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): AddFloatExpr = AddFloat(Tuple(input1, input2))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AddFloatExpr = apply(inputs.head, inputs(1))
}

case class SubIntExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SubIntExpr = SubIntExpr(newInnerIR)
}
object SubInt extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): SubIntExpr = SubIntExpr({
    val itv1 = SignedIntTypeVar()
    val itv2 = SignedIntTypeVar()
    val ttv = TupleTypeVar(itv1, itv2)
    val outit = SignedIntType(
      SimplifyIfThenElse(
        itv1.bitWidth.ae.gt(itv2.bitWidth.ae),
        itv1.bitWidth.ae,
        itv2.bitWidth.ae
      )
    )
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, HWFunType(ttv, outit))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SubIntExpr = apply(inputs.head)
  def unapply(expr: SubIntExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class MulIntExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MulIntExpr = MulIntExpr(newInnerIR)
}
object MulInt extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): MulIntExpr = MulIntExpr({
    val i1tv = IntTypeVar()
    val i2tv = IntTypeVar()
    val ttv = TupleTypeVar(i1tv, i2tv)
    val outit = IntType(i1tv.bitWidth.ae + i2tv.bitWidth.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, HWFunType(ttv, outit))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MulIntExpr = apply(inputs.head)
  def unapply(expr: MulIntExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object SquareInt extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): MulIntExpr = MulInt(VectorToTuple(VectorGenerator(input, 2)))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MulIntExpr = apply(inputs.head)
}

case class MulFloatExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MulFloatExpr = MulFloatExpr(newInnerIR)
}
object MulFloat extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): MulFloatExpr = MulFloatExpr({
    val floatTv = FloatTypeVar()
    val ttv = TupleTypeVar(floatTv, floatTv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(ttv, HWFunType(ttv, floatTv))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MulFloatExpr = apply(inputs.head)
  def unapply(expr: MulFloatExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class MaxIntExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MaxIntExpr = MaxIntExpr(newInnerIR)
}
object MaxInt extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): MaxIntExpr = MaxIntExpr({
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t.children.head match {
      case _: SignedIntTypeT =>
        val itv = SignedIntTypeVar()
        val inttv = TupleTypeVar(itv, itv)
        val outt = SignedIntType(itv.bitWidth)
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outt))), input.t), input)
      case _ =>
        val itv = IntTypeVar()
        val inttv = TupleTypeVar(itv, itv)
        val outt = IntType(itv.bitWidth)
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outt))), input.t), input)
    }
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxIntExpr = apply(inputs.head)
  def unapply(expr: MaxIntExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object MaxInt2 extends BuiltinExprFun {
  override def args: Int = 2
  def apply(input1: Expr, input2: Expr): MaxIntExpr = MaxInt(Tuple(input1, input2))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxIntExpr = apply(inputs.head, inputs(1))
}

object ReLUInt extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): MaxIntExpr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    MaxInt2(inputtc, ConstantValue(0, Some(inputtc.t)))
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MaxIntExpr = apply(inputs.head)
}

case class SelectExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SelectExpr = SelectExpr(newInnerIR, newTypes)
}
object Select extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, selection: ArithTypeT): SelectExpr = SelectExpr({
    // TODO allow more than 2 elements in a tuple
    assert(selection.ae.evalInt < 2)
    val inttv = TupleTypeVar(HWDataTypeVar(), HWDataTypeVar())
    val outt = inttv.children(selection.ae.evalInt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outt))), input.t), input)
  }, Seq(selection))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SelectExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SelectExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class Select3Expr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Select3Expr = Select3Expr(newInnerIR, newTypes)
}
object Select3 extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, selection: ArithTypeT): Select3Expr = Select3Expr({
    // TODO allow more than 2 elements in a tuple
    assert(selection.ae.evalInt < 3)
    val inttv = TupleTypeVar(HWDataTypeVar(), HWDataTypeVar(), HWDataTypeVar())
    val outt = inttv.children(selection.ae.evalInt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outt))), input.t), input)
  }, Seq(selection))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Select3Expr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: Select3Expr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

/*
 ***********************************************************************************************************************
 complex operations (map, reduce, slide, ...)
 ***********************************************************************************************************************
 */

case class MapOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MapOrderedStreamExpr = MapOrderedStreamExpr(newInnerIR)
}
object MapOrderedStream extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(function: Expr, input: Expr): MapOrderedStreamExpr = MapOrderedStreamExpr({
    val indtv = NonArrayTypeVar()
    val outdtv = NonArrayTypeVar()
    val ftv = FunTypeVar(indtv, outdtv)
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(indtv, lentv)
    val outst = OrderedStreamType(outdtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(ftv, instv), HWFunTypes(Seq(ftv, instv), outst))), Seq(function.t, input.t)), Seq(function, input))
  })
  private[arch] def apply(nestingLevels: Int, function: Expr, input: Expr): Expr = {
    if (nestingLevels <= 0) {
      FunctionCall(function, input)
    } else if (nestingLevels == 1) {
      apply(function, input)
    } else {
      val inputtc = TypeChecker.checkIfUnknown(input, input.t)
      val et = inputtc.t match {
        case st: OrderedStreamTypeT => st.et
        case _ => throw MalformedExprException(s"Cannot create $nestingLevels times nested Maps for input of type ${inputtc.t}")
      }
      MapOrderedStream(
        {
          val p = ParamDef(et)
          ArchLambda(p, apply(nestingLevels - 1, function, ParamUse(p)))
        },
        input
      )
    }
  }

  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MapOrderedStreamExpr = apply(inputs.head, inputs(1))
  def unapply(expr: MapOrderedStreamExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class MapSimpleOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MapSimpleOrderedStreamExpr = MapSimpleOrderedStreamExpr(newInnerIR)
}
object MapSimpleOrderedStream extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(function: Expr, input: Expr): MapSimpleOrderedStreamExpr = MapSimpleOrderedStreamExpr({
    val indtv = NonArrayTypeVar()
    val outdtv = NonArrayTypeVar()
    val ftv = FunTypeVar(indtv, outdtv)
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(indtv, lentv)
    val outst = OrderedStreamType(outdtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(ftv, instv), HWFunTypes(Seq(ftv, instv), outst))), Seq(function.t, input.t)), Seq(function, input))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MapSimpleOrderedStreamExpr = apply(inputs.head, inputs(1))
  def unapply(expr: MapSimpleOrderedStreamExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class MapUnorderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MapUnorderedStreamExpr = MapUnorderedStreamExpr(newInnerIR)
}
object MapUnorderedStream extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(function: Expr, input: Expr): MapUnorderedStreamExpr = MapUnorderedStreamExpr({
    val indtv = NonArrayTypeVar()
    val outdtv = NonArrayTypeVar()
    val ftv = FunTypeVar(indtv, outdtv)
    val lentv = ArithTypeVar()
    val instv = UnorderedStreamTypeVar(indtv, lentv)
    val outst = UnorderedStreamType(outdtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(ftv, instv), HWFunTypes(Seq(ftv, instv), outst))), Seq(function.t, input.t)), Seq(function, input))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MapUnorderedStreamExpr = apply(inputs.head, inputs(1))
  def unapply(expr: MapUnorderedStreamExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class MapVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): MapVectorExpr = MapVectorExpr(newInnerIR)
}
object MapVector extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(function: Expr, input: Expr): MapVectorExpr = MapVectorExpr({
    val indtv = BasicDataTypeVar()
    val outdtv = BasicDataTypeVar()
    val ftv = FunTypeVar(indtv, outdtv)
    val lentv = ArithTypeVar()
    val invtv = VectorTypeVar(indtv, lentv)
    val outvt = VectorType(outdtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(ftv, invtv), HWFunTypes(Seq(ftv, invtv), outvt))), Seq(function.t, input.t)), Seq(function, input))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): MapVectorExpr = apply(inputs.head, inputs(1))
  def unapply(expr: MapVectorExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class FoldOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): FoldOrderedStreamExpr = FoldOrderedStreamExpr(newInnerIR)
}
object FoldOrderedStream extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(function: Expr, input: Expr): FoldOrderedStreamExpr = FoldOrderedStreamExpr({
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case OrderedStreamType(_: IntTypeT, _) =>
        val dtv = ScalarTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val instv = OrderedStreamTypeVar(dtv, lentv)
        var outBitWidth: ArithTypeT = dtv.bitWidth.ae + ceil(arithmetic.Log(2, lentv))
        function.visit{
          case MaxInt(_, _) => outBitWidth = dtv.bitWidth.ae
          case _ =>
        }
        val outt = IntType(outBitWidth)
        //    ConditionalTypeVar(
        //      dtv,
        //      {
        //        case IntType(width) => IntType(width.ae + ceil(arithmetic.Log(2, lentv)))
        //         TODO "... + Log(2 ..." if function is ADD; "... * Log(2 ..." if function is MUL
        //        case ft: FloatTypeT => ft
        //      }
        //    )
        val ftv = FunTypeVar(Seq(outt, dtv), ScalarTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, instv),
            HWFunTypes(
              Seq(ftv, instv),
              outt
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
      case OrderedStreamType(OrderedStreamType(_: IntTypeT, _), _) =>
        val innerLenTv = ArithTypeVar(StartFromRange(1))
        val itv = ScalarTypeVar()
        val dtv = OrderedStreamTypeVar(itv, innerLenTv)
        val outerLenTv = ArithTypeVar(StartFromRange(1))
        val instv = OrderedStreamTypeVar(dtv, outerLenTv)
        var outBitWidth: ArithTypeT = itv.bitWidth.ae + ceil(arithmetic.Log(2, outerLenTv))
        function.visit{
          case MaxInt(_, _) => outBitWidth = itv.bitWidth.ae
          case _ =>
        }
        val outt = OrderedStreamType(IntType(outBitWidth), innerLenTv)
        val ftv = FunTypeVar(Seq(outt, dtv), OrderedStreamTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, instv),
            HWFunTypes(
              Seq(ftv, instv),
              outt
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
      case OrderedStreamType(_: BasicDataTypeT, _) =>
        val dtv = BasicDataTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val instv = OrderedStreamTypeVar(dtv, lentv)
        val ftv = FunTypeVar(Seq(dtv, dtv), dtv)
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, instv),
            HWFunTypes(
              Seq(ftv, instv),
              dtv
            )
          )),
          Seq(function.t, input.t)),
          Seq(function, input)
        )
    }
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): FoldOrderedStreamExpr = apply(inputs.head, inputs(1))
  def unapply(expr: FoldOrderedStreamExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

object FoldVector{
  private[arch] def apply(function: TypeFunctionT, input: Expr): Expr = {
    val inputtc = TypeChecker.checkAssert(input)
    inputtc.t match {
      case VectorType(_, len) if len.ae.evalInt > 2 && len.ae.evalInt % 2 == 0 =>
        val split = TypeChecker.check(SplitVector(inputtc, 2))
        FoldVector(
          function,
          MapVector(
            {
              val param = ParamDef(VectorTypeVar())
              ArchLambda(
                param,
                //                function.call(VectorToTuple(ParamUse(param)))
                {
                  var func: Expr = AddInt(VectorToTuple(ParamUse(param)))
                  function.visit{
                    case MaxInt(_, _) =>
                      func = MaxInt(VectorToTuple(ParamUse(param)))
                    case _ =>
                  }
                  func
                }
              )
            },
            split
          )
        )
      case VectorType(et, len) if len.ae.evalInt > 2 && len.ae.evalInt % 2 == 1 =>
        FoldVector(
          function,
          ConcatVector(Tuple(inputtc, VectorGenerator(ConstantValue(0, Some(et)), 1)))
        )
      case VectorType(et: IntTypeT, len) if len.ae.evalInt > 2 && (math.log(len.ae.evalInt)/math.log(2)) % 1 != 0 =>
        val pow2Order = math.floor(math.log(len.ae.evalInt)/math.log(2))
        val pow2Len = math.pow(2, pow2Order).toInt
        val splitRest = len.ae.evalInt - pow2Len
        val splitInput = SplitVector(inputtc, len)
        val splitInputtc = TypeChecker.checkAssert(splitInput)
        val foldSum = MapVector(
          {
            val p = ParamDef(splitInputtc.t.asInstanceOf[VectorTypeT].et)
            ArchLambda(p,
              {
                val inPow2 = DropVector(ParamUse(p), 0, splitRest)
                val inRest = DropVector(ParamUse(p), pow2Len, 0)
                val sumPow2 = FoldVector(function, inPow2)
                val sumRest = FoldVector(function, inRest)
                var sum: Expr = AddInt(Tuple(sumPow2, sumRest))
                function.visit{
                  case MaxInt(_, _) =>
                    sum = MaxInt(Tuple(sumPow2, sumRest))
                  case _ =>
                }
                val sumtc = TypeChecker.checkAssert(sum)
                TruncInteger(sumtc, ArithType(et.bitWidth.ae.evalInt + pow2Order.toInt + 1))
              }
            )
          }, splitInputtc
        )
        FoldVector(function, foldSum)
      case VectorType(_, len) if len.ae.evalInt == 2 =>
        //        function.call(VectorToTuple(inputtc))
        var func: Expr = AddInt(VectorToTuple(inputtc))
        function.visit{
          case MaxInt(_, _) =>
            func = MaxInt(VectorToTuple(inputtc))
          case _ =>
        }
        func
      case VectorType(et: IntTypeT, len) if len.ae.evalInt == 1 =>
        ArchConversion(inputtc, VectorType(et, len), IntType(et.bitWidth))
      case _ => throw MalformedExprException("input to ReduceVector must be a vector.")
    }
  }
}

case class ReduceOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ReduceOrderedStreamExpr = ReduceOrderedStreamExpr(newInnerIR)
}
object ReduceOrderedStream extends BuiltinExprFun {
  override def args: Int = 3
  private[arch] def apply(function: Expr, initialValue: Expr, input: Expr): ReduceOrderedStreamExpr = ReduceOrderedStreamExpr({
    val initialValuetc = TypeChecker.checkIfUnknown(initialValue, initialValue.t)
    initialValuetc.t match {
      case IntType(_) => // do not allow subtypes of IntTypeT here!
        val dtv = IntTypeVar()
        val initialValueTv = IntTypeVar()
        val lentv = ArithTypeVar(StartFromRange(1))
        val instv = OrderedStreamTypeVar(dtv, lentv)
        var outBitWidth: ArithTypeT = dtv.bitWidth.ae + ceil(arithmetic.Log(2, lentv))
        function.visit{
          case MaxInt(_, _) => outBitWidth = dtv.bitWidth.ae
          case _ =>
        }
        val outt = IntType(outBitWidth)
        // TODO "... + Log(2 ..." if function is ADD; "... * Log(2 ..." if function is MUL
        val ftv = FunTypeVar(Seq(outt, dtv), IntTypeVar())
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, initialValueTv, instv),
            HWFunTypes(
              Seq(ftv, initialValueTv, instv),
              outt
            )
          )),
          Seq(function.t, initialValue.t, input.t)),
          Seq(function, initialValue, input)
        )
      case _ =>
        val dtv = NonArrayTypeVar()
        val acctv = HWDataTypeVar()
        val ftv = FunTypeVar(Seq(acctv, dtv), acctv)
        val instv = OrderedStreamTypeVar(dtv)
        FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
          TypeFunType(
            Seq(ftv, acctv, instv),
            HWFunTypes(
              Seq(ftv, acctv, instv),
              acctv
            )
          )),
          Seq(function.t, initialValue.t, input.t)),
          Seq(function, initialValue, input)
        )
    }
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ReduceOrderedStreamExpr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: ReduceOrderedStreamExpr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

case class SlideOrderedStreamExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SlideOrderedStreamExpr = SlideOrderedStreamExpr(newInnerIR, newTypes)
}
object SlideOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, windowWidth: ArithTypeT): SlideOrderedStreamExpr = SlideOrderedStreamExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(VectorType(dtv, windowWidth), lentv - windowWidth.ae + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  }, Seq(windowWidth))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SlideOrderedStreamExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SlideOrderedStreamExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class AlternateExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): AlternateExpr = AlternateExpr(newInnerIR)
}
object Alternate extends BuiltinExprFun {
  override def args: Int = 3
  private[arch] def apply(f1: Expr, f2: Expr, input: Expr): AlternateExpr = AlternateExpr({
    val intv = HWDataTypeVar()
    val outt = HWDataTypeVar()
    val f1tv = FunTypeVar(intv, outt)
    val f2tv = FunTypeVar(intv, outt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(f1tv, f2tv, intv), HWFunTypes(Seq(f1tv, f2tv, intv), outt))), Seq(f1.t, f2.t, input.t)), Seq(f1, f2, input))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): AlternateExpr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: AlternateExpr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

case class IfelseExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): IfelseExpr = IfelseExpr(newInnerIR)
}
object Ifelse extends BuiltinExprFun {
  override def args: Int = 3
  private[arch] def apply(function1: Expr, function2: Expr, input: Expr): IfelseExpr = IfelseExpr({
    val seltv = IntTypeVar()
    val intv = HWDataTypeVar()
    val intupletv = TupleTypeVar(intv, seltv)
    val outtv = HWDataTypeVar()
    val f1tv = FunTypeVar(intv, outtv)
    val f2tv = FunTypeVar(intv, outtv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(f1tv, f2tv, intupletv), HWFunTypes(Seq(f1tv, f2tv, intupletv), outtv))), Seq(function1.t, function2.t, input.t)), Seq(function1, function2, input))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): IfelseExpr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: IfelseExpr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

/*
 ***********************************************************************************************************************
 data type conversions / rearrangement
 ***********************************************************************************************************************
 */

object ArchConversion {

  final case class ConversionException(private val message: String, private val cause: Throwable = None.orNull) extends Exception(message, cause)

  /**
    * calculates the required length of a collection A with elements of type 'elementType', when converted to / from a collection B of type 'collectionType'
    * this calculation includes possible unused bits, which are chopped of in conversion, e.g. when the bit widths of the elements from A and B are NOT a multiple of each other
    */
  def calcLength(collectionType: UnorderedStreamTypeT, elementType: HWDataTypeT): ArithTypeT = calcLength(1, collectionType: UnorderedStreamTypeT, elementType: HWDataTypeT)

  @scala.annotation.tailrec
  private def calcLength(outerLength: ArithTypeT, collectionType: UnorderedStreamTypeT, elementType: HWDataTypeT): ArithTypeT = collectionType.et match {
    // recursively go through collections and grab their length as 'outerLength', but only if the element Type is not a LogicType (We do not want to end up having streams of bits)
    case ct: UnorderedStreamTypeT if !ct.et.isInstanceOf[LogicTypeT] => calcLength(outerLength.ae * collectionType.len.ae, ct, elementType)
    case _ => SimplifyIfThenElse(
      collectionType.et.bitWidth.ae.gt(elementType.bitWidth.ae),
      outerLength.ae * collectionType.len.ae * (1 + ((collectionType.et.bitWidth.ae - 1) / elementType.bitWidth.ae)),
      1 + (outerLength.ae * collectionType.len.ae - 1) / // ceiling integer division
        SimplifyIfThenElse( // add this condition to prevent a division by zero error, because even if the condition of the first if is true, the else block is evaluated!
          collectionType.et.bitWidth.ae.gt(elementType.bitWidth.ae),
          1,
          elementType.bitWidth.ae / collectionType.et.bitWidth.ae
        )
    )
  }

  /**
    * generator for combinations of expressions to archive the desired data type conversion
    */
  def apply(input: Expr, from: Type, to: Type): Expr = (from, to) match {
    // no conversion needed if 'from' and 'to' are the same
    case (_, _) if from == to => input

    // if length of collections are equal => 'unwrap' and deal with inner elements
    case (OrderedStreamType(et1, len1), OrderedStreamType(et2, len2)) if len1 == len2 =>
      val p = ParamDef(et1)
      val f = ArchLambda(p, apply(ParamUse(p), et1, et2))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (UnorderedStreamType(et1, len1), UnorderedStreamType(et2, len2)) if len1 == len2 =>
      val p = ParamDef(et1)
      val f = ArchLambda(p, apply(ParamUse(p), et1, et2))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)
    case (VectorType(et1, len1), VectorType(et2, len2)) if len1 == len2 =>
      val p = ParamDef(et1)
      val f = ArchLambda(p, apply(ParamUse(p), et1, et2))
      val map = TypeChecker.check(MapVector(f, input))
      apply(map, map.t, to)

    // convert all scalar types (in streams / vectors) to vectors of logic values
    case (st: ScalarTypeT, _) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val conv = Conversion(input, newVecType)
      apply(conv, conv.t, to)
    case (_, st: ScalarTypeT) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val newInput = apply(input, from, newVecType)
      Conversion(newInput, to)
    case (_, UnorderedStreamType(st: ScalarTypeT, len)) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val newInput = apply(input, from, UnorderedStreamType(newVecType, len))
      val p = ParamDef(newVecType)
      val f = ArchLambda(p, Conversion(ParamUse(p), st))
      MapUnorderedStream(f, newInput)
    case (UnorderedStreamType(st: ScalarTypeT, _), _) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val p = ParamDef(st)
      val f = ArchLambda(p, Conversion(ParamUse(p), newVecType))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)
    case (_, OrderedStreamType(st: ScalarTypeT, len)) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val newInput = apply(input, from, OrderedStreamType(newVecType, len))
      val p = ParamDef(newVecType)
      val f = ArchLambda(p, Conversion(ParamUse(p), st))
      MapOrderedStream(f, newInput)
    case (OrderedStreamType(st: ScalarTypeT, _), _) =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val p = ParamDef(st)
      val f = ArchLambda(p, Conversion(ParamUse(p), newVecType))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (_, VectorType(st: ScalarTypeT, len)) if !st.isInstanceOf[LogicTypeT] =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val newInput = apply(input, from, VectorType(newVecType, len))
      val p = ParamDef(newVecType)
      val f = ArchLambda(p, Conversion(ParamUse(p), st))
      MapVector(f, newInput)
    case (VectorType(st: ScalarTypeT, _), _) if !st.isInstanceOf[LogicTypeT] =>
      val newVecType = VectorType(LogicType(), st.bitWidth)
      val p = ParamDef(st)
      val f = ArchLambda(p, Conversion(ParamUse(p), newVecType))
      val map = TypeChecker.check(MapVector(f, input))
      apply(map, map.t, to)

    // convert all tuple types (in streams / vectors) to vectors of logic values
    case (TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), _) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val conv = Conversion(input, newTupleType)
      apply(conv, conv.t, to)
    case (TupleType(Seq(_: VectorTypeT, _: VectorTypeT)), _) =>
      val concat = TypeChecker.check(ConcatVector(input))
      apply(concat, concat.t, to)
    case (_, TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT))) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val newInput = apply(input, from, newTupleType)
      Conversion(newInput, to)
    case (_, TupleType(Seq(VectorType(et1, len1), VectorType(et2, len2)))) if et1 != et2 || len1 != len2 =>
      throw ConversionException("missing primitive to split vector into 2 chunks of different length.")
    case (_, TupleType(Seq(VectorType(et1, len1), VectorType(_, _)))) => // vectors have the same element type and length
      val newInput = apply(input, from, VectorType(VectorType(et1, len1), 2))
      VectorToTuple(newInput)
    case (_, UnorderedStreamType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), len)) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val newInput = apply(input, from, UnorderedStreamType(newTupleType, len))
      val p = ParamDef(newTupleType)
      val f = ArchLambda(p, Conversion(ParamUse(p), tt))
      MapUnorderedStream(f, newInput)
    case (_, UnorderedStreamType(TupleType(Seq(VectorType(et1, len1), VectorType(et2, len2))), _)) if et1 != et2 || len1 != len2 =>
      throw ConversionException("missing primitive to split vector into 2 chunks of different length.")
    case (_, UnorderedStreamType(TupleType(Seq(VectorType(et1, len1), VectorType(_, _))), len)) =>
      val newInput = apply(input, from, UnorderedStreamType(VectorType(VectorType(et1, len1), 2), len))
      MapUnorderedStream(VectorToTuple.asFunction(), newInput)
    case (UnorderedStreamType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), _), _) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val p = ParamDef(tt)
      val f = ArchLambda(p, Conversion(ParamUse(p), newTupleType))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)
    case (UnorderedStreamType(TupleType(Seq(VectorType(_, _), VectorType(_, _))), _), _) =>
      val map = TypeChecker.check(MapUnorderedStream(ConcatVector.asFunction(), input))
      apply(map, map.t, to)
    case (_, OrderedStreamType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), len)) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val newInput = apply(input, from, OrderedStreamType(newTupleType, len))
      val p = ParamDef(newTupleType)
      val f = ArchLambda(p, Conversion(ParamUse(p), tt))
      MapOrderedStream(f, newInput)
    case (_, OrderedStreamType(TupleType(Seq(VectorType(et1, len1), VectorType(et2, len2))), _)) if et1 != et2 || len1 != len2 =>
      throw ConversionException("missing primitive to split vector into 2 chunks of different length.")
    case (_, OrderedStreamType(TupleType(Seq(VectorType(et1, len1), VectorType(_, _))), len)) =>
      val newInput = apply(input, from, OrderedStreamType(VectorType(VectorType(et1, len1), 2), len))
      MapOrderedStream(VectorToTuple.asFunction(), newInput)
    case (OrderedStreamType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), _), _) =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val p = ParamDef(tt)
      val f = ArchLambda(p, Conversion(ParamUse(p), newTupleType))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (OrderedStreamType(TupleType(Seq(VectorType(_, _), VectorType(_, _))), _), _) =>
      val map = TypeChecker.check(MapOrderedStream(ConcatVector.asFunction(), input))
      apply(map, map.t, to)
    case (_, VectorType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), len)) if !st1.isInstanceOf[LogicTypeT] && !st2.isInstanceOf[LogicTypeT] =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val newInput = apply(input, from, VectorType(newTupleType, len))
      val p = ParamDef(newTupleType)
      val f = ArchLambda(p, Conversion(ParamUse(p), tt))
      MapVector(f, newInput)
    case (_, VectorType(TupleType(Seq(VectorType(et1, len1), VectorType(et2, len2))), _)) if et1 != et2 || len1 != len2 =>
      throw ConversionException("missing primitive to split vector into 2 chunks of different length.")
    case (_, VectorType(TupleType(Seq(VectorType(et1, len1), VectorType(_, _))), len)) =>
      val newInput = apply(input, from, VectorType(VectorType(VectorType(et1, len1), 2), len))
      MapVector(VectorToTuple.asFunction(), newInput)
    case (VectorType(tt @ TupleType(Seq(st1: ScalarTypeT, st2: ScalarTypeT)), _), _) if !st1.isInstanceOf[LogicTypeT] && !st2.isInstanceOf[LogicTypeT] =>
      val newTupleType = TupleType(
        VectorType(LogicType(), st1.bitWidth),
        VectorType(LogicType(), st2.bitWidth)
      )
      val p = ParamDef(tt)
      val f = ArchLambda(p, Conversion(ParamUse(p), newTupleType))
      val map = TypeChecker.check(MapVector(f, input))
      apply(map, map.t, to)
    case (VectorType(TupleType(Seq(VectorType(_, _), VectorType(_, _))), _), _) =>
      val map = TypeChecker.check(MapVector(ConcatVector.asFunction(), input))
      apply(map, map.t, to)

    // drop or concat
    case (OrderedStreamType(et1, len1), OrderedStreamType(et2, len2)) if et1 == et2 && len1.ae.evalInt > len2.ae.evalInt =>
      val drop = TypeChecker.check(DropOrderedStream(input, 0, len1.ae - len2.ae))
      apply(drop, drop.t, to)
    case (OrderedStreamType(et1, len1), OrderedStreamType(et2, len2)) if et1 == et2 && len1.ae.evalInt < len2.ae.evalInt =>
      val constantZero = ConstantValue(0, Some(et1))
      val streamZero = Repeat(constantZero, len2.ae - len1.ae)
      val concat = TypeChecker.check(ConcatOrderedStream(input, streamZero))
      apply(concat, concat.t, to)
    case (VectorType(et1, len1), VectorType(et2, len2)) if et1 == et2 && len1.ae.evalInt > len2.ae.evalInt =>
      val drop = TypeChecker.check(DropVector(input, 0, len1.ae - len2.ae))
      apply(drop, drop.t, to)
    case (VectorType(et1, len1), VectorType(et2, len2)) if et1 == et2 && len1.ae.evalInt < len2.ae.evalInt =>
      val constantZero = ConstantValue(0, Some(et1))
      val vectorZero = VectorGenerator(constantZero, len2.ae - len1.ae)
      val concat = TypeChecker.check(ConcatVector(Tuple(input, vectorZero)))
      apply(concat, concat.t, to)
    case (VectorType(et1, len1), OrderedStreamType(et2, len2)) if et1 == et2 && len1.ae.evalInt > len2.ae.evalInt =>
      val drop = TypeChecker.check(DropVector(input, 0, len1.ae - len2.ae))
      apply(drop, drop.t, to)
    case (VectorType(et1, len1), UnorderedStreamType(et2, len2)) if et1 == et2 && len1.ae.evalInt > len2.ae.evalInt =>
      val drop = TypeChecker.check(DropVector(input, 0, len1.ae - len2.ae))
      apply(drop, drop.t, to)

    // collection conversion
    case (OrderedStreamType(et1, len1), UnorderedStreamType(et2, len2)) =>
      if (et1 == et2 && len1 == len2) {
        // strict condition here: we do not want to "loose" the ordering of a stream if we are not certain if that is the correct conversion to do
        val indexing = TypeChecker.check(OrderedStreamToUnorderedStream(input))
        apply(indexing, indexing.t, to)
      } else {
        // invoke indexing later
        val newInput = apply(input, from, OrderedStreamType(et2, len2))
        apply(newInput, newInput.t, to)
      }
    case (OrderedStreamType(OrderedStreamType(_, _), _), VectorType(_, _)) =>
      val join = TypeChecker.check(JoinOrderedStream(input))
      apply(join, join.t, to)
    case (OrderedStreamType(uos1 @ UnorderedStreamType(_, _), _), VectorType(_, _)) =>
      val p = ParamDef(uos1)
      val f = ArchLambda(p, UnorderedStreamToOrderedStream(ParamUse(p)))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (OrderedStreamType(_, _), VectorType(_, _)) =>
      // TODO this may lead to performance issues, if the 'from' stream is long and/or has high precision elements
      val vectorise = TypeChecker.check(OrderedStreamToVector(input))
      apply(vectorise, vectorise.t, to)
    case (VectorType(et1, len1), OrderedStreamType(et2, len2)) if et1 == et2 && len1 == len2 =>
      val toStream = TypeChecker.check(VectorToOrderedStream(input))
      apply(toStream, toStream.t, to)
    case (VectorType(et1, len1), UnorderedStreamType(et2, len2)) if et1 == et2 && len1 == len2 =>
      val toStream = TypeChecker.check(VectorToOrderedStream(input))
      apply(toStream, toStream.t, to)
    case (UnorderedStreamType(_, _), OrderedStreamType(_, _)) | (UnorderedStreamType(_, _), VectorType(_, _)) =>
      val sorting = TypeChecker.check(UnorderedStreamToOrderedStream(input))
      apply(sorting, sorting.t, to)
    // convert nested vector to stream
    case (OrderedStreamType(vt1 @ VectorType(et1, len1inner), len1outer), OrderedStreamType(et2, len2)) if (et1 == et2 && len1outer.ae.evalInt > 1) || len1inner.ae.evalInt * len1outer.ae.evalInt == len2.ae.evalInt =>
      // if there is only one vector we directly split the vector to get better performance
      val p = ParamDef(vt1)
      val f = ArchLambda(p, VectorToOrderedStream(ParamUse(p)))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (UnorderedStreamType(vt1 @ VectorType(et1, len1inner), len1outer), UnorderedStreamType(et2, len2)) if (et1 == et2 && len1outer.ae.evalInt > 1) || len1inner.ae.evalInt * len1outer.ae.evalInt == len2.ae.evalInt =>
      // if there is only one vector we directly split the vector to get better performance
      val p = ParamDef(vt1)
      val f = ArchLambda(p, VectorToOrderedStream(ParamUse(p)))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)

    // join nested
    case (VectorType(VectorType(et1, len1inner), len1outer), VectorType(et2, len2)) if et1 == et2 || len1inner.ae.evalInt * len1outer.ae.evalInt == len2.ae.evalInt =>
      val join = TypeChecker.check(JoinVector(input))
      apply(join, join.t, to)
    case (OrderedStreamType(OrderedStreamType(et1, len1inner), len1outer), OrderedStreamType(et2, len2)) if et1 == et2 || len1inner.ae.evalInt * len1outer.ae.evalInt == len2.ae.evalInt =>
      val join = TypeChecker.check(JoinOrderedStream(input))
      apply(join, join.t, to)
    case (OrderedStreamType(uos1 @ UnorderedStreamType(et1, _), _), OrderedStreamType(et2, _)) if et1 == et2 =>
      val p = ParamDef(uos1)
      val f = ArchLambda(p, UnorderedStreamToOrderedStream(ParamUse(p)))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (UnorderedStreamType(UnorderedStreamType(et1, len1inner), len1outer), UnorderedStreamType(et2, len2)) if et1 == et2 || len1inner.ae.evalInt * len1outer.ae.evalInt == len2.ae.evalInt =>
      val join = TypeChecker.check(JoinUnorderedStream(input))
      apply(join, join.t, to)
    case (UnorderedStreamType(os1 @ OrderedStreamType(et1, _), _), UnorderedStreamType(et2, _)) if et1 == et2 =>
      val p = ParamDef(os1)
      val f = ArchLambda(p, OrderedStreamToUnorderedStream(ParamUse(p)))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)

    // split vector
    case (OrderedStreamType(vt1 @ VectorType(et1, len1inner), len1outer), OrderedStreamType(VectorType(et2, len2inner), len2outer)) if et1 == et2 && len1inner.ae.evalInt > len2inner.ae.evalInt && (len1inner.ae.evalInt * len1outer.ae.evalInt >= len2inner.ae.evalInt * len2outer.ae.evalInt) =>
      val p = ParamDef(vt1)
      val f = ArchLambda(p, SplitVector(ParamUse(p), len2inner))
      val map = TypeChecker.check(MapOrderedStream(f, input))
      apply(map, map.t, to)
    case (UnorderedStreamType(vt1 @ VectorType(et1, len1inner), len1outer), UnorderedStreamType(VectorType(et2, len2inner), len2outer)) if et1 == et2 && len1inner.ae.evalInt > len2inner.ae.evalInt && (len1inner.ae.evalInt * len1outer.ae.evalInt >= len2inner.ae.evalInt * len2outer.ae.evalInt) =>
      val p = ParamDef(vt1)
      val f = ArchLambda(p, SplitVector(ParamUse(p), len2inner))
      val map = TypeChecker.check(MapUnorderedStream(f, input))
      apply(map, map.t, to)
    case (OrderedStreamType(VectorType(et1, width1), len1), OrderedStreamType(VectorType(et2, width2), _)) if et1 == et2 && width1.ae.evalInt < width2.ae.evalInt && len1.ae.evalInt % (width2.ae.evalInt / width1.ae.evalInt) > 0 =>
      // add elements to stream, if split is not possible
      val chunkSize = width2.ae.evalInt / width1.ae.evalInt
      val missingElements = chunkSize - (len1.ae.evalInt % chunkSize)
      val constantZero = ConstantValue(0, Some(et1))
      val vectorZero = VectorGenerator(constantZero, width1)
      val streamZero = Repeat(vectorZero, missingElements)
      val concat = TypeChecker.check(ConcatOrderedStream(input, streamZero))
      apply(concat, concat.t, to)
    case (UnorderedStreamType(VectorType(et1, width1), _), UnorderedStreamType(vt2 @ VectorType(et2, width2), len2)) if et1 == et2 && width1.ae.evalInt < width2.ae.evalInt =>
      val newInput = apply(input, from, OrderedStreamType(vt2, len2))
      apply(newInput, newInput.t, to)
    case (VectorType(et1, width1), OrderedStreamType(VectorType(et2, width2), len2)) if et1 == et2 && width1.ae.evalInt < width2.ae.evalInt * len2.ae.evalInt =>
      val newInput = apply(input, from, VectorType(et1, width2.ae * len2.ae))
      apply(newInput, newInput.t, to)
    case (VectorType(et1, width1), OrderedStreamType(VectorType(et2, width2), len2)) if et1 == et2 && width1.ae.evalInt == width2.ae.evalInt * len2.ae.evalInt =>
      val split = TypeChecker.check(SplitVector(input, width2))
      apply(split, split.t, to)
    case (VectorType(et1, width1), UnorderedStreamType(VectorType(et2, width2), len2)) if et1 == et2 && width1.ae.evalInt < width2.ae.evalInt * len2.ae.evalInt =>
      val newInput = apply(input, from, VectorType(et1, width2.ae * len2.ae))
      apply(newInput, newInput.t, to)
    case (VectorType(et1, width1), UnorderedStreamType(VectorType(et2, width2), len2)) if et1 == et2 && width1.ae.evalInt == width2.ae.evalInt * len2.ae.evalInt =>
      val split = TypeChecker.check(SplitVector(input, width2))
      apply(split, split.t, to)

    // split stream
    case (OrderedStreamType(et1, len1), OrderedStreamType(OrderedStreamType(et2, len2inner), len2outer)) if et1 != et2 && len1.ae.evalInt == len2inner.ae.evalInt * len2outer.ae.evalInt && len1.ae.evalInt % len2inner.ae.evalInt == 0 =>
      val split = TypeChecker.check(SplitOrderedStream(input, len2inner))
      apply(split, split.t, to)
    case (OrderedStreamType(et1, len1), OrderedStreamType(et2, len2)) if et1 != et2 && len1.ae.evalInt % len2.ae.evalInt == 0 =>
      val split = TypeChecker.check(SplitOrderedStream(input, len1.ae / len2.ae))
      apply(split, split.t, to)
    case (OrderedStreamType(et1, len1), OrderedStreamType(et2, len2)) if et1 != et2 && len2.ae.evalInt % len1.ae.evalInt == 0 =>
      // invoke split later
      val newInput = apply(input, from, OrderedStreamType(OrderedStreamType(et2, len2.ae / len1.ae), len1.ae))
      apply(newInput, newInput.t, to)
    case (UnorderedStreamType(et1, len1), UnorderedStreamType(et2, len2)) if et1 != et2 && len2.ae.evalInt % len1.ae.evalInt == 0 =>
      // invoke split later
      val newInput = apply(input, from, UnorderedStreamType(UnorderedStreamType(et2, len2.ae / len1.ae), len1.ae))
      apply(newInput, newInput.t, to)
    case (OrderedStreamType(et1, len1), OrderedStreamType(OrderedStreamType(et2, len2inner), len2outer)) if et1 != et2 && len1 != len2outer =>
      // invoke split later
      val newInput = apply(input, from, OrderedStreamType(et2, len2inner.ae * len2outer.ae))
      apply(newInput, newInput.t, to)
    case (UnorderedStreamType(et1, len1), UnorderedStreamType(UnorderedStreamType(et2, len2inner), len2outer)) if et1 != et2 && len1 != len2outer =>
      // invoke split later
      val newInput = apply(input, from, UnorderedStreamType(et2, len2inner.ae * len2outer.ae))
      apply(newInput, newInput.t, to)

    case _ =>
      throw ConversionException("Cannot convert type " + from + " to " + to)
  }
}

case class VectorToTupleExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): VectorToTupleExpr = VectorToTupleExpr(newInnerIR)
}
object VectorToTuple extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): VectorToTupleExpr = VectorToTupleExpr({
    val dtv = BasicDataTypeVar()
    // TODO make more generic for vector len > 2
    val len = 2
    val invtv = VectorTypeVar(dtv, len)
    val outst = TupleType(dtv, dtv) // Seq.fill(len)(dtv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): VectorToTupleExpr = apply(inputs.head)
  def unapply(expr: VectorToTupleExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class VectorToOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): VectorToOrderedStreamExpr = VectorToOrderedStreamExpr(newInnerIR)
}
object VectorToOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): VectorToOrderedStreamExpr = VectorToOrderedStreamExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val invtv = VectorTypeVar(dtv, lentv)
    val outst = OrderedStreamType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): VectorToOrderedStreamExpr = apply(inputs.head)
  def unapply(expr: VectorToOrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class OrderedStreamToVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): OrderedStreamToVectorExpr = OrderedStreamToVectorExpr(newInnerIR)
}
object OrderedStreamToVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): OrderedStreamToVectorExpr = OrderedStreamToVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outvt = VectorType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outvt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): OrderedStreamToVectorExpr = apply(inputs.head)
  def unapply(expr: OrderedStreamToVectorExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class IndexUnorderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): IndexUnorderedStreamExpr = IndexUnorderedStreamExpr(newInnerIR)
}
object IndexUnorderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): IndexUnorderedStreamExpr = IndexUnorderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar(StartFromRange(1))
    val instv = UnorderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(TupleType(dtv, instv.indexType), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(
      TypeFunType(
        instv,
        HWFunType(
          instv,
          outst
        )
      )),
      input.t),
      input
    )
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): IndexUnorderedStreamExpr = apply(inputs.head)
  def unapply(expr: IndexUnorderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}
object IndexStream {
  private[arch] def apply(input: Expr): Expr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case OrderedStreamType(_: BasicDataTypeT, _) => apply(OrderedStreamToUnorderedStream(inputtc))
      case UnorderedStreamType(_: BasicDataTypeT, _) => IndexUnorderedStream(inputtc)
      case st: UnorderedStreamTypeT =>
        SplitOrderedStream(apply(JoinStream(inputtc)), st.len)
      case _ =>
        throw MalformedExprException("input to IndexStream must be a stream, got instead: " + inputtc.t)
    }
  }
}

case class UnorderedStreamToOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): UnorderedStreamToOrderedStreamExpr = UnorderedStreamToOrderedStreamExpr(newInnerIR)
}
object UnorderedStreamToOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): UnorderedStreamToOrderedStreamExpr = UnorderedStreamToOrderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv = UnorderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): UnorderedStreamToOrderedStreamExpr = apply(inputs.head)
  def unapply(expr: UnorderedStreamToOrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}
//object UnorderedStreamToOrderedStream {
//  private[backend.hdl.arch] def apply(input: Expr): Expr = BufferUnorderedStream(input)
//}

case class OrderedStreamToUnorderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): OrderedStreamToUnorderedStreamExpr = OrderedStreamToUnorderedStreamExpr(newInnerIR)
}
object OrderedStreamToUnorderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): OrderedStreamToUnorderedStreamExpr = OrderedStreamToUnorderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outst = UnorderedStreamType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): OrderedStreamToUnorderedStreamExpr = apply(inputs.head)
  def unapply(expr: OrderedStreamToUnorderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class TupleExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TupleExpr = TupleExpr(newInnerIR)
}
object Tuple extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(input1: Expr, input2: Expr): TupleExpr = TupleExpr({
    val dtv1 = HWDataTypeVar()
    val dtv2 = HWDataTypeVar()
    val outtt = TupleType(dtv1, dtv2)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv1, dtv2), HWFunTypes(Seq(dtv1, dtv2), outtt))), Seq(input1.t, input2.t)), Seq(input1, input2))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): TupleExpr = apply(inputs.head, inputs(1))
  def unapply(expr: TupleExpr): Some[(Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.t))
}

case class FlipTupleExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): FlipTupleExpr = FlipTupleExpr(newInnerIR)
}
object FlipTuple extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): FlipTupleExpr = FlipTupleExpr({
    val dtv1 = HWDataTypeVar()
    val dtv2 = HWDataTypeVar()
    val intv = TupleTypeVar(dtv1, dtv2)
    val outtt = TupleType(dtv2, dtv1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(intv, HWFunType(intv, outtt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): FlipTupleExpr = apply(inputs.head)
  def unapply(expr: FlipTupleExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class ZipOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ZipOrderedStreamExpr = ZipOrderedStreamExpr(newInnerIR)
}
object ZipOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): ZipOrderedStreamExpr = ZipOrderedStreamExpr({
    val dtv1 = NonArrayTypeVar()
    val dtv2 = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv1 = OrderedStreamTypeVar(dtv1, lentv)
    val instv2 = OrderedStreamTypeVar(dtv2, lentv)
    val inttv = TupleTypeVar(instv1, instv2)
    val outst = OrderedStreamType(TupleType(dtv1, dtv2), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ZipOrderedStreamExpr = apply(inputs.head)
  def unapply(expr: ZipOrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class ZipVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ZipVectorExpr = ZipVectorExpr(newInnerIR)
}
object ZipVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): ZipVectorExpr = ZipVectorExpr({
    val dtv1 = BasicDataTypeVar()
    val dtv2 = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val instv1 = VectorTypeVar(dtv1, lentv)
    val instv2 = VectorTypeVar(dtv2, lentv)
    val inttv = TupleTypeVar(instv1, instv2)
    val outst = VectorType(TupleType(dtv1, dtv2), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ZipVectorExpr = apply(inputs.head)
  def unapply(expr: ZipVectorExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class SplitOrderedStreamExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SplitOrderedStreamExpr = SplitOrderedStreamExpr(newInnerIR, newTypes)
}
object SplitOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, chunkSize: ArithTypeT): SplitOrderedStreamExpr = apply(input, chunkSize, innerSize = true)
  private[arch] def apply(input: Expr, size: ArithTypeT, innerSize: Boolean): SplitOrderedStreamExpr = {
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)

    val (chunkSize: ArithExpr, numberOfChunks: ArithExpr) =
      if (innerSize)
        (size.ae, lentv.ae / size.ae)
      else
        (lentv.ae / size.ae, size.ae)

    SplitOrderedStreamExpr({
      val outst = OrderedStreamType(OrderedStreamType(dtv, chunkSize), numberOfChunks)
      FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
    }, Seq(chunkSize))
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SplitOrderedStreamExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SplitOrderedStreamExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class SplitVectorExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SplitVectorExpr = SplitVectorExpr(newInnerIR, newTypes)
}
object SplitVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, chunkSize: ArithTypeT): SplitVectorExpr = SplitVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val invtv = VectorTypeVar(dtv, lentv)
    val outvt = VectorType(VectorType(dtv, chunkSize), lentv.ae / chunkSize.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outvt))), input.t), input)
  }, Seq(chunkSize))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SplitVectorExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: SplitVectorExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class JoinOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): JoinOrderedStreamExpr = JoinOrderedStreamExpr(newInnerIR)
}
object JoinOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  /**
    * @param input an expression with type ordered stream of ordered stream. If you want to join an ordered stream of unordered stream, reorder the inner unordered stream first
    */
  private[arch] def apply(input: Expr): JoinOrderedStreamExpr = JoinOrderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val innerlentv = ArithTypeVar()
    val outerlentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(OrderedStreamTypeVar(dtv, innerlentv), outerlentv)
    val outst = OrderedStreamType(dtv, innerlentv.ae * outerlentv.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): JoinOrderedStreamExpr = apply(inputs.head)
  def unapply(expr: JoinOrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class JoinUnorderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): JoinUnorderedStreamExpr = JoinUnorderedStreamExpr(newInnerIR)
}
object JoinUnorderedStream extends BuiltinExprFun {
  override def args: Int = 1
  /**
    * @param input an expression with type unordered stream of unordered stream. If you want to join an unordered stream of ordered stream, convert the inner ordered stream to an unordered stream first
    */
  private[arch] def apply(input: Expr): JoinUnorderedStreamExpr = JoinUnorderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val innerlentv = ArithTypeVar()
    val outerlentv = ArithTypeVar()
    val instv = UnorderedStreamTypeVar(UnorderedStreamTypeVar(dtv, innerlentv), outerlentv)
    val outst = UnorderedStreamType(dtv, innerlentv.ae * outerlentv.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): JoinUnorderedStreamExpr = apply(inputs.head)
  def unapply(expr: JoinUnorderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

object JoinStream {
  private[arch] def apply(input: Expr): Expr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case OrderedStreamType(et, _) => et match {
        case OrderedStreamType(_, _) => JoinOrderedStream(inputtc)
        case UnorderedStreamType(_, _) => JoinUnorderedStream(OrderedStreamToUnorderedStream(inputtc))
        case _ => throw new RuntimeException("never happens")
      }
      case UnorderedStreamType(et, _) => et match {
        case os @ OrderedStreamType(_, _) =>
          JoinUnorderedStream(MapUnorderedStream(
            {
              val p = ParamDef(os)
              ArchLambda(p, OrderedStreamToUnorderedStream(ParamUse(p)))
            },
            inputtc
          ))
        case UnorderedStreamType(_, _) => JoinUnorderedStream(inputtc)
        case _ => throw new RuntimeException("never happens")
      }
    }
  }
}
object JoinStreamAll {
  private[arch] def apply(input: Expr): Expr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    inputtc.t match {
      case streamType: UnorderedStreamTypeT if streamType.dimensions.size > 1 => apply(JoinStream(inputtc))
      case _ => inputtc
    }
  }
}

case class JoinVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): JoinVectorExpr = JoinVectorExpr(newInnerIR)
}
object JoinVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): JoinVectorExpr = JoinVectorExpr({
    val dtv = BasicDataTypeVar()
    val innerlentv = ArithTypeVar()
    val outerlentv = ArithTypeVar()
    val invtv = VectorTypeVar(VectorTypeVar(dtv, innerlentv), outerlentv)
    val outvt = VectorType(dtv, innerlentv.ae * outerlentv.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outvt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): JoinVectorExpr = apply(inputs.head)
  def unapply(expr: JoinVectorExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class DropOrderedStreamExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): DropOrderedStreamExpr = DropOrderedStreamExpr(newInnerIR, newTypes)
}
object DropOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, firstElements: ArithTypeT, lastElements: ArithTypeT): DropOrderedStreamExpr = DropOrderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(dtv, lentv.ae - firstElements.ae - lastElements.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  }, Seq(firstElements, lastElements))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): DropOrderedStreamExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT], types(1).asInstanceOf[ArithTypeT])
  def unapply(expr: DropOrderedStreamExpr): Some[(Expr, ArithTypeT, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.types(1).asInstanceOf[ArithTypeT], expr.t))
}

case class DropVectorExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): DropVectorExpr = DropVectorExpr(newInnerIR, newTypes)
}
object DropVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, lowElements: ArithTypeT, highElements: ArithTypeT): DropVectorExpr = DropVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val invtv = VectorTypeVar(dtv, lentv)
    val outvt = VectorType(dtv, lentv.ae - lowElements.ae - highElements.ae)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outvt))), input.t), input)
  }, Seq(lowElements, highElements))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): DropVectorExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT], types(1).asInstanceOf[ArithTypeT])
  def unapply(expr: DropVectorExpr): Some[(Expr, ArithTypeT, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.types(1).asInstanceOf[ArithTypeT], expr.t))
}

case class ConcatOrderedStreamExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ConcatOrderedStreamExpr = ConcatOrderedStreamExpr(newInnerIR, newTypes)
}
object ConcatOrderedStream extends BuiltinExprFun {
  override def args: Int = 2
  private[arch] def apply(input1: Expr, input2: Expr, dimension: ArithTypeT = ArithType(0)): ConcatOrderedStreamExpr = ConcatOrderedStreamExpr({
    val (instv1, instv2, outst) = getStreamTypes(dimension.ae.evalInt)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(instv1, instv2), HWFunTypes(Seq(instv1, instv2), outst))), Seq(input1.t, input2.t)), Seq(input1, input2))
  }, Seq(dimension))

  private def getStreamTypes(dimension: Int): (OrderedStreamTypeVar, OrderedStreamTypeVar, OrderedStreamType) = dimension match {
    case 0 =>
      val dtv = NonArrayTypeVar()
      val lentv1 = ArithTypeVar()
      val lentv2 = ArithTypeVar()
      (
        OrderedStreamTypeVar(dtv, lentv1),
        OrderedStreamTypeVar(dtv, lentv2),
        OrderedStreamType(dtv, lentv1 + lentv2)
      )
    case _ =>
      val lentv = ArithTypeVar()
      val streamTypes = getStreamTypes(dimension - 1)
      (
        OrderedStreamTypeVar(streamTypes._1, lentv),
        OrderedStreamTypeVar(streamTypes._2, lentv),
        OrderedStreamType(streamTypes._3, lentv)
      )
  }

  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ConcatOrderedStreamExpr = apply(inputs.head, inputs(1), types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: ConcatOrderedStreamExpr): Some[(Expr, Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.args(1), expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class ConcatVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): ConcatVectorExpr = ConcatVectorExpr(newInnerIR)
}
object ConcatVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): ConcatVectorExpr = ConcatVectorExpr({
    val dtv = BasicDataTypeVar()
    val lowinlentv = ArithTypeVar()
    val highinlentv = ArithTypeVar()
    val lowinvtv = VectorTypeVar(dtv, lowinlentv)
    val highinvtv = VectorTypeVar(dtv, highinlentv)
    val inttv = TupleTypeVar(lowinvtv, highinvtv)
    val outst = VectorType(dtv, lowinlentv + highinlentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): ConcatVectorExpr = apply(inputs.head)
  def unapply(expr: ConcatVectorExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class UpdateOrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): UpdateOrderedStreamExpr = UpdateOrderedStreamExpr(newInnerIR)
}
object UpdateOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): UpdateOrderedStreamExpr = UpdateOrderedStreamExpr({
    val dtv = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): UpdateOrderedStreamExpr = apply(inputs.head)
  def unapply(expr: UpdateOrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class UpdateVectorExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): UpdateVectorExpr = UpdateVectorExpr(newInnerIR)
}
object UpdateVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): UpdateVectorExpr = UpdateVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val invtv = TupleTypeVar(
      VectorTypeVar(dtv, lentv), // the original vector
      TupleTypeVar(IntTypeVar(), BasicDataTypeVar()) // the index to update and its value
    )
    val outvt = VectorType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(invtv, HWFunType(invtv, outvt))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): UpdateVectorExpr = apply(inputs.head)
  def unapply(expr: UpdateVectorExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

case class RotateVectorExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): RotateVectorExpr = RotateVectorExpr(newInnerIR, newTypes)
}
object RotateVector extends BuiltinExprFun {
  override def args: Int = 1
  /**
    * @param input an expression with type tuple of VectorType and IntType. The integer identifies the distance to rotate.
    * @param rotateLeft a flag for left rotation. If 0, the input vector is rotated right (LSB). Otherwise the vector is rotated left (MSB).
    */
  private[arch] def apply(input: Expr, rotateLeft: ArithTypeT): RotateVectorExpr = RotateVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val intv1 = VectorTypeVar(dtv, lentv)
    val intv2 = IntTypeVar()
    val intv = TupleTypeVar(intv1, intv2)
    val outtv = VectorType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(intv, HWFunType(intv, outtv))), input.t), input)
  }, Seq(rotateLeft))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): RotateVectorExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: RotateVectorExpr): Some[(Expr, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t))
}

case class PermuteVectorExpr private[arch] (innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): PermuteVectorExpr = PermuteVectorExpr(newInnerIR, newTypes)
}
object PermuteVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, rule: ArithLambdaType): PermuteVectorExpr = PermuteVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val inseqtv = VectorTypeVar(dtv, lentv)
    val outseqt = VectorType(dtv, lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, HWFunType(inseqtv, outseqt))), input.t), input)
  }, Seq(rule))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): PermuteVectorExpr = apply(inputs.head, types.head.asInstanceOf[ArithLambdaType])
  def unapply(expr: PermuteVectorExpr): Some[(Expr, ArithLambdaType, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithLambdaType], expr.t))
}

case class SlideVectorExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SlideVectorExpr = SlideVectorExpr(newInnerIR, newTypes)
}
object SlideVector extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, windowWidth: ArithTypeT, stepSize: ArithTypeT): SlideVectorExpr = SlideVectorExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val instv = VectorTypeVar(dtv, lentv)
    val outst = VectorType(VectorType(dtv, windowWidth), (lentv  - windowWidth.ae) / stepSize.ae + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  }, Seq(windowWidth, stepSize))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SlideVectorExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT], types(1).asInstanceOf[ArithTypeT])
  def unapply(expr: SlideVectorExpr): Some[(Expr, ArithTypeT, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.types(1).asInstanceOf[ArithTypeT], expr.t))
}

case class SlideGeneralOrderedStreamExpr private[arch](innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): SlideGeneralOrderedStreamExpr = SlideGeneralOrderedStreamExpr(newInnerIR, newTypes)
}
object SlideGeneralOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, windowWidth: ArithTypeT, stepSize: ArithTypeT): SlideGeneralOrderedStreamExpr = SlideGeneralOrderedStreamExpr({
    val dtv = BasicDataTypeVar()
    val lentv = ArithTypeVar()
    val instv = OrderedStreamTypeVar(dtv, lentv)
    val outst = OrderedStreamType(OrderedStreamType(dtv, windowWidth), (lentv  - windowWidth.ae) / stepSize.ae + 1)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(instv, HWFunType(instv, outst))), input.t), input)
  }, Seq(windowWidth, stepSize))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): SlideGeneralOrderedStreamExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT], types(1).asInstanceOf[ArithTypeT])
  def unapply(expr: SlideGeneralOrderedStreamExpr): Some[(Expr, ArithTypeT, ArithTypeT, Type)] =
    Some((expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.types(1).asInstanceOf[ArithTypeT], expr.t))
}

// Added Tuple of 3
case class Tuple3Expr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Tuple3Expr = Tuple3Expr(newInnerIR)
}
object Tuple3 extends BuiltinExprFun {
  override def args: Int = 3
  private[arch] def apply(input1: Expr, input2: Expr, input3: Expr): Tuple3Expr = Tuple3Expr({
    val dtv1 = HWDataTypeVar()
    val dtv2 = HWDataTypeVar()
    val dtv3 = HWDataTypeVar()
    val outtt = TupleType(dtv1, dtv2, dtv3)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(Seq(dtv1, dtv2, dtv3), HWFunTypes(Seq(dtv1, dtv2, dtv3), outtt))), Seq(input1.t, input2.t, input3.t)), Seq(input1, input2, input3))
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Tuple3Expr = apply(inputs.head, inputs(1), inputs(2))
  def unapply(expr: Tuple3Expr): Some[(Expr, Expr, Expr, Type)] =
    Some((expr.args.head, expr.args(1), expr.args(2), expr.t))
}

case class Zip3OrderedStreamExpr private[arch](innerIR: Expr) extends ArchExpr {
  override def types: Seq[Type] = Seq()
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): Zip3OrderedStreamExpr = Zip3OrderedStreamExpr(newInnerIR)
}
object Zip3OrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr): Zip3OrderedStreamExpr = Zip3OrderedStreamExpr({
    val dtv1 = NonArrayTypeVar()
    val dtv2 = NonArrayTypeVar()
    val dtv3 = NonArrayTypeVar()
    val lentv = ArithTypeVar()
    val instv1 = OrderedStreamTypeVar(dtv1, lentv)
    val instv2 = OrderedStreamTypeVar(dtv2, lentv)
    val instv3 = OrderedStreamTypeVar(dtv3, lentv)
    val inttv = TupleTypeVar(instv1, instv2, instv3)
    val outst = OrderedStreamType(TupleType(dtv1, dtv2, dtv3), lentv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inttv, HWFunType(inttv, outst))), input.t), input)
  })
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): Zip3OrderedStreamExpr = apply(inputs.head)
  def unapply(expr: Zip3OrderedStreamExpr): Some[(Expr, Type)] =
    Some((expr.args.head, expr.t))
}

// Dummy primitive
case class TransposeNDOrderedStreamExpr private[arch] (innerIR: Expr, types: Seq[Type]) extends ArchExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): TransposeNDOrderedStreamExpr = TransposeNDOrderedStreamExpr(newInnerIR, newTypes)
}
object TransposeNDOrderedStream extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr, dimOrder: Seq[ArithTypeT]): TransposeNDOrderedStreamExpr = TransposeNDOrderedStreamExpr({
    def genReorderedStmType(dimOrder: Seq[ArithTypeT], lentvSeq: Seq[ArithTypeT], dtv: BasicDataTypeVar): OrderedStreamType = {
      assert(dimOrder.nonEmpty, "dimOrder must have at least one element")
      if (dimOrder.length <= 1) {
        OrderedStreamType(dtv, lentvSeq(dimOrder.last.ae.evalInt))
      } else {
        OrderedStreamType(genReorderedStmType(dimOrder.init, lentvSeq, dtv), lentvSeq(dimOrder.last.ae.evalInt))
      }
    }

    val dtv = BasicDataTypeVar()
    val lentvSeq = Seq.fill(dimOrder.length)(ArithTypeVar())
    val inseqtv = OrderedStreamTypeVar(lentvSeq, dtv)
    val outseqt = genReorderedStmType(dimOrder, lentvSeq, dtv)
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(inseqtv, HWFunType(inseqtv, outseqt))), input.t), input)
  }, dimOrder)
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): TransposeNDOrderedStreamExpr = apply(inputs.head, types.asInstanceOf[Seq[ArithTypeT]])
  def unapply(expr: TransposeNDOrderedStreamExpr): Some[(Expr, Seq[ArithTypeT], Type)] =
    Some((expr.args.head, expr.types.asInstanceOf[Seq[ArithTypeT]], expr.t))
}

object Transpose extends BuiltinExprFun {
  override def args: Int = 1
  def apply(input: Expr): BuiltinExpr = {
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val height = inputtc.t.asInstanceOf[VectorTypeT].len.ae.evalInt
    val width = inputtc.t.asInstanceOf[VectorTypeT].et.asInstanceOf[VectorTypeT].len.ae.evalInt
    val v = ArithTypeVar()
    SplitVector(PermuteVector(JoinVector(inputtc), ArithLambdaType(v, v/width + (v%width) * height)), height)
  }
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): BuiltinExpr = apply(inputs.head)
}