package backend.hdl.arch.rewrite.transpose

import backend.hdl.{IntType, NamedTupleTypeT, OrderedStreamTypeT}
import backend.hdl.arch.{AddInt, ArchLambda, CounterInteger, IndexUnorderedStream, JoinOrderedStream, JoinStreamAll, JoinVector, MapOrderedStream, OrderedStreamToUnorderedStream, OrderedStreamToVector, PermuteVector, SplitOrderedStream, Transpose, TransposeNDOrderedStream, TruncInteger, Tuple, VectorToOrderedStream, ZipOrderedStream}
import backend.hdl.arch.rewrite.RulesHelper
import core.{ArithType, ArithTypeT, Expr, ParamDef, ParamUse, TypeChecker}
import core.rewrite.{Rule, RulesT}
import lift.arithmetic.{Cst, IntDiv, Mod, Prod, Sum}

object TransposeConversionRules extends RulesT {

  override protected def all(config: Option[Int]): Seq[Rule] =
    Seq(
      cleanTranspose,
      mergeTransposes,
      replacePermuteVecWithTransposeStm,
    )

  def cleanTranspose: Rule = Rule("cleanTranspose", {
    case TransposeNDOrderedStream(input, dimOrder, _) if RulesHelper.checkDimOrder(dimOrder) =>
      input
  })

  def mergeTransposes: Rule = Rule("mergeTransposes", {
    case TransposeNDOrderedStream(TransposeNDOrderedStream(input, dimOrder1, _), dimOrder2, _) if
      dimOrder1.length == dimOrder2.length =>
      val newDimOrder = RulesHelper.mergeDimOrders(dimOrder2, dimOrder1)
      TransposeNDOrderedStream(input, newDimOrder)
  })

  def convertTranspose2DIntoPermuteVec: Rule = Rule("convertTranspose2DIntoPermuteVec", {
    case TransposeNDOrderedStream(MapOrderedStream(ArchLambda(p1, VectorToOrderedStream(JoinVector(OrderedStreamToVector(
    MapOrderedStream(ArchLambda(p2, body, _), input2, _), _), _), _), _), input1, _), dimOrder, t: OrderedStreamTypeT) if dimOrder.length == 2 &&
      dimOrder.head.ae.evalInt == 1 && dimOrder.last.ae.evalInt == 0 &&
      t.len.ae.evalInt % t.et.asInstanceOf[OrderedStreamTypeT].len.ae.evalInt== 0 &&
      t.len.ae.evalInt > t.et.asInstanceOf[OrderedStreamTypeT].len.ae.evalInt =>
      JoinOrderedStream(
        MapOrderedStream(
          ArchLambda(
            p2,
            TransposeNDOrderedStream(MapOrderedStream(
              ArchLambda(
                p1,
                VectorToOrderedStream(body)
              ),
              input1
            ), Seq(1, 0))
          ), input2
        )
      )
    case TransposeNDOrderedStream(input, dimOrder, _) if dimOrder.length == 2 && dimOrder.head.ae.evalInt == 1 && dimOrder.last.ae.evalInt == 0 =>
      val inputVec = TypeChecker.check(OrderedStreamToVector(MapOrderedStream(OrderedStreamToVector.asFunction(), input)))
      val transposedInput = Transpose(inputVec)
      MapOrderedStream(VectorToOrderedStream.asFunction(), VectorToOrderedStream(transposedInput))

    case TransposeNDOrderedStream(input, dimOrder, _) if dimOrder.length == 3 && dimOrder.head.ae.evalInt == 0 && dimOrder(1).ae.evalInt == 2 && dimOrder.last.ae.evalInt == 1 =>
      val inputVec = TypeChecker.check(OrderedStreamToVector(MapOrderedStream(OrderedStreamToVector.asFunction(), MapOrderedStream(2, OrderedStreamToVector.asFunction(), input))))
      val transposedInput = Transpose(inputVec)
      MapOrderedStream(2, VectorToOrderedStream.asFunction(), MapOrderedStream(VectorToOrderedStream.asFunction(), VectorToOrderedStream(transposedInput)))

    case TransposeNDOrderedStream(input, dimOrder, _) if dimOrder.length == 3 && dimOrder.head.ae.evalInt == 1 && dimOrder(1).ae.evalInt == 2 && dimOrder.last.ae.evalInt == 0 =>
      val chunkSize = input.t.asInstanceOf[OrderedStreamTypeT].dimensions(1).ae.evalInt
      val inputVec = TypeChecker.check(OrderedStreamToVector(MapOrderedStream(OrderedStreamToVector.asFunction(),  JoinOrderedStream(input))))
      val transposedInput = Transpose(inputVec)
      MapOrderedStream(SplitOrderedStream.asFunction(Seq(None), Seq(chunkSize)), MapOrderedStream(VectorToOrderedStream.asFunction(), VectorToOrderedStream(transposedInput)))
  })

  def generateSumOfCounters(start: ArithTypeT, increment: ArithTypeT, loop: ArithTypeT, dimensions: Seq[ArithTypeT],
                            repetitions: Seq[ArithTypeT], dimOrder: Seq[ArithTypeT]): Expr = {
    val countIncr = dimensions.map(_.ae.evalInt).slice(0, dimOrder.last.ae.evalInt).product * increment.ae.evalInt
    val countStart = if(dimOrder.length == 1) start else ArithType(0)
    val counter = CounterInteger(countStart, countIncr, loop, Seq(dimensions(dimOrder.last.ae.evalInt)), Seq(repetitions(dimOrder.last.ae.evalInt)), None)
    val countertc = TypeChecker.check(counter)
    if(dimOrder.length <= 1){
      countertc
    } else {
      val prevSum = generateSumOfCounters(start, increment, loop, dimensions, repetitions, dimOrder.dropRight(1))
      val prevSumtc = TypeChecker.check(prevSum)
      val prevSumLevel = prevSumtc.t.asInstanceOf[OrderedStreamTypeT].dimensions.length
      MapOrderedStream(
        {
          val param1 = ParamDef(countertc.t.asInstanceOf[OrderedStreamTypeT].et)
          ArchLambda(
            param1,
            MapOrderedStream(
              prevSumLevel,
              {
                val param2 = ParamDef(prevSumtc.t.asInstanceOf[OrderedStreamTypeT].leafType)
                ArchLambda(
                  param2,
                  AddInt(Tuple(ParamUse(param1), ParamUse(param2)))
                )
              }, prevSumtc
            )
          )
        }, countertc
      )
    }
  }

  def mergeIntoCounter: Rule = Rule("mergeIntoCounter", {
    case TransposeNDOrderedStream(CounterInteger(start, increment, loop, dimensions, repetitions, t: OrderedStreamTypeT), dimOrder, _)  =>
      val bitWidth = t.leafType.bitWidth
      val sum = generateSumOfCounters(start, increment, loop, dimensions, repetitions, dimOrder)
      val sumtc = TypeChecker.check(sum)
      val sumLevel = dimOrder.length
      MapOrderedStream(
        sumLevel,
        {
          val param = ParamDef(sumtc.t.asInstanceOf[OrderedStreamTypeT].leafType)
          ArchLambda(
            param,
            TruncInteger(ParamUse(param), bitWidth)
          )
        }, sumtc
      )
  })

  def replacePermuteVecWithTransposeStm = Rule("replacePermuteVecWithTransposeStm", {
    case SplitOrderedStream(VectorToOrderedStream(PermuteVector(
    OrderedStreamToVector(JoinOrderedStream(input, _), _), arithFun, _), _), chunkSize, _) if {
      var exprFound = false
      arithFun.b.ae match {
        case Sum(IntDiv(v1, c1)::Prod(Cst(_)::Mod(v2, c2)::Nil)::Nil) if v1 == v2 && c1 == c2 =>
          exprFound = true
        case _=>
      }
      exprFound
    } =>
      TransposeNDOrderedStream(input, Seq(1,0))
  })

  def mergeIntoIndexStm: Rule = Rule("mergeIntoIndexStm", {
    case IndexUnorderedStream(OrderedStreamToUnorderedStream(joinTransposeInput, _), t) if {
      var exprFound = false
      val innerMostInput = RulesHelper.getCurrentOrJoinNDInput(joinTransposeInput)
      innerMostInput match {
        case TransposeNDOrderedStream(input, dimOrder, transposedType: OrderedStreamTypeT) if
          dimOrder.length - 1 == RulesHelper.getJoinNDLevels(joinTransposeInput) =>
          exprFound = true
        case _ =>
      }
      exprFound
    } =>
      // Get transposition's info
      var transposedDims: Seq[ArithTypeT] = Seq()
      var dimOrder: Seq[ArithTypeT] = Seq()
      var input: Expr = joinTransposeInput

      val innerMostInput = RulesHelper.getCurrentOrJoinNDInput(joinTransposeInput)
      innerMostInput match {
        case TransposeNDOrderedStream(i, order, tT: OrderedStreamTypeT) if
          order.length - 1 == RulesHelper.getJoinNDLevels(joinTransposeInput) =>
          transposedDims = tT.dimensions
          dimOrder = order
          input = i
        case _ =>
      }

      // Reshape address counter
      val addrBitWidth = t.asInstanceOf[OrderedStreamTypeT].et.asInstanceOf[NamedTupleTypeT].namedTypes.last._2.bitWidth
      val currentRange = math.pow(2, addrBitWidth.ae.evalInt).toInt
      val requiredBitWidth = transposedDims.map(_.ae.evalInt).product
      val inverseDimOrder = RulesHelper.invertDimOrder(dimOrder)
      val addrCounter = if(currentRange <= requiredBitWidth) {
        val tmpCounter = TypeChecker.check(TransposeNDOrderedStream(CounterInteger(0, 1, transposedDims, Some(IntType(addrBitWidth.ae.evalInt + 1))), inverseDimOrder))
        MapOrderedStream(
          transposedDims.length,
          {
            val param = ParamDef(tmpCounter.t.asInstanceOf[OrderedStreamTypeT].leafType)
            ArchLambda(
              param,
              TruncInteger(ParamUse(param), addrBitWidth)
            )
          }, tmpCounter
        )
      } else {
        TransposeNDOrderedStream(CounterInteger(0, 1, transposedDims, Some(IntType(addrBitWidth))), inverseDimOrder)
      }

      ZipOrderedStream(Tuple(JoinStreamAll(input), JoinStreamAll(addrCounter)))
  })

  def limitedConvertTransposeNDIntoPermute: Rule = Rule("limitedConvertTransposeNDIntoPermute", {
    case TransposeNDOrderedStream(input, dimOrder, _) if dimOrder.length == 4 && dimOrder(1).ae.evalInt == 2 && dimOrder(2).ae.evalInt == 1 && dimOrder(3).ae.evalInt == 3 && dimOrder(0).ae.evalInt == 0 =>
      val inputAsVec =  TypeChecker.check(MapOrderedStream(OrderedStreamToVector.asFunction(), MapOrderedStream(2, OrderedStreamToVector.asFunction(), MapOrderedStream(3, OrderedStreamToVector.asFunction(), input))))
      val transposed = MapOrderedStream({
        val param = ParamDef(inputAsVec.t.asInstanceOf[OrderedStreamTypeT].et)
        ArchLambda(
          param,
          Transpose(ParamUse(param))
        )
      }, inputAsVec)
      MapOrderedStream(3, VectorToOrderedStream.asFunction(), MapOrderedStream(2, VectorToOrderedStream.asFunction(), MapOrderedStream(VectorToOrderedStream.asFunction(), transposed)))

    case TransposeNDOrderedStream(input, dimOrder, t) if dimOrder.length == 3 && dimOrder(1).ae.evalInt == 0 && dimOrder(0).ae.evalInt == 1 =>
      val inputAsVec =  TypeChecker.check(MapOrderedStream(OrderedStreamToVector.asFunction(), MapOrderedStream(2, OrderedStreamToVector.asFunction(), input)))
      val transposed = MapOrderedStream({
        val param = ParamDef(inputAsVec.t.asInstanceOf[OrderedStreamTypeT].et)
        ArchLambda(
          param,
          Transpose(ParamUse(param))
        )
      }, inputAsVec)
      MapOrderedStream(2, VectorToOrderedStream.asFunction(), MapOrderedStream(VectorToOrderedStream.asFunction(), transposed))
  })
}
