package backend.hdl.vhdl

import backend.hdl.{IntType, LogicType, OrderedStreamTypeT, ScalarTypeT, SignedIntType, SignedIntTypeT, VectorTypeT}
import backend.hdl.graph._
import core.{ArithType, MalformedExprException}

import scala.collection.immutable.ListMap

object VhdlGenerator {

  def generate(topNode: NodeT): VhdlModule = {
    assert(topNode.checkConnections, "Graph contains invalid / is missing connections.")
    _generate(topNode)
  }

  private def generateModulePorts(ports: Seq[PortT]): Seq[(PortSimpleT, VhdlPort)] = ports.zipWithIndex.flatMap(p => generatePorts("p" + p._2, p._1))

  private def generatePorts(id: String, port: PortT): Seq[(PortSimpleT, VhdlPort)] = {
    val (direction, invDirection) = port match {
      case _: PortInT => (VhdlSignalDirection.IN, VhdlSignalDirection.OUT)
      case _: PortOutT => (VhdlSignalDirection.OUT, VhdlSignalDirection.IN)
    }

    port match {
      case ps: PortSimpleT =>
        Seq((ps, VhdlPort(id, "data", direction, ps.dataType)))
      case preq: PortRequestT =>
        Seq(
          (preq.data, VhdlPort(id, "req_data", direction, preq.data.dataType)),
          (preq.valid, VhdlPort(id, "req_valid", direction, preq.valid.dataType)),
          (preq.ready, VhdlPort(id, "req_ready", invDirection, preq.ready.dataType))
        )
      case prsp: PortResponseT =>
        Seq(
          (prsp.data, VhdlPort(id, "rsp_data", direction, prsp.data.dataType)),
          (prsp.valid, VhdlPort(id, "rsp_valid", direction, prsp.valid.dataType))
        )
      case ph: PortHandshakeT =>
        Seq(
          (ph.data, VhdlPort(id, "data", direction, ph.data.dataType)),
          (ph.last, VhdlPort(id, "last", direction, ph.last.dataType)),
          (ph.valid, VhdlPort(id, "valid", direction, ph.valid.dataType)),
          (ph.ready, VhdlPort(id, "ready", invDirection, ph.ready.dataType))
        )
      case onBoardRamPort: PortOnBoardRamBank =>
        Seq(
          (onBoardRamPort.waitRequest, VhdlPort(id, "bank_waitrequest", direction, onBoardRamPort.waitRequest.dataType)),
          (onBoardRamPort.readData, VhdlPort(id, "bank_readdata", direction, onBoardRamPort.readData.dataType)),
          (onBoardRamPort.readDataValid, VhdlPort(id, "bank_readdatavalid", direction, onBoardRamPort.readDataValid.dataType)),
          (onBoardRamPort.burstCount, VhdlPort(id, "bank_burstcount", invDirection, onBoardRamPort.burstCount.dataType)),
          (onBoardRamPort.writeData, VhdlPort(id, "bank_writedata", invDirection, onBoardRamPort.writeData.dataType)),
          (onBoardRamPort.address, VhdlPort(id, "bank_address", invDirection, onBoardRamPort.address.dataType)),
          (onBoardRamPort.write, VhdlPort(id, "bank_write", invDirection, onBoardRamPort.write.dataType)),
          (onBoardRamPort.read, VhdlPort(id, "bank_read", invDirection, onBoardRamPort.read.dataType)),
          (onBoardRamPort.byteEnable, VhdlPort(id, "bank_byteenable", invDirection, onBoardRamPort.byteEnable.dataType)),
          (onBoardRamPort.numCyclesWrites, VhdlPort(id, "num_cycles_writes", invDirection, onBoardRamPort.numCyclesWrites.dataType)),
          (onBoardRamPort.numCyclesReads, VhdlPort(id, "num_cycles_reads", invDirection, onBoardRamPort.numCyclesReads.dataType))
          //          (onBoardRamPort.write_req_count, VhdlPort(id, "write_req_count", invDirection, onBoardRamPort.write_req_count.dataType)),
//          (onBoardRamPort.write_rsp_count, VhdlPort(id, "write_rsp_count", invDirection, onBoardRamPort.write_rsp_count.dataType)),
//          (onBoardRamPort.read_req_count, VhdlPort(id, "read_req_count", invDirection, onBoardRamPort.read_req_count.dataType)),
//          (onBoardRamPort.read_rsp_count, VhdlPort(id, "read_rsp_count", invDirection, onBoardRamPort.read_rsp_count.dataType)),
//          (onBoardRamPort.obr_read_req_count, VhdlPort(id, "obr_read_req_count", invDirection, onBoardRamPort.obr_read_req_count.dataType)),
//          (onBoardRamPort.obr_write_req_count, VhdlPort(id, "obr_write_req_count", invDirection, onBoardRamPort.obr_write_req_count.dataType)),
//          (onBoardRamPort.obr_read_rsp_count, VhdlPort(id, "obr_read_rsp_count", invDirection, onBoardRamPort.obr_read_rsp_count.dataType)),
//          (onBoardRamPort.write_req_id_req_sum, VhdlPort(id, "write_req_id_req_sum", invDirection, onBoardRamPort.write_req_id_req_sum.dataType)),
//          (onBoardRamPort.write_req_id_rsp_sum, VhdlPort(id, "write_req_id_rsp_sum", invDirection, onBoardRamPort.write_req_id_rsp_sum.dataType)),
//          (onBoardRamPort.read_req_id_req_sum, VhdlPort(id, "read_req_id_req_sum", invDirection, onBoardRamPort.read_req_id_req_sum.dataType)),
//          (onBoardRamPort.read_req_id_rsp_sum, VhdlPort(id, "read_req_id_rsp_sum", invDirection, onBoardRamPort.read_req_id_rsp_sum.dataType)),
//          (onBoardRamPort.client_read_req_count, VhdlPort(id, "client_read_req_count", invDirection, onBoardRamPort.client_read_req_count.dataType)),
//          (onBoardRamPort.idx_buffer_count, VhdlPort(id, "idx_buffer_count", invDirection, onBoardRamPort.idx_buffer_count.dataType)),
//          (onBoardRamPort.information, VhdlPort(id, "OBR_controller_information", invDirection, onBoardRamPort.information.dataType))
        )
      case mem: PortMemoryRead =>
        Seq(
          (mem.memReady, VhdlPort("p", "read_mem_ready", direction, mem.memReady.dataType)),
          (mem.reqAddr, VhdlPort("p", "read_req_address", invDirection, mem.reqAddr.dataType)),
          (mem.reqMData, VhdlPort("p", "read_req_mdata", invDirection, mem.reqMData.dataType)),
          (mem.reqValid, VhdlPort("p", "read_req_valid", invDirection, mem.reqValid.dataType)),
          (mem.reqTotal, VhdlPort("p", "read_req_total", invDirection, mem.reqTotal.dataType)),
          (mem.reqPending, VhdlPort("p", "read_req_pending", invDirection, mem.reqPending.dataType)),
          (mem.reqALMFull, VhdlPort("p", "read_req_almFull", direction, mem.reqALMFull.dataType)),
          (mem.rspData, VhdlPort("p", "read_rsp_data", direction, mem.rspData.dataType)),
          (mem.rspMData, VhdlPort("p", "read_rsp_mdata", direction, mem.rspMData.dataType)),
          (mem.rspValid, VhdlPort("p", "read_rsp_valid", direction, mem.rspValid.dataType)),
        )
      case mem: PortMemoryWrite =>
        Seq(
          (mem.memReady, VhdlPort("p", "write_mem_ready", invDirection, mem.memReady.dataType)),
          (mem.reqAddr, VhdlPort("p", "write_req_address", direction, mem.reqAddr.dataType)),
          (mem.reqMData, VhdlPort("p", "write_req_mdata", direction, mem.reqMData.dataType)),
          (mem.reqData, VhdlPort("p", "write_req_data", direction, mem.reqData.dataType)),
          (mem.reqValid, VhdlPort("p", "write_req_valid", direction, mem.reqValid.dataType)),
          (mem.reqTotal, VhdlPort("p", "write_req_total", direction, mem.reqTotal.dataType)),
          (mem.reqPending, VhdlPort("p", "write_req_pending", direction, mem.reqPending.dataType)),
          (mem.reqALMFull, VhdlPort("p", "write_req_almFull", invDirection, mem.reqALMFull.dataType)),
          (mem.rspMData, VhdlPort("p", "write_rsp_mdata", invDirection, mem.rspMData.dataType)),
          (mem.rspValid, VhdlPort("p", "write_rsp_valid", invDirection, mem.rspValid.dataType)),
        )
    }
  }

  private def generateUniqueNodeName(node: NodeT): String = node match {
    case TopNode(_) => "top"
    case BehaviourNodeTestbench(uut, _) => "tb_" + uut.name
    case BehaviourNodeMarker(_, _, label, _, _) => node.name + "_" + node.id + "_" + label.s
    case _ => node.name + "_" + node.id
  }

  private def _generate(node: NodeT): VhdlModule = {
    // (direct!) subnodes of this node
    val subNodes: Seq[NodeT] = node.subGraph.toSeq.flatMap(_.nodes)

    // create VhdlPorts for all inputs and outputs of the current node
    val nodePorts: ListMap[PortSimpleT, VhdlPort] = ListMap(generateModulePorts(node.ports): _*)
    // create VhdlPorts for all inputs and outputs of direct subnodes of the current node
    val subNodePorts: Map[PortSimpleT, VhdlPort] = subNodes.flatMap(n => generateModulePorts(n.ports)).toMap ++
      (node match {
          // ... and for the behaviour of this node
        case bh: BehaviourNodeT => generateModulePorts(bh.behaviourPorts)
        case _ => Map()
      })

    // connections in this node
    val subConnections: Seq[Connection] = node.subGraph.toSeq.flatMap(_.connections)
    // external connections go from a subnode to a parent node; internal connections are between subnodes only
    val (externalConnections, internalConnections): (Seq[Connection], Seq[Connection]) = subConnections.partition(con => nodePorts.keys.exists(con.contains))
    // map external connections to existing vhdl signals from this node's ports
    val externalConnectionSignals: Map[Connection, VhdlSignalT] = externalConnections.map(con => con -> nodePorts.find(ps => con.contains(ps._1)).get._2).toMap
    // create new signals for internal connections
    val internalConnectionSignals: ListMap[Connection, VhdlSignalT] = {
      // distinct by port, because there may be more than 1 connection going (partly!) to a single target
      // here: source port (pout) defines the internal signal (source port type = internal signal type). (Could also be changed, but must be consistent!)
      // These connections must then have the same name/string. So they get the same index (using zipWithIndex)
      val ids: Map[PortSimpleT, Int] = internalConnections.map(_.pout).distinct.zipWithIndex.toMap
      ListMap(internalConnections.map(con => (con, VhdlSignal("s%02d".format(ids(con.pout)), subNodePorts(con.pout).purpose, con.pout.dataType))): _*)
    }
    val signals = nodePorts ++ (externalConnectionSignals ++ internalConnectionSignals).flatMap(con => Map(con._1.pin -> con._2, con._1.pout -> con._2))

    // add clock and reset port to every module, except testbenches
    val ports = node match {
      case BehaviourNodeTestbench(_, _) => nodePorts.values.toSeq
      case _ => Seq(VhdlPort.CLOCK, VhdlPort.RESET) ++ nodePorts.values.toSeq
    }

    // avoid multiple definitions of the same signal, by grouping them and taking the head of the group only
    val internalSignals = internalConnectionSignals.values.groupBy(sig => sig.id + "" + sig.purpose).values.map(_.head).toSeq.sortBy(_.id.s)
    // grouping of subnodes in order to generate only one component declaration per entity (and not per instance)
    val components = subNodes.groupBy(generateUniqueNodeName).values.map(_.head).map(generate).toSeq
    // create instantiations of hierarchy subnodes
    val instantiations = subNodes.zipWithIndex.map(sn =>
      VhdlModuleInstance("U" + sn._2 + "_" + sn._1.name, generateUniqueNodeName(sn._1), {
        val snPorts = sn._1.ports.flatMap(_.flatten)
        // connect instance's clock and reset to parent clock and reset
        Seq(VhdlConnection(VhdlPort.CLOCK, VhdlPort.CLOCK), VhdlConnection(VhdlPort.RESET, VhdlPort.RESET)) ++
          // go through all connections that are relevant for the current subnode
          subConnections.filter(con => snPorts.exists(con.contains)).map(con =>
            // rearrange connections: port of the component must be left in assignment
            if (internalConnectionSignals.contains(con)) {
              if (snPorts.contains(con.pout))
                VhdlConnection(subNodePorts(con.pout), internalConnectionSignals(con)) // directly + fully connect outgoing port to internal signal (NO PARTIAL CONNECTION)
              else
                VhdlConnection(subNodePorts(con.pin), internalConnectionSignals(con), con.pinPart, con.poutPart) // from the internal signal only connect the specified parts to the target (PARTIAL CONNECTION)
            } else {
              if (snPorts.contains(con.pout))
                VhdlConnection(subNodePorts(con.pout), nodePorts(con.pin), con.pinPart, con.poutPart) // no internal signal -> always apply partial connections!
              else
                VhdlConnection(subNodePorts(con.pin), nodePorts(con.pout), con.pinPart, con.poutPart) // no internal signal -> always apply partial connections!
            }
          ).distinct
      })
    )

    // NOTE if a new behaviour is added, only change the code below this line!

    VhdlModule(
      generateUniqueNodeName(node),
      // generics
      node match {
        case BehaviourNodeConstantValue(_, value, vType, _) =>
          Seq(
            vType match {
              case LogicType() => VhdlGeneric("value", VhdlConstantLogic(value.ae.evalInt != 0))
              case SignedIntType(_) => VhdlGeneric("value", VhdlConstantLogicVector(value.ae.evalInt, vType.bitWidth.ae.evalInt))
              case _ => VhdlGeneric("value", VhdlConstantLogicVector(value.ae.evalInt))
            },
            VhdlGeneric("loop_all", VhdlConstantLogic(true))
          )
        case BehaviourNodeCounterInteger(_, start, increment, loop, dimensions, repetitions, _) =>
          Seq(
            VhdlGeneric("start", VhdlConstantNatural(start.ae.evalInt)),
            VhdlGeneric("increment", VhdlConstantNatural(increment.ae.evalInt)),
            // reverse the dimensions! in HDL the representation is different to this IR
            VhdlGeneric("dimensions", VhdlConstantNaturalVector(dimensions.reverse.map(_.ae.evalInt))),
            VhdlGeneric("repetitions", VhdlConstantLogicVector(repetitions.reverse.map(_.evalBool))),
            VhdlGeneric("loop_all", VhdlConstantLogic(loop.evalBool))
          )
        case BehaviourNodeRepeat(_, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeRepeatHidden(_, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeDelay(_, _, cycles, _) => Seq(VhdlGeneric("delay_cycles", VhdlConstantNatural(cycles.ae.evalInt)))
        case BehaviourNodeMapOrderedStream(_, _, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeMapSimpleOrderedStream(_, _, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeMapUnorderedStream(_, _, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeMapBufferedUnorderedStream(_, _, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeFoldOrderedStream(_, _, _, dimensionsIn, _, _) if dimensionsIn.length == 2 => Seq(VhdlGeneric("inner_length", VhdlConstantNatural(dimensionsIn.head.ae.evalInt)))
        case BehaviourNodeSlideOrderedStream(_, _, windowWidth, _) => Seq(VhdlGeneric("window_width", VhdlConstantNatural(windowWidth.ae.evalInt)))
        case BehaviourNodeUnorderedStreamToOrderedStream(_, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeOrderedStreamToUnorderedStream(_, _, len, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)))
        case BehaviourNodeSplitOrderedStream(_, _, innerLength, outerLength, _) => Seq(VhdlGeneric("inner_stream_length", VhdlConstantNatural(innerLength.ae.evalInt)), VhdlGeneric("outer_stream_length", VhdlConstantNatural(outerLength.ae.evalInt)))
        case BehaviourNodeJoinUnorderedStream(_, _, innerLength, _) => Seq(VhdlGeneric("inner_stream_length", VhdlConstantNatural(innerLength.ae.evalInt)))
        case BehaviourNodeDropOrderedStream(_, _, len, firstElements, lastElements, _) => Seq(VhdlGeneric("stream_length", VhdlConstantNatural(len.ae.evalInt)), VhdlGeneric("first_elements", VhdlConstantNatural(firstElements.ae.evalInt)), VhdlGeneric("last_elements", VhdlConstantNatural(lastElements.ae.evalInt)))
        case BehaviourNodeDropVector(_, _, lowElements, highElements, _) => Seq(VhdlGeneric("low_elements", VhdlConstantNatural(lowElements.ae.evalInt)), VhdlGeneric("high_elements", VhdlConstantNatural(highElements.ae.evalInt)))
        case BehaviourNodeConcatOrderedStream(_, _, _, dimension, _) => Seq(VhdlGeneric("concat_dimension", VhdlConstantNatural(dimension.ae.evalInt)))
        case BehaviourNodeRotateVector(_, _, rotateLeft, _) => Seq(VhdlGeneric("rotate_left", VhdlConstantLogic(rotateLeft.evalBool)))
        case BehaviourNodeSlideVector(_, _, _, stepSize, _) => Seq(VhdlGeneric("step_size", VhdlConstantNatural(stepSize.ae.evalInt)))
        case BehaviourNodeBlockRam(_, pout, entries, _) => Seq(VhdlGeneric("entries", VhdlConstantNatural(entries.ae.evalInt)), VhdlGeneric("data_width", VhdlConstantNatural(pout.dataType.bitWidth.ae.evalInt)))
        case BehaviourNodeBankedBlockRam(_, pout, entries, numBanks, _) => Seq(VhdlGeneric("entries", VhdlConstantNatural(entries.ae.evalInt)), VhdlGeneric("data_width", VhdlConstantNatural(pout.dataType.bitWidth.ae.evalInt)), VhdlGeneric("num_banks", VhdlConstantNatural(numBanks.ae.evalInt)))
        case BehaviourNodeReadAsync(_, _, _, _, numElements, _) =>  Seq(VhdlGeneric("data_length", VhdlConstantNatural(numElements.ae.evalInt)))
        case BehaviourNodeReadAsyncOrdered(_, _, _, _, numElements, _) =>  Seq(VhdlGeneric("data_length", VhdlConstantNatural(numElements.ae.evalInt)))
        case BehaviourNodeWriteAsync(_, _, _, _, numElements, _) =>  Seq(VhdlGeneric("data_length", VhdlConstantNatural(numElements.ae.evalInt)))
        case BehaviourNodeReadHostMemoryController(_, _, reqBufferSize, _) => Seq(VhdlGeneric("req_buffer_size", VhdlConstantNatural(reqBufferSize.ae.evalInt)))
        case BehaviourNodeWriteHostMemoryController(_, _, reqBufferSize, _) => Seq(VhdlGeneric("req_buffer_size", VhdlConstantNatural(reqBufferSize.ae.evalInt)))
        case BehaviourNodeDistributor(_, toClientsData, _, _, _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(toClientsData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)))
        case BehaviourNodeCollectorArbiter(fromClientsData, _, _, _, _, _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(fromClientsData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)))
        case BehaviourNodeVectorFork(_, toClientsData, _, _, _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(toClientsData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)))
        case BehaviourNodeVectorJoin(fromClientsData, _, _, _, _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(fromClientsData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)))
        case BehaviourNodeArbiterSyncFunction(_, _, toClientsData, _, _, _, _, _, _, _,  _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(toClientsData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)))
        case BehaviourNodeArbiterAsyncFunction(_, _, toClientsReqData, _, _, _, _, _, _) => Seq(VhdlGeneric("num_clients", VhdlConstantNatural(toClientsReqData.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt)), VhdlGeneric("req_list_size", VhdlConstantNatural(64))) // TODO get this number 64 from the ramType somehow?? in the Graph Generator
        case _ => Seq()
      },
      ports,
      components,
      internalSignals,
      instantiations,
      // statements
      node match {
        case BehaviourNodeBooleanLogic(logic, pins, pout, _) =>
          Seq(VhdlStatement(s"${nodePorts(pout)} <= ${pins.map(pin => nodePorts(pin)).mkString("", " and " ,"")};")) // TODO generate logic code + unfold vectorType / forbid vector Type?
        case BehaviourNodePermuteVector(pin, pout, permuteFun, _) =>
          val vectorLen = pin.dataType.asInstanceOf[VectorTypeT].len.ae.evalInt
          val indexSeq = Seq.range(0, vectorLen, 1)
          indexSeq.map(idx =>
            VhdlStatement(s"${nodePorts(pout.data)}(%d) <= ${nodePorts(pin.data)}(%d);".format(permuteFun.call(ArithType(idx)).ae.evalInt, idx))
          )
        case _ => Seq()
      },
      // template
      template = node match {
        case b: BehaviourNodeT => b match {
          /*
           *************************************************************************************************************
           value/input generators
           *************************************************************************************************************
           */
          case BehaviourNodeConstantValue(pout, _, vType, _) =>
            Some(VhdlBehaviourTemplate(
              vType match {
                case LogicType() => "const_bit"
                case _ => "const_bit_vector"
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid))
            ))
          case BehaviourNodeCounterInteger(pout, _, _, _, dimensions, _, _) =>
            Some(VhdlBehaviourTemplate("counter",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("natural_vector_type" -> VhdlConstantNaturalVector(dimensions.map(_.ae.evalInt)).vType)
            ))
          case b @ BehaviourNodeRepeat(_, pout, _, _) =>
            Some(VhdlBehaviourTemplate("repeat",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready))
            ))
          case b @ BehaviourNodeRepeatHidden(_, pout, _, _) =>
            Some(VhdlBehaviourTemplate("repeat_hidden",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready))
            ))
          case b @ BehaviourNodeVectorGenerator(_, pout, _) =>
            Some(VhdlBehaviourTemplate("vector_generator",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready))
            ))
          case BehaviourNodeSink(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("sink",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeRegistered(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("registered",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          case BehaviourNodeDelay(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("delay",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case b @ BehaviourNodeMarker(_, pout, _, _, _) =>
            Some(VhdlBehaviourTemplate("marker",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready))
            ))
          /*
           *************************************************************************************************************
           simple operations
           *************************************************************************************************************
           */
          case BehaviourNodeId(pin, pout, _) =>
            (pin, pout) match {
              case (pin: PortInHandshake, pout: PortOutHandshake) =>
                Some(VhdlBehaviourTemplate("id_hs",
                  Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                    "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                    "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
                ))
              case (pin: PortInRequest, pout: PortOutRequest) =>
                Some(VhdlBehaviourTemplate("id_req",
                  Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                    "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready), "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
                ))
              case (pin: PortInResponse, pout: PortOutResponse) =>
                Some(VhdlBehaviourTemplate("id_rsp",
                  Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                    "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid)),
                ))
              case _ => throw VhdlCodeException("Node with id behaviour cannot be generated for these ports (different type maybe?): " + pin + " and " + pout)
            }
          case BehaviourNodeSigned(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("signed",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeUnsigned(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("unsigned",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeAddInt(pin, pout, _) =>
            Some(VhdlBehaviourTemplate(
              pout.dataType match {
                case SignedIntType(_) => "add_signed_int"
                case IntType(_) => "add_int"
                case _ => ???
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeAddFloat(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("add_float",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeSubInt(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("sub_int",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeMulInt(pin, pout, _) =>
            Some(VhdlBehaviourTemplate(
              pout.dataType match {
                case SignedIntType(_) => "mul_signed_int"
                case IntType(_) => "mul_int"
                case _ => ???
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeMulFloat(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("mul_float",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeMaxInt(pin, pout, _) =>
            pout.dataType match {
              case t: SignedIntTypeT =>
                Some(VhdlBehaviourTemplate("max_signed_int",
                  Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                    "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                    "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
                ))
              case _ =>
                Some(VhdlBehaviourTemplate("max_int",
                  Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                    "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                    "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
                ))
            }
          case BehaviourNodeSelect(pin, pout, selection, reduction, _) =>
            Some(VhdlBehaviourTemplate(
              (selection.ae.evalInt, reduction) match {
                case (0, false) => "sel0"
                case (1, false) => "sel1"
                case (0, true) => "sel0_reduced"
                case (1, true) => "sel1_reduced"
                case _ => throw MalformedExprException("Selection integer value must be 0 or 1.")
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pout.dataType)
            ))
          case BehaviourNodeSelect3(pin, pout, selection, _) =>
            Some(VhdlBehaviourTemplate(
              selection.ae.evalInt match {
                case 0 => "sel0"
                case 1 => "sel1"
                case 2 => "sel2"
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          /*
           *************************************************************************************************************
           complex operations (map, reduce, slide, ...)
           *************************************************************************************************************
           */
          case b@BehaviourNodeMapOrderedStream(_, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("map_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_from_f_data" -> signals(b.function.pin.data), "port_from_f_last" -> signals(b.function.pin.last), "port_from_f_valid" -> signals(b.function.pin.valid), "port_from_f_ready" -> signals(b.function.pin.ready),
                "port_to_f_data" -> signals(b.function.pouts.head.data), "port_to_f_last" -> signals(b.function.pouts.head.last), "port_to_f_valid" -> signals(b.function.pouts.head.valid), "port_to_f_ready" -> signals(b.function.pouts.head.ready),
                "delay_ctrl_last" -> signals(b.delayCtrlLastOut), "delay_ctrl_ready" -> signals(b.delayCtrlReadyOut)),
              Map("data_from_f_type" -> b.function.pin.dataType, "data_to_f_type" -> b.function.pouts.head.dataType)
            ))
          case b@BehaviourNodeMapSimpleOrderedStream(_, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("map_simple_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready),
                "port_f_out_data" -> signals(b.function.pouts.head.data), "port_f_out_last" -> signals(b.function.pouts.head.last), "port_f_out_valid" -> signals(b.function.pouts.head.valid), "port_f_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeMapUnorderedStream(_, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("map_unordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready),
                "port_f_out_data" -> signals(b.function.pouts.head.data), "port_f_out_last" -> signals(b.function.pouts.head.last), "port_f_out_valid" -> signals(b.function.pouts.head.valid), "port_f_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeMapBufferedUnorderedStream(_, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("map_buffered_unordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready),
                "port_f_out_data" -> signals(b.function.pouts.head.data), "port_f_out_last" -> signals(b.function.pouts.head.last), "port_f_out_valid" -> signals(b.function.pouts.head.valid), "port_f_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case BehaviourNodeMapStreamDelay(pin, pout, controlLastPin, controlReadyPin, _) =>
            Some(VhdlBehaviourTemplate("map_stream_delay",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_in_control_last" -> signals(controlLastPin),
                "port_in_control_ready" -> signals(controlReadyPin))
            ))
          case b@BehaviourNodeFoldOrderedStream(_, pin, pout, dimensionsIn, bufferDataType, _) =>
              Some(VhdlBehaviourTemplate(
                (dimensionsIn.length, bufferDataType) match {
                  case (1, IntType(_)) => "fold_integer_ordered_stream"
                  case (1, _) => "fold_ordered_stream"
                  case (2, _) => "fold_ordered_stream_2D"
                  case _ => ??? // unsupported for now!
                },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_out_acc_data" -> signals(b.function.pouts.head.data), "port_f_out_acc_last" -> signals(b.function.pouts.head.last), "port_f_out_acc_valid" -> signals(b.function.pouts.head.valid), "port_f_out_acc_ready" -> signals(b.function.pouts.head.ready),
                "port_f_out_data_data" -> signals(b.function.pouts(1).data), "port_f_out_data_last" -> signals(b.function.pouts(1).last), "port_f_out_data_valid" -> signals(b.function.pouts(1).valid), "port_f_out_data_ready" -> signals(b.function.pouts(1).ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready)),
              Map("data_word_type" -> pout.dataType)
            ))
          case b@BehaviourNodeReduceOrderedStream(_, pinInitialValue, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("reduce_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_init_data" -> signals(pinInitialValue.data), "port_in_init_last" -> signals(pinInitialValue.last), "port_in_init_valid" -> signals(pinInitialValue.valid), "port_in_init_ready" -> signals(pinInitialValue.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_f_out_acc_data" -> signals(b.function.pouts.head.data), "port_f_out_acc_last" -> signals(b.function.pouts.head.last), "port_f_out_acc_valid" -> signals(b.function.pouts.head.valid), "port_f_out_acc_ready" -> signals(b.function.pouts.head.ready),
                "port_f_out_data_data" -> signals(b.function.pouts(1).data), "port_f_out_data_last" -> signals(b.function.pouts(1).last), "port_f_out_data_valid" -> signals(b.function.pouts(1).valid), "port_f_out_data_ready" -> signals(b.function.pouts(1).ready),
                "port_f_in_data" -> signals(b.function.pin.data), "port_f_in_last" -> signals(b.function.pin.last), "port_f_in_valid" -> signals(b.function.pin.valid), "port_f_in_ready" -> signals(b.function.pin.ready)),
              Map("data_word_type" -> pout.dataType)
            ))
          case BehaviourNodeSlideOrderedStream(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("slide_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          case b@BehaviourNodeAlternate(_, _, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("alternate",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_to_f1_data" -> signals(b.function1.pouts.head.data), "port_to_f1_last" -> signals(b.function1.pouts.head.last), "port_to_f1_valid" -> signals(b.function1.pouts.head.valid), "port_to_f1_ready" -> signals(b.function1.pouts.head.ready),
                "port_from_f1_data" -> signals(b.function1.pin.data), "port_from_f1_last" -> signals(b.function1.pin.last), "port_from_f1_valid" -> signals(b.function1.pin.valid), "port_from_f1_ready" -> signals(b.function1.pin.ready),
                "port_to_f2_data" -> signals(b.function2.pouts.head.data), "port_to_f2_last" -> signals(b.function2.pouts.head.last), "port_to_f2_valid" -> signals(b.function2.pouts.head.valid), "port_to_f2_ready" -> signals(b.function2.pouts.head.ready),
                "port_from_f2_data" -> signals(b.function2.pin.data), "port_from_f2_last" -> signals(b.function2.pin.last), "port_from_f2_valid" -> signals(b.function2.pin.valid), "port_from_f2_ready" -> signals(b.function2.pin.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          case b@BehaviourNodeIfelse(_, _, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("ifelse",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_to_f1_data" -> signals(b.function1.pouts.head.data), "port_to_f1_last" -> signals(b.function1.pouts.head.last), "port_to_f1_valid" -> signals(b.function1.pouts.head.valid), "port_to_f1_ready" -> signals(b.function1.pouts.head.ready),
                "port_from_f1_data" -> signals(b.function1.pin.data), "port_from_f1_last" -> signals(b.function1.pin.last), "port_from_f1_valid" -> signals(b.function1.pin.valid), "port_from_f1_ready" -> signals(b.function1.pin.ready),
                "port_to_f2_data" -> signals(b.function2.pouts.head.data), "port_to_f2_last" -> signals(b.function2.pouts.head.last), "port_to_f2_valid" -> signals(b.function2.pouts.head.valid), "port_to_f2_ready" -> signals(b.function2.pouts.head.ready),
                "port_from_f2_data" -> signals(b.function2.pin.data), "port_from_f2_last" -> signals(b.function2.pin.last), "port_from_f2_valid" -> signals(b.function2.pin.valid), "port_from_f2_ready" -> signals(b.function2.pin.ready)),
              Map("data_word_tuple_type" -> pin.dataType, "data_word_type" -> pout.dataType)
            ))
          /*
           *************************************************************************************************************
           data type conversions / rearrangement
           *************************************************************************************************************
           */
          case BehaviourNodeConversion(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("id_hs",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeVectorToTuple(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("vector_to_tuple",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeVectorToOrderedStream(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("vector_to_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeOrderedStreamToVector(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("ordered_stream_to_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_vector_type" -> pout.dataType, "data_word_type" -> pin.dataType)
            ))
          case BehaviourNodeIndexUnorderedStream(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("index_unordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeUnorderedStreamToOrderedStream(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("unordered_stream_to_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pout.dataType)
            ))
          case BehaviourNodeOrderedStreamToUnorderedStream(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("ordered_stream_to_unordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeTuple(pin1, pin2, pout, _) =>
            Some(VhdlBehaviourTemplate("tuple",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_1_data" -> signals(pin1.data), "port_in_1_last" -> signals(pin1.last), "port_in_1_valid" -> signals(pin1.valid), "port_in_1_ready" -> signals(pin1.ready),
                "port_in_2_data" -> signals(pin2.data), "port_in_2_last" -> signals(pin2.last), "port_in_2_valid" -> signals(pin2.valid), "port_in_2_ready" -> signals(pin2.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeTuple3(pin1, pin2, pin3, pout, _) =>
            Some(VhdlBehaviourTemplate("tuple3",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_1_data" -> signals(pin1.data), "port_in_1_last" -> signals(pin1.last), "port_in_1_valid" -> signals(pin1.valid), "port_in_1_ready" -> signals(pin1.ready),
                "port_in_2_data" -> signals(pin2.data), "port_in_2_last" -> signals(pin2.last), "port_in_2_valid" -> signals(pin2.valid), "port_in_2_ready" -> signals(pin2.ready),
                "port_in_3_data" -> signals(pin3.data), "port_in_3_last" -> signals(pin3.last), "port_in_3_valid" -> signals(pin3.valid), "port_in_3_ready" -> signals(pin3.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeFlipTuple(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("flip_tuple",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeZipStream(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("zip_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeZip3Stream(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("zip3_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeZipVector(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("zip_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeSplitOrderedStream(pin, pout, _, _, _) =>
            Some(VhdlBehaviourTemplate("split_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeSplitVector(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("split_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeJoinOrderedStream(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("join_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeJoinUnorderedStream(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("join_unordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeJoinVector(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("join_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeDropOrderedStream(pin, pout, _, _, _, _) =>
            Some(VhdlBehaviourTemplate("drop_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          case BehaviourNodeDropVector(pin, pout, _, _, _) =>
            Some(VhdlBehaviourTemplate("drop_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_vector_type" -> pin.dataType, "data_word_vector_slice_type" -> pout.dataType)
            ))
          case BehaviourNodeConcatOrderedStream(pin1, pin2, pout, _, _) =>
            Some(VhdlBehaviourTemplate("concat_ordered_stream",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_1_data" -> signals(pin1.data), "port_in_1_last" -> signals(pin1.last), "port_in_1_valid" -> signals(pin1.valid), "port_in_1_ready" -> signals(pin1.ready),
                "port_in_2_data" -> signals(pin2.data), "port_in_2_last" -> signals(pin2.last), "port_in_2_valid" -> signals(pin2.valid), "port_in_2_ready" -> signals(pin2.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeConcatVector(pin, pout, lowDataType, highDataType, wordType, _) =>
            Some(VhdlBehaviourTemplate("concat_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_vector_slice_low_type" -> lowDataType, "data_word_vector_slice_high_type" -> highDataType, "data_word_vector_type" -> pout.dataType, "data_word_type" -> wordType)
            ))
          case BehaviourNodeUpdateVector(pin, pout, _) =>
            Some(VhdlBehaviourTemplate("update_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case BehaviourNodeRotateVector(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("rotate_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pout.dataType.asInstanceOf[VectorTypeT].et)
            ))
          case BehaviourNodePermuteVector(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("permute_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          case BehaviourNodeSlideVector(pin, pout, _, _, _) =>
            Some(VhdlBehaviourTemplate("slide_vector",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready)),
              Map("data_word_type" -> pin.dataType)
            ))
          /*
           *************************************************************************************************************
           memory related behaviour
           *************************************************************************************************************
           */
          case b@BehaviourNodeReadSync(_, pinBaseAddr, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("read_sync",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeWriteSync(_, pinBaseAddr, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("write_sync",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeReadSyncMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("read_sync_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeWriteSyncMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("write_sync_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case BehaviourNodeBlockRam(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("ram_buffer",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          case b@BehaviourNodeReadSyncBanked(_, pinBaseAddr, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("read_sync_banked",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeWriteSyncBanked(_, pinBaseAddr, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("write_sync_banked",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeReadSyncBankedMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("read_sync_banked_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case b@BehaviourNodeWriteSyncBankedMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("write_sync_banked_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_last" -> signals(b.function.pin.last), "port_mem_in_valid" -> signals(b.function.pin.valid), "port_mem_in_ready" -> signals(b.function.pin.ready),
                "port_mem_out_data" -> signals(b.function.pouts.head.data), "port_mem_out_last" -> signals(b.function.pouts.head.last), "port_mem_out_valid" -> signals(b.function.pouts.head.valid), "port_mem_out_ready" -> signals(b.function.pouts.head.ready))
            ))
          case BehaviourNodeBankedBlockRam(pin, pout, _, _, _) =>
            Some(VhdlBehaviourTemplate("banked_ram_buffer",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          //TODO: Change made here
          case b@BehaviourNodeOnBoardRam(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("on_board_ram_buffer",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "bank_waitrequest" -> signals(b.onBoardRamBankPort.waitRequest), "bank_read_data" -> signals(b.onBoardRamBankPort.readData), "bank_read_valid" -> signals(b.onBoardRamBankPort.readDataValid),
                "bank_burstcount" -> signals(b.onBoardRamBankPort.burstCount), "bank_write_data" -> signals(b.onBoardRamBankPort.writeData), "bank_address" -> signals(b.onBoardRamBankPort.address),
                "bank_write_enable" -> signals(b.onBoardRamBankPort.write), "bank_read_enable" -> signals(b.onBoardRamBankPort.read), "bank_byteenable" -> signals(b.onBoardRamBankPort.byteEnable),
//                "port_out_write_req_count" -> signals(b.onBoardRamBankPort.write_req_count), "port_out_write_rsp_count" -> signals(b.onBoardRamBankPort.write_rsp_count),
//                "port_out_read_req_count" -> signals(b.onBoardRamBankPort.read_req_count), "port_out_read_rsp_count" -> signals(b.onBoardRamBankPort.read_rsp_count),
//                "port_out_obr_write_req_count" -> signals(b.onBoardRamBankPort.obr_write_req_count), "port_out_obr_read_req_count" -> signals(b.onBoardRamBankPort.obr_read_req_count), "port_out_obr_read_rsp_count" -> signals(b.onBoardRamBankPort.obr_read_rsp_count),
//                "port_out_write_req_id_req_sum" -> signals(b.onBoardRamBankPort.write_req_id_req_sum), "port_out_write_req_id_rsp_sum" -> signals(b.onBoardRamBankPort.write_req_id_rsp_sum),
//                "port_out_read_req_id_req_sum" -> signals(b.onBoardRamBankPort.read_req_id_req_sum), "port_out_read_req_id_rsp_sum" -> signals(b.onBoardRamBankPort.read_req_id_rsp_sum),
//                "port_out_client_read_req_count" -> signals(b.onBoardRamBankPort.client_read_req_count), "port_out_idx_buffer_count" -> signals(b.onBoardRamBankPort.idx_buffer_count),
//                "port_out_information" -> signals(b.onBoardRamBankPort.information),
                "port_out_num_cycles_writes" -> signals(b.onBoardRamBankPort.numCyclesWrites),
                "port_out_num_cycles_reads" -> signals(b.onBoardRamBankPort.numCyclesReads),
                "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid))
            ))

          case b@BehaviourNodeReadLocalMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("read_local_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid),
                "port_in_rsp_data" -> signals(b.function.pin.data), "port_in_rsp_valid" -> signals(b.function.pin.valid),
                "port_out_req_data" -> signals(b.function.pout.data), "port_out_req_valid" -> signals(b.function.pout.valid), "port_in_mem_ready" -> signals(b.function.pout.ready))
            ))

          case b@BehaviourNodeWriteLocalMemoryController(_, pin, pout, _) =>
            Some(VhdlBehaviourTemplate("write_local_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid),
                "port_in_rsp_data" -> signals(b.function.pin.data), "port_in_rsp_valid" -> signals(b.function.pin.valid),
                "port_out_req_data" -> signals(b.function.pout.data), "port_out_req_valid" -> signals(b.function.pout.valid), "port_in_mem_ready" -> signals(b.function.pout.ready))
            ))

          case b@BehaviourNodeReadAsync(_, pinBaseAddr, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("read_async",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_valid" -> signals(b.function.pin.valid),
                "port_mem_out_data" -> signals(b.function.pout.data), "port_mem_out_valid" -> signals(b.function.pout.valid), "port_mem_out_ready" -> signals(b.function.pout.ready))
            ))
          case b@BehaviourNodeReadAsyncOrdered(_, pinBaseAddr, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("read_async_ordered",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_valid" -> signals(b.function.pin.valid),
                "port_mem_out_data" -> signals(b.function.pout.data), "port_mem_out_valid" -> signals(b.function.pout.valid), "port_mem_out_ready" -> signals(b.function.pout.ready))
            ))
          case b@BehaviourNodeWriteAsync(_, pinBaseAddr, pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("write_async",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_baseaddr_data" -> signals(pinBaseAddr.data), "port_in_baseaddr_last" -> signals(pinBaseAddr.last), "port_in_baseaddr_valid" -> signals(pinBaseAddr.valid), "port_in_baseaddr_ready" -> signals(pinBaseAddr.ready),
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready),
                "port_mem_in_data" -> signals(b.function.pin.data), "port_mem_in_valid" -> signals(b.function.pin.valid),
                "port_mem_out_data" -> signals(b.function.pout.data), "port_mem_out_valid" -> signals(b.function.pout.valid), "port_mem_out_ready" -> signals(b.function.pout.ready)),
            ))
          case b@BehaviourNodeReadHostMemoryController(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("read_host_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready), "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid),
                "port_in_mem_ready" -> signals(b.memReadPort.memReady),
                "port_out_req_address" -> signals(b.memReadPort.reqAddr), "port_out_req_mdata" -> signals(b.memReadPort.reqMData), "port_out_req_valid" -> signals(b.memReadPort.reqValid), "port_out_req_total" -> signals(b.memReadPort.reqTotal), "port_out_req_pending" -> signals(b.memReadPort.reqPending), "port_in_req_almFull" -> signals(b.memReadPort.reqALMFull),
                "port_in_rsp_data" -> signals(b.memReadPort.rspData), "port_in_rsp_mdata" -> signals(b.memReadPort.rspMData), "port_in_rsp_valid" -> signals(b.memReadPort.rspValid))
            ))
          case b@BehaviourNodeWriteHostMemoryController(pin, pout, _, _) =>
            Some(VhdlBehaviourTemplate("write_host_memory_controller",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready), "port_out_data" -> signals(pout.data), "port_out_valid" -> signals(pout.valid),
                "port_in_mem_ready" -> signals(b.memWritePort.memReady),
                "port_out_req_address" -> signals(b.memWritePort.reqAddr), "port_out_req_mdata" -> signals(b.memWritePort.reqMData), "port_out_req_data" -> signals(b.memWritePort.reqData), "port_out_req_valid" -> signals(b.memWritePort.reqValid), "port_out_req_total" -> signals(b.memWritePort.reqTotal), "port_out_req_pending" -> signals(b.memWritePort.reqPending), "port_in_req_almFull" -> signals(b.memWritePort.reqALMFull),
                "port_in_rsp_mdata" -> signals(b.memWritePort.rspMData), "port_in_rsp_valid" -> signals(b.memWritePort.rspValid))
            ))
          /*
           ***********************************************************************************************************************
           device specific behaviour
           ***********************************************************************************************************************
           */
          case BehaviourNodeMul2AddInt(pin, pout, _) =>
            Some(VhdlBehaviourTemplate(
              pout.dataType match {
                case SignedIntType(_) => "mul2add_signed_int"
                case IntType(_) => "mul2add_int"
                case _ => ???
              },
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(pin.data), "port_in_last" -> signals(pin.last), "port_in_valid" -> signals(pin.valid), "port_in_ready" -> signals(pin.ready),
                "port_out_data" -> signals(pout.data), "port_out_last" -> signals(pout.last), "port_out_valid" -> signals(pout.valid), "port_out_ready" -> signals(pout.ready))
            ))
          /*
           *************************************************************************************************************
           general purpose behaviour
           *************************************************************************************************************
           */
          case BehaviourNodeTestbench(uut, _) =>
            //TODO use? ports.flatMap(_.flatten).map(signals)
            Some(VhdlBehaviourTemplate("testbench",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET)
            ))
          case BehaviourNodeDistributor(fromMaster, toClientsData, toClientsLast, toClientsValid, toClientsReady, _) =>
            Some(VhdlBehaviourTemplate("distributor",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(fromMaster.data), "port_in_last" -> signals(fromMaster.last), "port_in_valid" -> signals(fromMaster.valid), "port_in_ready" -> signals(fromMaster.ready),
                "port_out_clients_data" -> signals(toClientsData), "port_out_clients_last" -> signals(toClientsLast), "port_out_clients_valid" -> signals(toClientsValid), "port_out_clients_ready" -> signals(toClientsReady))
            ))
          case BehaviourNodeCollectorArbiter(fromClientsData, fromClientsLast, fromClientsValid, fromClientsReady, toMaster, strategy, _) =>
            // TODO strategy
            Some(VhdlBehaviourTemplate("collector_arbiter",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_clients_data" -> signals(fromClientsData), "port_in_clients_last" -> signals(fromClientsLast), "port_in_clients_valid" -> signals(fromClientsValid), "port_in_clients_ready" -> signals(fromClientsReady),
                "port_out_data" -> signals(toMaster.data), "port_out_last" -> signals(toMaster.last), "port_out_valid" -> signals(toMaster.valid), "port_out_ready" -> signals(toMaster.ready))
            ))
          case BehaviourNodeVectorFork(fromMaster, toClientsData, toClientsLast, toClientsValid, toClientsReady, _) =>
            Some(VhdlBehaviourTemplate("vector_fork",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(fromMaster.data), "port_in_last" -> signals(fromMaster.last), "port_in_valid" -> signals(fromMaster.valid), "port_in_ready" -> signals(fromMaster.ready),
                "port_out_clients_data" -> signals(toClientsData), "port_out_clients_last" -> signals(toClientsLast), "port_out_clients_valid" -> signals(toClientsValid), "port_out_clients_ready" -> signals(toClientsReady))
            ))
          case BehaviourNodeVectorJoin(fromClientsData, fromClientsLast, fromClientsValid, fromClientsReady, toMaster, _) =>
            Some(VhdlBehaviourTemplate("vector_join",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_clients_data" -> signals(fromClientsData), "port_in_clients_last" -> signals(fromClientsLast), "port_in_clients_valid" -> signals(fromClientsValid), "port_in_clients_ready" -> signals(fromClientsReady),
                "port_out_data" -> signals(toMaster.data), "port_out_last" -> signals(toMaster.last), "port_out_valid" -> signals(toMaster.valid), "port_out_ready" -> signals(toMaster.ready))
            ))
          case BehaviourNodeArbiterSyncFunction(toMaster, fromMaster, toClientsData, toClientsLast, toClientsValid, toClientsReady, fromClientsData, fromClientsLast, fromClientsValid, fromClientsReady, strategy, _) =>
            // TODO strategy
            Some(VhdlBehaviourTemplate("arbiter_sync_function",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_in_data" -> signals(fromMaster.data), "port_in_last" -> signals(fromMaster.last), "port_in_valid" -> signals(fromMaster.valid), "port_in_ready" -> signals(fromMaster.ready),
                "port_out_data" -> signals(toMaster.data), "port_out_last" -> signals(toMaster.last), "port_out_valid" -> signals(toMaster.valid), "port_out_ready" -> signals(toMaster.ready),
                "port_in_clients_data" -> signals(fromClientsData), "port_in_clients_last" -> signals(fromClientsLast), "port_in_clients_valid" -> signals(fromClientsValid), "port_in_clients_ready" -> signals(fromClientsReady),
                "port_out_clients_data" -> signals(toClientsData), "port_out_clients_last" -> signals(toClientsLast), "port_out_clients_valid" -> signals(toClientsValid), "port_out_clients_ready" -> signals(toClientsReady))
            ))
          case BehaviourNodeArbiterAsyncFunction(fromMasterRequest, fromMasterResponse, toClientsReqData, toClientsReqValid, toClientsReqReady, toClientsRspData, toClientsRspValid, schedulingStrategy, _) =>
            // TODO strategy
            Some(VhdlBehaviourTemplate("arbiter_async_function",
              Map("clk" -> VhdlPort.CLOCK, "reset" -> VhdlPort.RESET,
                "port_out_data" -> signals(fromMasterRequest.data), "port_out_valid" -> signals(fromMasterRequest.valid), "port_out_ready" -> signals(fromMasterRequest.ready),
                "port_in_data" -> signals(fromMasterResponse.data), "port_in_valid" -> signals(fromMasterResponse.valid),
                "port_in_clients_data" -> signals(toClientsReqData), "port_in_clients_valid" -> signals(toClientsReqValid), "port_in_clients_ready" -> signals(toClientsReqReady),
                "port_out_clients_data" -> signals(toClientsRspData), "port_out_clients_valid" -> signals(toClientsRspValid))
            ))
        }
        case _ => None
      }
    )
  }
}
