package cGen.eqsat

import core.Expr

/**
 * An applier that extracts the smallest expression from an e-class and rewrites it, then adds the expression to the
 * graph.
 */
case class AdditiveExpressionRewriter(rewrite: Expr => Expr) extends Applier {
  /**
   * Rewrites an e-class based on an expression and type substitution map.
   *
   * @param eGraph    The graph that contains the e-class to rewrite.
   * @param matchInfo The match to rewrite.
   * @param stopwatch A stopwatch that times the application.
   * @return A set of all e-classes that were _directly_ modified as a result of the application.
   */
  override def apply[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                        (rule: Rewrite,
                                                         matchInfo: Match[EClass],
                                                         stopwatch: HierarchicalStopwatch): Set[EClass] = {
    // Extract the smallest expression.
    val smallestExpr = matchInfo.where.toSmallestExpr

    // Increment all unbound variable indices.
    val rewrittenExpr = rewrite(smallestExpr)

    // Add the rewritten expression to the graph if it does not exist already.
    if (Pattern(rewrittenExpr).search(eGraph, stopwatch).isEmpty) {
      Set(eGraph.add(rewrittenExpr, AddedByRewrite(rule, eGraph.add(smallestExpr, AlreadyThere), matchInfo)))
    } else {
      Set()
    }
  }
}
