package cGen.eqsat

import cGen.eqsat.explain.{Explainer, Justification}
import core.{BuiltinExpr, Expr, ParamDef, ShirIR}

import scala.collection.mutable
import scala.collection.mutable.ListBuffer

/**
  * An e-graph implementation that uses egg's merging algorithm to process batches of merge operations on every
  * `rebuild` call.
  */
final class Egg2022EGraph(val observer: PEGObserver = PEGNullObserver,
                    val analyses: Analyses = Analyses(Seq()),
                    val applyMatchesOnce: Boolean = true,
                    val useBatchSearch: Boolean = true,
                    val useBatchApplication: Boolean = true,
                    val shouldPrintProgress: Boolean = false) extends HashConsEGraphT {
  // Based on the algorithms described in egg: Fast and extensible equality saturation
  // by Max Willsey, Chandrakana Nandi, Yisu Remy Wang, Oliver Flatt, Zachary Tatlock, Pavel Panchekha
  // Link to paper: https://dl.acm.org/doi/pdf/10.1145/3434304

  override protected val explainerImpl: Option[Explainer[HashConsENode, HashConsEClass]] = None

  /**
    * A work list of e-classes to unify.
    */
  private val classesToUnify = new ListBuffer[(HashConsEClass, HashConsEClass)]()

  private val pending = new mutable.Queue[(HashConsENode, HashConsEClass)]()
  private val analysisPending = new mutable.Queue[(HashConsENode, HashConsEClass)]()

  override def defaultSaturationEngine: SaturationEngine[HashConsENode, HashConsEClass] =
    SaturationEngine(this, observer, useBatchSearch, applyMatchesOnce, shouldPrintProgress, !useBatchApplication)

  override def performMerge(left: HashConsEClass, right: HashConsEClass, reason: Justification): Boolean = {
    assert(left.t == right.t)
    val (leftRoot, rightRoot) = (unionFind.find(left), unionFind.find(right))
    if (leftRoot == rightRoot) {
      return false
    }

    classesToUnify.append((left, right))
    true
  }

  /**
    * Actually merges two e-classes.
    * @param left The first e-class to merge.
    * @param right The second e-class to merge.
    * @return `true` if merging the e-classes changed the graph; otherwise, `false`.
    */
  private def mergeImpl(left: HashConsEClass, right: HashConsEClass): Boolean = {
    val (leftRoot, rightRoot) = (unionFind.find(left), unionFind.find(right))
    if (leftRoot == rightRoot) {
      return false
    }

    val newId = unionFind.union(leftRoot, rightRoot)
    val oldId = if (newId == leftRoot) rightRoot else leftRoot
    newId.nodeSet ++= oldId.nodeSet
    newId.parents ++= oldId.parents
    pending ++= oldId.parents

    // Update the new e-class' analysis results.
    var haveAnalysesChanged = false
    for (analysis <- analyses.analyses) {
      val newResult = analysis.join(newId.analysisResult(analysis), oldId.analysisResult(analysis))
      val oldResult = newId.analysisResult(analysis)
      if (newResult != oldResult) {
        newId.setAnalysisResult(analysis, newResult)
        haveAnalysesChanged = true
      }
    }

    if (haveAnalysesChanged) {
      analysisPending ++= newId.parents
    }

    true
  }

  private def processUnions(): Unit = {
    while (pending.nonEmpty || analysisPending.nonEmpty) {
      while (pending.nonEmpty) {
        val (eNode, eClass) = pending.dequeue()
        repairNode(eNode, eClass)
      }

      while (analysisPending.nonEmpty) {
        val (eNode, eClass) = analysisPending.dequeue()
        val classId = find(eClass)
        for (analysis <- analyses.analyses) {
          val nodeData = analysis.make(eNode)
          val classData = classId.analysisResult(analysis)
          val mergedData = analysis.join(classData, nodeData)
          if (mergedData != classData) {
            classId.setAnalysisResult(analysis, mergedData)
            analysisPending ++= classId.parents
            analysis.modify(this)(eClass, mergedData)
          }
        }
      }
    }

    assert(pending.isEmpty)
    assert(analysisPending.isEmpty)
  }

  private def rebuildClasses(): Unit = {
    var anyChanged = true
    while (anyChanged) {
      anyChanged = false
      for ((eNode, eClass) <- hashCons) {
        if (eNode.args.exists(_.isStale)) {
          repairNode(eNode, eClass)
          anyChanged = true
        }
      }
    }
  }

  /**
    * Rebuilds the e-graph's invariants, which may be temporarily broken during rule application.
    */
  override def rebuild(): Unit = {
    assertNoDanglingClasses()

    while (classesToUnify.nonEmpty) {
      // Unify all e-classes that need unifying.
      for ((left, right) <- classesToUnify) {
        mergeImpl(left, right)
      }

      // Reset the list of e-classes to unify.
      classesToUnify.clear()

      processUnions()
      rebuildClasses()
    }

    assertNoDanglingClasses()
  }

  private def repairNode(node: HashConsENode, nodeClass: HashConsEClass): Unit = {
    if (!hashCons.contains(node)) {
      return
    }

    val canonicalClass = find(nodeClass)
    val canonicalNode = canonicalize(node)
    canonicalClass.nodeSet.remove(node)
    canonicalClass.nodeSet.add(canonicalNode)

    hashCons.get(canonicalNode) match {
      case Some(otherClass) if canonicalClass != find(otherClass) =>
        mergeImpl(canonicalClass, otherClass)
        hashCons.put(canonicalNode, nodeClass)

      case _ =>
        hashCons.remove(node)
        hashCons.put(canonicalNode, nodeClass)
    }
  }
}

object Egg2022EGraph {
  /**
    * Gets an expression's arguments or expression children.
    * @param expr The expression to inspect.
    * @param stripParamDefs Strips `ParamDef` values from the results if set to `true`.
    * @return A sequence of child expressions.
    */
  def args(expr: Expr, stripParamDefs: Boolean = true): Seq[Expr] = expr match {
    case builtin: BuiltinExpr => removeParamDefs(builtin.args, stripParamDefs)
    case other => removeParamDefs(other.children collect { case e: Expr => e }, stripParamDefs)
  }

  private def removeParamDefs(values: Seq[Expr], proceed: Boolean = true) = {
    if (proceed)
      values collect { case e if !e.isInstanceOf[ParamDef] => e }
    else
      values
  }

  /**
    * Gets all parameter-valued arguments or expression children.
    * @param expr The expression to examine.
    * @return A sequence of parameter definitions.
    */
  def params(expr: Expr): Seq[ParamDef] =
    args(expr, stripParamDefs = false).collect({ case p: ParamDef => p })

  /**
    * Builds a version of an expression with updated arguments or expression children.
    * @param expr The expression to build.
    * @param newArgs A sequence of updated arguments or expression children.
    * @return An expression that is identical to `expr` save for its arguments.
    */
  def buildFromArgs(expr: Expr, newArgs: Seq[Expr]): Expr = expr match {
    case builtin: BuiltinExpr =>
      val newChildren = substituteArgs(builtin.args.toList, newArgs.toList).map(x => x.asInstanceOf[Expr])
      builtin.buildFromArgs(newChildren)
    case _ =>
      val newChildren = substituteArgs(expr.children.toList, newArgs.toList)
      expr.build(newChildren)
  }

  private def substituteArgs(oldChildren: List[ShirIR], args: List[Expr]): List[ShirIR] = oldChildren match {
    case Nil => Nil
    case (x: ParamDef) :: xs => x :: substituteArgs(xs, args) // Skip ParamDef values.
    case (_: Expr) :: xs => args.head :: substituteArgs(xs, args.tail)
    case x :: xs => x :: substituteArgs(xs, args)
  }
}
