package cGen.eqsat

import cGen.{BinaryOp, Constant}
import core.{ArithType, BuiltinExpr, ShirIR, ParamDef, TypeVarT}

import scala.collection.mutable

object ExprToCode {
  def apply(expr: ShirIR): String = {
    val initStatements = mutable.ArrayBuffer[String]()
    val paramNames = mutable.Map[Long, String]()
    val typeVarNames = mutable.Map[TypeVarT, String]()

    def shortBaseName(expr: ShirIR): String = {
      val result = expr.getClass.getSimpleName.split('$').head
      if (result.endsWith("Expr")) {
        result.substring(0, result.length - "Expr".length)
      } else {
        result
      }
    }

    def toCode: ShirIR => String = {
      case p: ParamDef if paramNames.contains(p.id) => paramNames(p.id)
      case p: ParamDef =>
        val name = s"p${paramNames.size + 1}"
        paramNames(p.id) = name
        initStatements.append(s"val $name = ParamDef(${toCode(p.t)})")
        name
      case tv: TypeVarT if typeVarNames.contains(tv) => typeVarNames(tv)
      case tv: TypeVarT =>
        val name = s"tv${typeVarNames.size + 1}"
        typeVarNames(tv) = name
        initStatements.append(s"val $name = ${shortBaseName(tv)}(${tv.children.map(toCode).mkString(", ")})")
        name

      case ArithType(ae) if ae.isEvaluable => ae.evalLong.toString

      case p: DeBruijnParamUse => s"${shortBaseName(p)}(${p.index}, ${toCode(p.t)})"
      case lam: DeBruijnLambda => s"${shortBaseName(lam)}(${toCode(lam.param.t)}, ${toCode(lam.body)})"

      case BinaryOp(x, y, op, _, _) => s"BinaryOp(${toCode(x)}, ${toCode(y)}, AlgoOp.$op)"
      case Constant(value, t) => s"Constant($value, ${toCode(t)})"

      case builtin: BuiltinExpr =>
        s"${shortBaseName(builtin)}(${(builtin.args.map(toCode) ++ builtin.types.map(toCode)).mkString(", ")})"
      case other => s"${shortBaseName(other)}(${other.children.map(toCode).mkString(", ")})"
    }

    val exprCode = toCode(expr)
    initStatements.append(exprCode)
    initStatements.mkString("\n")
  }
}
