package cGen.eqsat

import cGen.AlgoOp.AlgoOp
import cGen.{AlgoOp, ArrayType, BinaryOp, Build, CLambda, Constant, DoubleType, DoubleTypeT, ExternFunctionCall, If, Ifold, IntType, IntTypeT, NamedTupleStructType, Tuple}
import cGen.idioms.TypeArith
import cGen.primitive.{Fst, Idx, Snd}
import core.{ArithType, Expr, FunctionCall, ParamDef, ParamUse, ShirIR, Type}

import scala.annotation.tailrec
import scala.collection.mutable

object ExprToScala {
  private def nameParam(param: ParamDef): String = s"p${param.id}"

  def apply(expr: ShirIR, mustBeAtomic: Boolean = false): String = {
    def parenthesize(value: String): String = if (mustBeAtomic) s"($value)" else value

    def opToText(algoOp: AlgoOp): String = algoOp match {
      case AlgoOp.Add => "+"
      case AlgoOp.Sub => "-"
      case AlgoOp.Mul => "*"
      case AlgoOp.Div => "/"
      case AlgoOp.Mod => "%"
      case AlgoOp.Lt => "<"
    }

    expr match {
      case Build(idxf, n, _, _) =>
        s"Array.range(0, ${apply(n)}).map(${apply(idxf)})"
      case Ifold(idxf, init, n, _) =>
        s"BuiltinIfold(${apply(n, mustBeAtomic = true)}, ${apply(init, mustBeAtomic = true)}, ${apply(idxf, mustBeAtomic = true)})"
      case Tuple(a, b, _) => s"(${apply(a, mustBeAtomic = true)}, ${apply(b, mustBeAtomic = true)})"
      case Fst(a, _, _) => parenthesize(s"${apply(a, mustBeAtomic = true)}._1")
      case Snd(a, _, _) => parenthesize(s"${apply(a, mustBeAtomic = true)}._2")
      case Constant(value, _: IntTypeT) => value.toLong.toString
      case Constant(value, _: DoubleTypeT) => value.toString
      case TypeArith(value, _) => apply(value, mustBeAtomic = mustBeAtomic)
      case Idx(array, index, _, _) => s"${apply(array, mustBeAtomic = true)}(${apply(index)})"
      case ExternFunctionCall(name, args, _, _) => s"$name(stopwatch, ${args.reverse.map(apply(_)).mkString(", ")})"
      case FunctionCall(f, arg, _) => s"${apply(f, mustBeAtomic = true)}(${apply(arg, mustBeAtomic = true)})"

      case CLambda(p, body, _) => parenthesize(s"(${nameParam(p)}: ${apply(p.t)}) => ${apply(body)}")
      case ParamUse(pDef) => nameParam(pDef)

      case BinaryOp(left, right, op, _, _) =>
        parenthesize(
          s"${apply(left, mustBeAtomic = true)} ${opToText(op)} ${apply(right, mustBeAtomic = true)}")

      case If(cond, ifTrue, ifFalse, _) =>
        parenthesize(
          s"if (${apply(cond)}) { ${apply(ifTrue)} } " +
            s"else { ${apply(ifFalse)} }")

      case IntType(_, _) => "Int"
      case DoubleType() => "Double"
      case ArrayType(et, _) => s"Array[${apply(et)}]"
      case NamedTupleStructType(_, fields) =>
        val parts = fields.map(_._2).map(apply(_)).mkString(", ")
        s"($parts)"
      case ArithType(ae) => ae.evalLong.toString
    }
  }

  private def generateRandomInput(t: Type): String = t match {
    case IntType(_, _) => "Random.nextInt(4000)"
    case DoubleType() => "Random.nextDouble()"
    case ArrayType(et, len) => s"Array.fill(${this(len)})(${generateRandomInput(et)})"
  }

  def generateTest(expr: Expr, name: String): String = {
    val lines = mutable.ArrayBuffer[String]()

    @tailrec
    def lambdaToFixture(expr: Expr): Expr = expr match {
      case CLambda(p, body, _) =>
        lines.append(s"val ${nameParam(p)} = ${generateRandomInput(p.t)}")
        lambdaToFixture(body)

      case _ => expr
    }

    val inner = lambdaToFixture(expr)

    val nameInQuotes = "\"" + name + "\""
    lines.append(s"val stopwatch = HierarchicalStopwatch.start($nameInQuotes)")
    lines.append(this(inner))
    lines.append("stopwatch.complete()")
    val body = lines.mkString("\n")
    s"def $name(): MeasuredActivity = {\n$body\n}\n"
  }
}
