package cGen.eqsat

import cGen.AlgoOp.AlgoOp
import cGen.eqsat.explain.Annotation
import cGen.{AlgoOp, ArrayType, BinaryOp, BoolType, Build, CFC, Constant, DoubleType, DoubleTypeT, ExternFunctionCall, If, Ifold, IntType, IntTypeT, Map, NamedTupleStructType, Tuple}
import cGen.idioms.TypeArith
import cGen.primitive.{Fst, Id, Idx, Snd}
import core.{ArithType, ExprVar, FunctionCall, LambdaT, ParamDef, ParamUse, ShirIR}

import scala.collection.mutable

object ExprToText {
  private final class Instance(renameVars: Boolean) {
    private val paramRenaming = mutable.Map[Long, Long]()
    private val exprVarRenaming = mutable.Map[Long, Long]()

    private def name(id: Long, renamingMap: mutable.Map[Long, Long]): Long = {
      if (renameVars)
        renamingMap.getOrElseUpdate(id, renamingMap.size)
      else
        id
    }

    def apply(expr: ShirIR, mustBeAtomic: Boolean = false): String = {
      def parenthesize(value: String): String = if (mustBeAtomic) s"($value)" else value

      def opToText(algoOp: AlgoOp): String = algoOp match {
        case AlgoOp.Add => "+"
        case AlgoOp.Sub => "-"
        case AlgoOp.Mul => "*"
        case AlgoOp.Div => "/"
        case AlgoOp.Mod => "%"
        case AlgoOp.Lt => "<"
      }

      expr match {
        case Build(idxf, n, vt, _) =>
          parenthesize(s"build${vt.toShortString.head} ${apply(n, mustBeAtomic = true)} ${apply(idxf, mustBeAtomic = true)}")
        case Ifold(idxf, init, n, _) =>
          parenthesize(s"ifold ${apply(n, mustBeAtomic = true)} ${apply(init, mustBeAtomic = true)} ${apply(idxf, mustBeAtomic = true)}")
        case Map(f, array, _, _) =>
          parenthesize(s"map ${apply(f, mustBeAtomic = true)} ${apply(array, mustBeAtomic = true)}")
        case Tuple(a, b, _) =>
          parenthesize(s"tuple ${apply(a, mustBeAtomic = true)} ${apply(b, mustBeAtomic = true)}")
        case Id(value, _) => parenthesize(s"id ${apply(value, mustBeAtomic = true)}")
        case Fst(a, _, _) => parenthesize(s"fst ${apply(a, mustBeAtomic = true)}")
        case Snd(a, _, _) => parenthesize(s"snd ${apply(a, mustBeAtomic = true)}")
        case Constant(value, _: IntTypeT) => value.toLong.toString
        case Constant(value, _: DoubleTypeT) => value.toString
        case TypeArith(value, _) => apply(value, mustBeAtomic = mustBeAtomic)
        case Idx(array, index, _, _) => s"${apply(array, mustBeAtomic = true)}[${apply(index)}]"
        case ExternFunctionCall(name, args, _, _) => s"$name(${args.reverse.map(apply(_)).mkString(", ")})"
        case FunctionCall(f, arg, _) =>
          parenthesize(s"${apply(f, mustBeAtomic = true)} ${apply(arg, mustBeAtomic = true)}")
        case CFC(f, arg, _) =>
          parenthesize(s"${apply(f, mustBeAtomic = true)} ${apply(arg, mustBeAtomic = true)}")
        case lambda: DeBruijnLambda =>
          parenthesize(s"λ${apply(lambda.param.t)}. ${apply(lambda.body)}")
        case lambda: LambdaT =>
          parenthesize(s"λ${name(lambda.param.id, paramRenaming)}: ${apply(lambda.param.t)}. ${apply(lambda.body)}")
        case param: DeBruijnParamUse => s"p_${param.index}(${apply(param.t)})"
        case p@ParamDef(t, _) => parenthesize(s"param ${name(p.id, paramRenaming)}: ${apply(t, mustBeAtomic = true)}")
        case DeBruijnShift(e, _) => parenthesize(s"shift ${apply(e, mustBeAtomic = true)}")
        case ParamUse(p@ParamDef(t, _)) =>
          parenthesize(s"param ${name(p.id, paramRenaming)}: ${apply(t, mustBeAtomic = true)}")

        case exprVar@ExprVar(_, _) => parenthesize(s"expr ${name(exprVar.id, exprVarRenaming)}")

        case BinaryOp(left, right, op, _, _) =>
          parenthesize(
            s"${apply(left, mustBeAtomic = true)} ${opToText(op)} ${apply(right, mustBeAtomic = true)}")

        case If(cond, ifTrue, ifFalse, _) =>
          parenthesize(
            s"if ${apply(cond, mustBeAtomic = true)} then ${apply(ifTrue, mustBeAtomic = true)} " +
              s"else ${apply(ifFalse, mustBeAtomic = true)}")

        case IntType(width, _) => s"i${apply(width)}"
        case DoubleType() => "double"
        case BoolType() => "bool"
        case ArrayType(et, len) => s"[${apply(et)} x ${apply(len)}]"
        case NamedTupleStructType(_, fields) =>
          val parts = fields.map(_._2).map(apply(_)).mkString(", ")
          s"($parts)"
        case ArithType(ae) => ae.evalLong.toString

        case Annotation(name, args) =>
          if (args.length == 1) {
            parenthesize(s"[$name] ${ExprToText(args.head)}")
          } else {
            val argString = args.map(apply(_, mustBeAtomic = true)).mkString(" ")
            parenthesize(s"[$name] { $argString }")
          }
      }
    }
  }

  def apply(expr: ShirIR, mustBeAtomic: Boolean = false, renameVars: Boolean = true): String = {
    new Instance(renameVars)(expr, mustBeAtomic)
  }
}
