package cGen.eqsat

import core.Expr

/**
 * An e-graph expression extractor that greedily extracts expressions based on total cost in a top-down fashion.
 * Selected ancestor nodes may affect the selection eligibility of descendant nodes.
 */
final case class GreedyExtractor[C](costFunction: CostFunction[C],
                                    mayInclude: (Expr, List[ENodeT]) => Boolean)
                                   (implicit num: Numeric[C]) extends Extractor {

  override def apply(eClass: EClassT): Option[Expr] = extract(eClass, Nil)

  /**
   * Extracts an expression from an e-class.
   * @return An expression that is equivalent to the computation represented by `eClass`, if one can be found;
   *         otherwise, `None`.
   */
  private def extract(eClass: EClassT, ancestors: List[ENodeT]): Option[Expr] = {
    val costs = eClass.computeAnalysisForReachable(
      ExtractionAnalysis(
        "greedy-extractor-analysis",
        costFunction,
        mayInclude(_, ancestors)))

    val nodesSortedByCost = eClass
      .nodes
      .filter(node => mayInclude(node.expr, ancestors))
      .filter(node => node.args.forall(costs(_).isDefined))
      .sortBy(node => costFunction(node.expr, node.args.map(costs(_).get._2)))

    for (node <- nodesSortedByCost) {
      val extracted = node.args.flatMap(c => extract(c, node :: ancestors))
      if (extracted.size == node.args.size) {
        return Some(Egg2022EGraph.buildFromArgs(node.expr, extracted))
      }
    }

    None
  }
}
