package cGen.eqsat

import core.Expr

// The classes and traits in this file are inspired by egg's counterparts.
// See https://github.com/egraphs-good/egg/blob/main/src/rewrite.rs

/**
  * A named rewrite rule that consists of a searcher and an applier.
  * @param name A name for the rewrite rule.
  * @param searcher An object that finds e-classes matching the rewrite rule.
  * @param applier An object that applies the rewrite rule.
  */
final case class Rewrite(name: String, searcher: Searcher, applier: Applier) {
  /**
    * Tries to reverse this rewrite rule, producing a rewrite rule that accomplishes the opposite rewrite as this rule.
    * Returns None if either `searcher` is not also an `Applier` or `applier` is not also a `Searcher`.
    * @return A reversed rewrite rule if one can be created; otherwise, `None`.
    */
  def tryReverse: Option[Rewrite] = {
    (searcher.tryReverseIntoApplier, applier.tryReverseIntoSearcher) match {
      case (Some((searcherAsApplier, updateSearcher)), Some((applierAsSearcher, updateApplier))) =>
        val prefix = "Reverse"
        val newName = if (name.startsWith(prefix)) {
          name.substring(prefix.length)
        } else {
          prefix + name
        }
        Some(Rewrite(newName, updateSearcher(applierAsSearcher), updateApplier(searcherAsApplier)))

      case _ => None
    }
  }

  /**
    * Reverses this rewrite rule, producing a rewrite rule that accomplishes the opposite rewrite as this rule.
    * Throws an exception when either `searcher` is not also an `Applier` or `applier` is not also a `Searcher`.
    * @return A reversed rewrite rule.
    */
  def reverse: Rewrite = {
    tryReverse match {
      case Some(result) => result
      case _ => throw new Exception(s"Rewrite rule $name cannot be reversed.")
    }
  }
}

object Rewrite {
  /**
    * Creates a named rewrite rule from a pattern.
    * @param name A name to assign to the rewrite rule.
    * @param from The pattern that will be replaced.
    * @param to The pattern to replace `from` with.
    */
  def apply(name: String, from: Expr, to: Expr): Rewrite = {
    Rewrite(name, Pattern(from), Pattern(to))
  }
}

/**
  * Finds e-classes that match a pattern.
  */
trait Searcher {
  /**
    * Tries to reverse this searcher, producing an applier and a function that updates the searcher generated by
    * reversing an applier.
    *
    * @return A pair of an applier and a searcher-updating function.
    */
  def tryReverseIntoApplier: Option[(Applier, Searcher => Searcher)] = this match {
    case applier: Applier => Some(applier, identity)
    case _ => None
  }

  /**
    * Tests if an e-class matches the pattern this searcher looks for; returns all resulting matches.
    * @param eGraph The e-graph that defines the e-class.
    * @param eClass An e-class to search.
    * @param stopwatch A stopwatch that times the search.
    * @return All matches for this pattern and e-class combo.
    */
  def searchEClass[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                      (eClass: EClass,
                                                       stopwatch: HierarchicalStopwatch): Seq[Match[EClass]]

  /**
    * Finds all e-classes in an e-graph that match the pattern this searcher looks for.
    * @param eGraph An e-graph to search.
    * @param createStopwatch Creates a stopwatch to measure a search.
    * @return A sequence of matches.
    */
  def search[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass],
                                                 createStopwatch: () => HierarchicalStopwatch): Seq[Match[EClass]] = {

    eGraph
      .classes
      .flatMap(eClass => {
        val stopwatch = createStopwatch()
        val matches = searchEClass(eGraph)(eClass, stopwatch)
        stopwatch.complete()
        matches
      })
  }

  /**
    * Finds all e-classes in an e-graph that match the pattern this searcher looks for.
    * @param eGraph An e-graph to search.
    * @param stopwatch A stopwatch to use for measuring all searches.
    * @return A sequence of matches.
    */
  def search[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass],
                                                 stopwatch: HierarchicalStopwatch): Seq[Match[EClass]] =
    search(eGraph, () => stopwatch)

  /**
    * Finds all e-classes in an e-graph that match the pattern this searcher looks for.
    * @param eGraph An e-graph to search.
    * @return A sequence of matches.
    */
  def search[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass]): Seq[Match[EClass]] =
    search(eGraph, HierarchicalStopwatch.disabled)

  /**
    * Instantiates the match location. This produces an e-class that is equivalent to `found.where`, but it might be
    * more specific. For instance, if the pattern being matched by the `Searcher` is `x + 0`, then `found.where` will
    * be any e-class in which `x + 0` appears. `instantiateMatch` may produce a more specific (non-canonical) e-class
    * for the `x + 0` pattern match.
    * @param eGraph The e-graph in which the match appears. The e-graph may be changed by this method.
    * @param found A match previously produced by this `Searcher`.
    * @tparam ENode The type of an e-node.
    * @tparam EClass The class of an e-node.
    * @return A non-canonical, but possibly more specific e-class that is equivalent to `found.where`.
    */
  def instantiateMatch[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass],
                                                           found: Match[EClass]): EClass = {
    found.where
  }

  /**
    * Instantiates the match location if and only if explanations are enabled for the e-graph.
    * @param eGraph The e-graph in which the match appears. The e-graph may be changed by this method.
    * @param found A match previously produced by this `Searcher`.
    * @tparam ENode The type of an e-node.
    * @tparam EClass The class of an e-node.
    * @return A non-canonical, but possibly more specific e-class that is equivalent to `found.where`.
    */
  final def instantiateMatchIfUseful[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass],
                                                                         found: Match[EClass]): EClass = {
    if (eGraph.explainer.isDefined) {
      instantiateMatch(eGraph, found)
    } else {
      found.where
    }
  }
}

/**
  * Rewrites patterns found by `Searcher` instances.
  */
trait Applier {
  /**
    * Tries to reverse this applier, producing a searcher and a function that updates the applier generated by
    * reversing a searcher.
    * @return A pair of a searcher and an applier-updating function.
    */
  def tryReverseIntoSearcher: Option[(Searcher, Applier => Applier)] = this match {
    case searcher: Searcher => Some(searcher, identity)
    case _ => None
  }

  /**
    * Rewrites an e-class based on an expression and type substitution map.
    * @param eGraph The graph that contains the e-class to rewrite.
    * @param rule The rewrite rule that triggered this applier.
    * @param matchInfo The match to rewrite.
    * @param stopwatch A stopwatch that times the application.
    * @return A set of all e-classes that were _directly_ modified as a result of the application.
    */
  def apply[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                               (rule: Rewrite,
                                                matchInfo: Match[EClass],
                                                stopwatch: HierarchicalStopwatch): Set[EClass]

  /**
    * Rewrites multiple e-classes based on matches as produced by a `Searcher`.
    * @param eGraph The e-graph to rewrite.
    * @param rule The rewrite rule that triggered this applier.
    * @param matches A sequence of matches to rewrite.
    * @param createStopwatch Creates a stopwatch for a rewrite rule application.
    * @return A set of all e-classes that were _directly_ modified as a result of the applications.
    */
  def applyManyTimed[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                        (rule: Rewrite,
                                                         matches: Seq[Match[EClass]],
                                                         createStopwatch: () => HierarchicalStopwatch): Set[EClass] =
    matches
      .map(m => {
        val stopwatch = createStopwatch()
        val changes = apply(eGraph)(rule, m, stopwatch)
        stopwatch.complete()
        changes
      })
      .foldLeft(Set[EClass]())(_ union _)

  /**
    * Rewrites multiple e-classes based on matches as produced by a `Searcher`.
    * @param eGraph The e-graph to rewrite.
    * @param matches A sequence of matches to rewrite.
    * @return A set of all e-classes that were _directly_ modified as a result of the applications.
    */
  def applyMany[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                   (rule: Rewrite,
                                                    matches: Seq[Match[EClass]]): Set[EClass] =
    applyManyTimed(eGraph)(rule, matches, () => HierarchicalStopwatch.disabled)
}

/**
 * An applier that does nothing.
 */
object NopApplier extends Applier {
  /**
   * Rewrites an e-class based on an expression and type substitution map.
   *
   * @param eGraph    The graph that contains the e-class to rewrite.
   * @param matchInfo The match to rewrite.
   * @param stopwatch A stopwatch that times the application.
   * @return A set of all e-classes that were _directly_ modified as a result of the application.
   */
  override def apply[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                        (rule: Rewrite,
                                                         matchInfo: Match[EClass],
                                                         stopwatch: HierarchicalStopwatch): Set[EClass] = Set()
}
