package cGen.eqsat

import core.Expr

import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer
import scala.util.Random

/**
  * An expression extractor that randomly samples expressions from an e-class until some criterion is met and one of
  * the samples (or none of the samples) is returned as the extracted expression for the e-class.
  * @param createSampleSet Creates a new sample evaluation set. These sets accumulate samples until they are saturated,
  *                        at which point they select a best expression (or no expression).
  * @param r A random number generator that is used to perform random walks through e-graphs.
  */
final case class StochasticExtractor(createSampleSet: () => SampleEvaluationSet[Expr], r: Random)
  extends Extractor {

  /**
    * Extracts an expression from an e-class.
    * @param eClass The e-class to extract an expression from.
    * @return An expression that is equivalent to the computation represented by `eClass`, if one can be found;
    *         otherwise, `None`.
    */
  override def apply(eClass: EClassT): Option[Expr] = {
    val samples = createSampleSet()
    while (true) {
      sample(eClass) match {
        case Some(sampled) =>
          // Sample a fresh expression from the e-class and add it to the sample set.
          samples.add(sampled)

          // See if we have a result.
          if (samples.isSaturated) {
            return samples.best
          }

        case _ =>
          samples.onSampleFailure()
      }
    }
    ???
  }

  /**
    * Randomly samples an expression from an e-class. This expression may or may not be acceptable as expressed by
    * the `accept` predicate.
    * @param eClass An e-class to sample from.
    * @return An expression, randomly extracted from `eClass`.
    */
  def sample(eClass: EClassT, maxDepth: Int = 100): Option[Expr] = {
    if (maxDepth >= 0) {
      val nodes = eClass.nodes
      val selectedNode = nodes(r.nextInt(nodes.length))
      val argSamples = new mutable.ArrayBuffer[Expr]()
      for (arg <- selectedNode.args) {
        sample(arg, maxDepth - 1) match {
          case Some(argSample) => argSamples.append(argSample)
          case None => return None
        }
      }
      Some(selectedNode.instantiate(argSamples))
    } else {
      None
    }
  }
}

object StochasticExtractor {
  /**
    * Creates a stochastic expression extractor that samples an e-class until either `minAcceptedSamples` samples that
    * conform to predicate `accept` are found or `maxTotalSamples` total samples have been taken. At that point, the
    * best sample (according to some cost function) is chosen.
    * @param cost A function that determines the cost of an expression. The extractor's goal is to find an expression
    *             that minimizes this function.
    * @param accept Determines whether a sampled expression is eligible for inclusion in the set of samples from which
    *               the best expression is drawn.
    * @param r A random number generator that is used to perform random walks through e-graphs.
    * @param minAcceptedSamples The minimum number of samples conforming to `accept` that must be found before the
    *                           extractor feels comfortable to render judgment by selecting the lowest-cost accepted
    *                           sample.
    * @param maxTotalSamples The maximum number of samples taken before the extractor either selects the lowest-cost
    *                        accepted sample or returns nothing at all (if no samples were accepted).
    * @param num Defines numeric operations on the cost function's metric.
    * @tparam C The cost function's metric. Lower values indicate better expressions.
    * @return A stochastic expression extractor.
    */
  def apply[C](cost: CostFunction[C],
               accept: Expr => Boolean,
               r: Random = Random,
               minAcceptedSamples: Int = 1,
               maxTotalSamples: Option[Int] = None)
              (implicit num: Numeric[C]): StochasticExtractor = {

    StochasticExtractor(() => new SampleEvaluationSet[Expr] {
      var totalSampleCount = 0
      val samples = new ArrayBuffer[Expr]()

      override def add(sample: Expr): Unit = {
        if (accept(sample)) {
          samples.append(sample)
        }
        totalSampleCount += 1
      }

      override def onSampleFailure(): Unit = {
        totalSampleCount += 1
      }

      override def isSaturated: Boolean = {
        maxTotalSamples match {
          case Some(max) if totalSampleCount >= max => true
          case _ => samples.length >= minAcceptedSamples
        }
      }

      override def best: Option[Expr] = samples.map(expr => (expr, cost(expr))).sortBy(_._2).headOption.map(_._1)
    }, r)
  }
}
