package cGen.eqsat

import cGen.primitive.{Fst, Idx, Snd}
import cGen.{ArrayTypeVar, CGenDataTypeVar, EagerType, ViewTypeVar}
import core.{AnyTypeVar, ArithType, ExprVar, FunctionCall}
import lift.arithmetic.Cst

import scala.collection.mutable.ArrayBuffer

/**
 * An applier that merges e-classes corresponding to composite data types based on e-classes corresponding to element
 * data types.
 */
trait StructuralEqualityApplier extends Applier {
  /**
   * Finds the set of equivalent composites in an e-class identified by a match.
   * @param eGraph The e-graph that contains the match.
   * @param where The e-class to examine.
   * @param stopwatch A stopwatch that times the search.
   */
  def findEquivalentComposites[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                                  (where: EClass,
                                                                   stopwatch: HierarchicalStopwatch): Seq[EClass]

  /**
   * Rewrites an e-class based on an expression and type substitution map.
   *
   * @param eGraph    The graph that contains the e-class to rewrite.
   * @param matchInfo The match to rewrite.
   * @param stopwatch A stopwatch that times the application.
   * @return A set of all e-classes that were _directly_ modified as a result of the application.
   */
  override def apply[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                        (rule: Rewrite,
                                                         matchInfo: Match[EClass],
                                                         stopwatch: HierarchicalStopwatch): Set[EClass] = {

    if (matchInfo.where.nodes.length < 2) {
      return Set()
    }

    // First collect all equivalent composites.
    val equivalentComposites = findEquivalentComposites(eGraph)(matchInfo.where, stopwatch)

    // Partition equivalent composites by type, then merge the composites, if they need merging.
    val changed = ArrayBuffer[EClass]()
    for (group <- equivalentComposites.groupBy(_.t).values.filter(_.length > 2)) {
      for (composite <- group.tail) {
        val reason = AddedByRewrite(rule, matchInfo.where, matchInfo)
        if (eGraph.merge(group.head, composite, reason)) {
          changed.append(group.head)
          changed.append(composite)
        }
      }
    }

    // Return the modified e-classes.
    changed.toSet
  }
}

/**
 * An applier that finds structurally equivalent arrays (xs[x_0] = ys[x_0]) and unifies them (xs = ys).
 */
object StructuralArrayEquality extends StructuralEqualityApplier {
  private val arrayExprVar = ExprVar(ArrayTypeVar())

  /**
   * The pattern this structural function equality applier rewrites.
   */
  val pattern: Pattern = Pattern(
    Idx(DeBruijnShift(arrayExprVar), DeBruijnParamUse(0, CGenDataTypeVar()), ViewTypeVar()))

  def findEquivalentComposites[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                                  (where: EClass,
                                                                   stopwatch: HierarchicalStopwatch): Seq[EClass] = {
    val searchMatches = stopwatch.child(
      "Search",
      (s: HierarchicalStopwatch) => pattern.searchEClass(eGraph)(where, s))
    searchMatches.map(_.instantiate(arrayExprVar))
  }
}

/**
 * An applier that finds structurally equivalent functions (f x_0 = g x_0) and unifies them (f = g).
 */
object StructuralFunctionEquality extends StructuralEqualityApplier {
  private val functionExprVar = ExprVar(AnyTypeVar())

  /**
   * The pattern this structural function equality applier rewrites.
   */
  val pattern: Pattern = Pattern(FunctionCall(DeBruijnShift(functionExprVar), DeBruijnParamUse(0, AnyTypeVar())))

  def findEquivalentComposites[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                                                  (where: EClass,
                                                                   stopwatch: HierarchicalStopwatch): Seq[EClass] = {
    val searchMatches = stopwatch.child(
      "Search",
      (s: HierarchicalStopwatch) => pattern.searchEClass(eGraph)(where, s))
    searchMatches.map(_.instantiate(functionExprVar))
  }
}

/**
 * An applier that finds structurally equivalent tuples (t._1 = u._1 and t._2 = u._2) and unifies them (t = u).
 */
object StructuralTupleEquality extends Applier {
  def apply[ENode <: ENodeT, EClass <: EClassT](eGraph: EGraphT[ENode, EClass])
                                               (rule: Rewrite,
                                                matchInfo: Match[EClass],
                                                stopwatch: HierarchicalStopwatch): Set[EClass] = {

    // Gather a list of equivalent expressions of the form `t._1`.
    val fstEquivalent = ArrayBuffer[EClass]()
    for (node <- matchInfo.where.nodes) {
      node.expr match {
        case Fst(array: EClass, _, _) =>
          fstEquivalent.append(array)

        case _ =>
      }
    }

    // Exit early if we have two matches or fewer.
    if (fstEquivalent.length < 2) {
      return Set()
    }

    // Group each element `t` of `fstEquivalent` by the class membership of `t._2`. Discard singleton groups since we
    // will be merging the elements of each group. Drop the keys. We don't need them.
    val groupedBySnd = fstEquivalent
      .groupBy(t => eGraph.add(Snd(t, EagerType()), AddedByRewrite(rule, matchInfo.where, matchInfo)))
      .values
      .filter(_.nonEmpty)

    // Merge the elements of each group.
    val changed = ArrayBuffer[EClass]()
    for (group <- groupedBySnd) {
      for (elem <- group.tail) {
        val reason = AddedByRewrite(rule, matchInfo.where, matchInfo)
        if (eGraph.merge(group.head, elem, reason)) {
          changed.append(group.head)
          changed.append(elem)
        }
      }
    }

    // Return the modified e-classes.
    changed.toSet
  }
}
