package cGen.eqsat

import core.{Expr, Type}

import scala.collection.mutable

/**
  * A mutable subgraph that can be added to an e-graph as a unit.
  *
  */
class SubgraphBuilder[ENode <: ENodeT, EClass <: EClassT] {
  case class Promise(id: Int, t: Option[Type])

  private sealed trait Instruction {
    def t: Option[Type]

    def tryApply(promise: Promise,
                 graph: EGraphT[ENode, EClass],
                 reason: NodeReason,
                 mapping: mutable.HashMap[Promise, EClass]): Boolean
  }

  private case class AddClass(node: Expr, args: Seq[Promise]) extends Instruction {
    override def t: Option[Type] = Some(node.t)

    override def tryApply(promise: Promise,
                          graph: EGraphT[ENode, EClass],
                          reason: NodeReason,
                          mapping: mutable.HashMap[Promise, EClass]): Boolean = {
      val children = args.map(mapping.get).collect({ case Some(value) => value })
      if (children.length == args.length) {
        mapping(promise) = graph.add(Egg2022EGraph.buildFromArgs(node, children), reason)
        true
      } else {
        false
      }
    }
  }

  private case class AddGhostClass(classType: Type) extends Instruction {
    override def t: Option[Type] = Some(classType)

    override def tryApply(promise: Promise,
                          graph: EGraphT[ENode, EClass],
                          reason: NodeReason,
                          mapping: mutable.HashMap[Promise, EClass]): Boolean = {
      // This instruction cannot meaningfully apply itself. Merge will fold ghost classes into real classes.
      // All we do here is note whether that's happened yet.
      mapping.contains(promise)
    }
  }

  private case class ReferenceClass(eClass: EClass) extends Instruction {
    override def t: Option[Type] = Some(eClass.t)

    override def tryApply(promise: Promise,
                          graph: EGraphT[ENode, EClass],
                          reason: NodeReason,
                          mapping: mutable.HashMap[Promise, EClass]): Boolean = {
      mapping(promise) = eClass
      true
    }
  }

  private case class Merge(left: Promise, right: Promise) extends Instruction {
    override def t: Option[Type] = None

    override def tryApply(promise: Promise,
                          graph: EGraphT[ENode, EClass],
                          reason: NodeReason,
                          mapping: mutable.HashMap[Promise, EClass]): Boolean = {
      (mapping.get(left), mapping.get(right), instructions(left.id), instructions(right.id)) match {
        case (Some(newLeft), Some(newRight), _, _) =>
          for (node1 <- newLeft.nodes) {
            node1.expr match {
              case use1: DeBruijnParamUse =>
                for (node2 <- newRight.nodes) {
                  node2.expr match {
                    case use2: DeBruijnParamUse =>
                      assert(use1.index == use2.index)
                      assert(use1.t == use2.t)

                    case _ =>
                  }
                }
              case _ =>
            }
          }
          graph.merge(newLeft, newRight, reason)
          true

        case (Some(newLeft), None, _, AddGhostClass(t)) =>
          assert(t == newLeft.t)
          mapping(right) = newLeft
          true

        case (None, Some(newRight), AddGhostClass(t), _) =>
          assert(t == newRight.t)
          mapping(left) = newRight
          true

        case _ => false
      }
    }
  }

  private val instructions = new mutable.ArrayBuffer[Instruction]()

  private def createPromise(instruction: Instruction): Promise = {
    val promise = Promise(instructions.length, instruction.t)
    instructions.append(instruction)
    promise
  }

  def reference(eClass: EClass): Promise = {
    createPromise(ReferenceClass(eClass))
  }

  def addEmpty(t: Type): Promise = {
    createPromise(AddGhostClass(t))
  }

  def add(node: Expr, args: Seq[Promise]): Promise = {
    createPromise(AddClass(node, args))
  }

  def merge(left: Promise, right: Promise): Promise = {
    assert(left.t.isDefined)
    assert(right.t.isDefined)
    assert(left.t == right.t)
    createPromise(Merge(left, right))
  }

  def transferTo(graph: EGraphT[ENode, EClass], reason: NodeReason, root: Promise): EClass = {
    val mapping = mutable.HashMap[Promise, EClass]()
    var remainingInstructions = instructions.zipWithIndex
    while (remainingInstructions.nonEmpty) {
      val newRemainingInstructions = mutable.ArrayBuffer[(Instruction, Int)]()
      for ((instruction, id) <- remainingInstructions) {
        val promise = Promise(id, instruction.t)
        if (!instruction.tryApply(promise, graph, reason, mapping)) {
          newRemainingInstructions.append((instruction, id))
        }
      }
      remainingInstructions = newRemainingInstructions
    }

    mapping(root)
  }

  def clear(): Unit = {
    this.instructions.clear()
  }
}
