package cGen.eqsat

import core.{Expr, LambdaT}

import scala.collection.mutable

/**
  * A rewriter that replaces De Bruijn parameters with values.s
  * @param eGraph The e-graph to rewrite.
  * @param builder A subgraph builder that collects rewritten nodes.
  * @tparam ENode The type of an e-node.
  * @tparam EClass The type of an e-class.
  */
case class SubstitutingRewriter[ENode <: ENodeT, EClass <: EClassT](override val eGraph: EGraphT[ENode, EClass],
                                                                    override val builder: SubgraphBuilder[ENode, EClass],
                                                                    substitutionValues: List[Expr])
  extends ParameterUseRewriter[ENode, EClass, Expr] {

  /**
    * Remember shifters to avoid duplicating work.
    */
  private val shifters = new mutable.HashMap[Int, ShiftingRewriter[ENode, EClass]]()

  override def lambdaValue(lambda: LambdaT): Expr = DeBruijnParamUse(0, lambda.param.t)

  override protected def initialContext(eClass: EClass): List[(Expr, Int)] = substitutionValues.map((_, 0))

  private def shift(expr: Expr, amount: Int): builder.Promise = expr match {
    case use: DeBruijnParamUse => builder.add(DeBruijnParamUse(use.index + amount, use.t), Seq())
    case _ =>
      val shifter = shifters.getOrElse(amount, ShiftingRewriter(eGraph, builder, amount))
      shifters.update(amount, shifter)
      shifter.rewrite(eGraph.add(expr, AlreadyThere)).asInstanceOf[builder.Promise]
  }

  override protected def rewriteAndAdd(node: ENode, context: List[(Expr, Int)]): builder.Promise = node.expr match {
    case use: DeBruijnParamUse if use.index < context.length =>
      val (value, depth) = context(use.index)
      if (use.t == value.t) {
        shift(value, depth)
      } else {
        // FIXME: is this legal?
        builder.add(use, Seq())
      }

    case use: DeBruijnParamUse if use.index >= context.length =>
      builder.add(DeBruijnParamUse(use.index - substitutionValues.length, use.t), Seq())

    case _ => rewriteAndAddDefault(node, context)
  }
}
