package drivers

import cGen.eqsat.Extractor.DefaultSmallest
import cGen.{CCodeGen, DPSPass, FlatLinux64}
import cGen.eqsat.{DeBruijnLambda, EClassT, EGraphT, ENodeT, ExprToPython, ExprToScala, ExprToText, ExternCallInResultObserver, Extractor, HierarchicalStopwatch, StepObserver}
import cGen.idioms.{BlasIdiomRules, ExtractionBasedSemanticRules, PrepareForC, PrepareForPytorch, PrepareForTensorflow, PytorchIdiomRules, TensorflowIdiomRules, TimeComplexityExtractor}
import cGen.kernels.Kernel
import core.{Expr, LambdaT, Type}
import core.util.Prune

import java.time.{ZoneId, ZonedDateTime}
import java.util.{Timer, TimerTask}
import scala.language.postfixOps
import sys.process._

object OptimizeKernel {
  val kernels = Seq(
    cGen.kernels.VectorSumKernel(1000000),
    cGen.kernels.AxpyKernel(1000000),
    cGen.kernels.GemvKernel(4000),
    cGen.kernels.GesummvKernel(2800),
    cGen.kernels.GemverKernel(4000),
    cGen.kernels.GEMMKernel(2000, 2000, 2000),
    cGen.kernels.OneMMKernel(2000),
    cGen.kernels.OuterMMKernel(2000),
    cGen.kernels.DoitgenKernel(256, 256, 256),
    cGen.kernels.MvtKernel(4000),
    cGen.kernels.MemsetZeroKernel(100000),
    cGen.kernels.BicgKernel(1800, 2200),
    cGen.kernels.AtaxKernel(2200, 2200),
    cGen.kernels.Blur1DKernel(100000),
    cGen.kernels.Jacobi1DKernel(4000, 1000),
    cGen.kernels.Stencil2DKernel(1024),
    cGen.kernels.SlimTwoMMKernel(256),
    cGen.kernels.TwoMMKernel(256)
  )

  sealed trait Target
  object BlasIdiomsTarget extends Target {
    override def toString: String = "blas"
  }
  object TensorflowIdiomsTarget extends Target {
    override def toString: String = "tensorflow"
  }

  object PytorchIdiomsTarget extends Target {
    override def toString: String = "pytorch"
  }
  object NoIdiomsTarget extends Target {
    override def toString: String = "none"
  }

  val BLAS_FLAG = "--idioms=blas"
  val TENSORFLOW_FLAG = "--idioms=tensorflow"
  val PYTORCH_FLAG = "--idioms=pytorch"
  val NO_IDIOMS_FLAG = "--idioms=none"

  def main(args: Array[String]): Unit = {
    if (args.length != 4
      || !Seq(BLAS_FLAG, TENSORFLOW_FLAG, PYTORCH_FLAG, NO_IDIOMS_FLAG).contains(args(3))) {

      println("usage: RunKernel kernel-name timeout max-steps --idioms=blas|tensorflow|pytorch|none")
      sys.exit(1)
    }

    val kernelName = args.head
    val timeout = args(1).toLong
    kernels.find(k => k.name == kernelName || k.qualifiedName == kernelName) match {
      case None =>
        println(s"unknown kernel $kernelName")
        val kernelNames = kernels.map(_.qualifiedName).mkString(", ")
        println(s"known kernels: $kernelNames")

      case Some(kernel) =>
        optimizeKernel(
          kernel,
          timeout,
          args(2).toInt,
          args(3) match {
            case BLAS_FLAG => BlasIdiomsTarget
            case TENSORFLOW_FLAG => TensorflowIdiomsTarget
            case PYTORCH_FLAG => PytorchIdiomsTarget
            case NO_IDIOMS_FLAG => NoIdiomsTarget
          })
    }
  }

  /**
    * Runs a specific kernel, logging its status throughout the running process.
    * @param kernel The kernel to run.
    * @param timeout The amount of time the kernel has to run, in milliseconds.
    */
  def optimizeKernel(kernel: Kernel, timeout: Long, maxSteps: Int, target: Target): Unit = {
    println(s"running kernel: ${kernel.qualifiedName}")

    val time = ZonedDateTime.now(ZoneId.of("UTC"))
    println(s"date/time: $time")
    println(s"commit id: ${"git rev-parse HEAD" !!}")
    println(s"machine name: $machineName")
    println(s"optimizer scala version: ${util.Properties.versionString}")
    println(s"timeout: $timeout")
    println(s"idiom recognition enabled: ${target != NoIdiomsTarget}")
    println(s"target: $target")

    val loweredKernel = kernel.transformedInput
    println(s"original kernel: ${ExprToText(loweredKernel)}")

    val (paramTypes, returnType) = paramTypesAndReturnType(loweredKernel)
    println(s"parameter types: <<<")
    paramTypes.map(ExprToText(_)).foreach(println)
    println(">>>")
    println(s"return type: ${ExprToText(returnType)}")

    Prune.findUnsolvedTypeVar(loweredKernel).foreach(smallest =>
      println(s"warning: found unsolved type var in ${ExprToText(smallest)}"))

    val extractor = TimeComplexityExtractor
    val idiomExtractor = DefaultSmallest
    val rules = ExtractionBasedSemanticRules(extractor).rules
    val idioms = target match {
      case BlasIdiomsTarget => BlasIdiomRules(idiomExtractor).rules
      case NoIdiomsTarget => Seq()
      case TensorflowIdiomsTarget => TensorflowIdiomRules(idiomExtractor).rules
      case PytorchIdiomsTarget => PytorchIdiomRules(idiomExtractor).rules
    }
    val instance = kernel.instantiate(extractor.analysis, rules, idioms)

    val observer = new ProgressPrinter(None, kernel.extractor, kernel, target)
    val (graph, root) = instance.build(observer)
    observer.root = Some(root)

    val timer = new Timer(true)
    timer.schedule(new TimerTask {
      override def run(): Unit = {
        System.exit(0)
      }
    }, timeout)

    instance.saturate(graph, maxSteps)
  }

  private def paramTypesAndReturnType(expr: Expr): (Seq[Type], Type) = {
    expr match {
      case lam: LambdaT =>
        val (innerParams, innerReturn) = paramTypesAndReturnType(lam.body)
        (lam.param.t +: innerParams, innerReturn)
      case _ => (Seq.empty, expr.t)
    }
  }

  private def machineName: String = {
    val env = System.getenv
    if (env.containsKey("COMPUTERNAME")) env.get("COMPUTERNAME")
    else if (env.containsKey("HOSTNAME")) env.get("HOSTNAME")
    else "hostname" !!
  }

  private final class ProgressPrinter(var root: Option[EClassT],
                                      extractor: Extractor,
                                      kernel: Kernel,
                                      target: Target) extends StepObserver {
    private val stopwatch = HierarchicalStopwatch.start("")

    private def printStatistics[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass]): Unit = {
      // HACK: recompute extraction analysis
      graph.recomputeAnalysis(kernel.extractionAnalysis)

      val bestSoFar = extractor(root.get).get
      val named = DeBruijnLambda.toNamedRec(bestSoFar).asInstanceOf[Expr]
      val externCalls = ExternCallInResultObserver.countExternCalls(bestSoFar)
        .map(t => s"${t._2} x ${t._1}")
        .mkString(", ")
      println(s"node count: ${graph.classes.map(_.nodes.length).sum}")
      println(s"class count: ${graph.classes.length}")
      println(s"elapsed nanoseconds: ${stopwatch.elapsedNanoseconds}")
      println(s"extern calls: $externCalls")
      println(s"solution: ${ExprToText(bestSoFar)}")
      println("scala solution: <<<")
      println(ExprToScala.generateTest(named, kernel.name))
      println(">>>")

      target match {
        case TensorflowIdiomsTarget | PytorchIdiomsTarget =>
          generatePython(named)

        case BlasIdiomsTarget =>
          generateC(named)

        case NoIdiomsTarget =>
          generateC(named)
          generatePython(named)
      }
    }

    private def generateC(expr: Expr): Unit = {
      val prepared = PrepareForC(expr)
      println(s"C codegen input: ${ExprToText(prepared)}")
      println("C solution: <<<")
      println(CCodeGen(DPSPass(prepared), platform = FlatLinux64).ccode())
      println(">>>")
    }

    private def generatePython(expr: Expr): Unit = {
      val prepared = target match {
        case TensorflowIdiomsTarget => PrepareForTensorflow(expr)
        case PytorchIdiomsTarget => PrepareForPytorch(expr)
        case NoIdiomsTarget => expr
        case _ => ???
      }
      println(s"python codegen input: ${ExprToText(prepared)}")
      println("python solution: <<<")
      println(target match {
        case TensorflowIdiomsTarget => ExprToPython.toTensorflowProgram(prepared)
        case PytorchIdiomsTarget => ExprToPython.toPytorchProgram(prepared)
        case NoIdiomsTarget => ExprToPython.toPythonProgram(prepared)
        case _ => ???
      })
      println(">>>")
    }

    override def onStepOrFinish[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass]): Unit = {
      println("===   step   ===")
      printStatistics(graph)
      println("=== step end ===")
    }
  }
}
