package backend.hdl

import algo._
import algo.util.Matrix
import backend.hdl.arch.mem.MemoryImage
import core._
import org.junit.Test

class FullCompilerTest {

  @Test
  def testGenerateConstantStream(): Unit = {
    Counter.resetAll()
    val tree = Repeat(ConstantInteger(60000, Some(algo.IntType(16))), ArithType(10))

    HDLProject("constantstream", tree).compileHDL
  }

  @Test
  def testGenerateCounterStream(): Unit = {
    Counter.resetAll()
    val tree = Repeat(CounterInteger(2, 2, 10), ArithType(10))

    HDLProject("counterstream", tree).compileHDL
  }

  @Test
  def testReshapeStream(): Unit = {
    Counter.resetAll()
    val tree = Sink(Split(Join(Repeat(Repeat(CounterInteger(2, 2, 10), 3), 2)), 2))

    HDLProject("reshapestream", tree).compileHDL
  }

  @Test
  def testMapEmptyLambda(): Unit = {
    Counter.resetAll()
    val tree =
      Map(
        {
          val p = ParamDef()
          AlgoLambda(p, ParamUse(p))
        },
        CounterInteger(0, 1, 10)
      )

    HDLProject("mapemptylambda", tree).compileHDL
  }

  @Test
  def testDotProduct(): Unit = {
    Counter.resetAll()
    val tree = DotProduct(CounterInteger(1, 0, 8), CounterInteger(4, 0, 8))

    HDLProject("dotproduct", tree).compileHDL //.simulateStandalone
  }

  @Test
  def testMemInput(): Unit = {
    Counter.resetAll()
    val tree = Input("input", algo.IntType(512), 4)
    HDLProject("meminput", tree).simulateWithHostRam(
      Predef.Map(
        "input" -> Seq(1, 2, 3, 4),
        "result" -> Seq(1, 2, 3, 4),
      )
    )
  }

  @Test
  def testMemInputPrecisions(): Unit = {
    Counter.resetAll()
    val tree1 = Input("input", 2, 4, algo.IntType(3)) // row smaller than cacheline
    HDLProject("meminput", tree1).compileHDL
    Counter.resetAll()
    val tree2 = Input("input", 3, 1, algo.IntType(200)) // row bigger than cacheline
    HDLProject("meminput", tree2).compileHDL
    Counter.resetAll()
    val tree3 = Input("input", 2, 2, algo.IntType(500)) // element almost as big as cacheline
    HDLProject("meminput", tree3).compileHDL
    Counter.resetAll()
    val tree4 = Input("input", 2, 2, algo.IntType(800)) // element bigger than cacheline
    HDLProject("meminput", tree4).compileHDL
    Counter.resetAll()
    val tree5 = Input("input", 2, 2, algo.IntType(1200)) // element bigger two cachelines
    HDLProject("meminput", tree5).compileHDL
  }

  @Test
  def testDotProductMemInput(): Unit = {
    Counter.resetAll()
    val tree = DotProduct(Input("varA", algo.IntType(16), 10), Input("varB", algo.IntType(16), 10))
    HDLProject("dotproduct", tree).compileHDL
  }

  @Test
  def testSumRow1(): Unit = {
    Counter.resetAll()
    val matrix = Matrix.fromRandom(4, 4, 65536)
    val tree = Map(
      Reduce.asFunction(Seq(Some(Add2.asFunction()), Some(ConstantInteger(0)), None)),
      Input("matrix", 4, 4, algo.IntType(16))
    )
    HDLProject("rowsum", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.sumRows().data.head)
    ).assertCorrect.print()
  }

  @Test
  def testSumRow2(): Unit = {
    Counter.resetAll()
    val matrix = Matrix.fromRandom(128, 128, 65536)
    val tree = Map(
      Reduce.asFunction(Seq(Some(Add2.asFunction()), Some(ConstantInteger(0)), None)),
      Input("matrix", 128, 128, algo.IntType(16))
    )
    HDLProject("rowsum", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.sumRows().data.head)
    ).assertCorrect.print()
  }

  @Test
  def testSumRow3(): Unit = {
    Counter.resetAll()
    val matrix = Matrix.fromRandom(64, 64, 65536)
    val tree = Map(
      Reduce.asFunction(Seq(Some(Add2.asFunction()), Some(ConstantInteger(0)), None)),
      Input("matrix", 64, 64, algo.IntType(16))
    )
    HDLProject("rowsum", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.sumRows().data.head)
    ).assertCorrect.print()
  }

  @Test
  def testIncMatrix1(): Unit = {
    val matrix = Matrix.fromRandom(64, 64, 64)
    Counter.resetAll()
    val tree = Map(
      Map.asFunction(Seq(Some(Inc.asFunction()), None)),
      Input("matrix", 64, 64, algo.IntType(8))
    )
    HDLProject("inc", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.inc().data)
    ).assertCorrect.print()
  }

  @Test
  def testIncMatrix2(): Unit = {
    val matrix = Matrix.fromRandom(128, 128,256)
    Counter.resetAll()
    val tree = Map(
      Map.asFunction(Seq(Some(Inc.asFunction()), None)),
      Input("matrix", 128, 128, algo.IntType(16))
    )
    HDLProject("inc", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.inc().data)
    ).assertCorrect.print()
  }

  @Test
  def testIncMatrix3(): Unit = {
    val matrix = Matrix.fromRandom(1, 128,256)
    Counter.resetAll()
    val tree = Map(
      Map.asFunction(Seq(Some(Inc.asFunction()), None)),
      Input("matrix", 128, 1, algo.IntType(16))
    )
    HDLProject("inc", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.inc().data)
    ).assertCorrect.print()
  }

  @Test
  def testIncMatrix4(): Unit = {
    val matrix = Matrix.fromRandom(2, 128,256)
    Counter.resetAll()
    val tree = Map(
      Map.asFunction(Seq(Some(Inc.asFunction()), None)),
      Input("matrix", 128, 2, algo.IntType(16))
    )
    HDLProject("inc", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.inc().data)
    ).assertCorrect.print()
  }

  @Test
  def testIncMatrix5(): Unit = {
    val matrix = Matrix.fromRandom(1, 16*128,256)
    Counter.resetAll()
    val tree = Map(
      Map.asFunction(Seq(Some(Inc.asFunction()), None)),
      Input("matrix", 16*128, 1, algo.IntType(16))
    )
    HDLProject("inc", tree).simulateWithHostRam(
      Predef.Map("matrix" -> matrix.data, MemoryImage.RESULT_VAR_NAME -> matrix.inc().data)
    ).assertCorrect.print()
  }

  @Test
  def testMV(): Unit = {
    Counter.resetAll()
    val tree = MVMul(CounterInteger(1, 1, Seq(2, 2), Some(algo.IntType(8))), CounterInteger(1, 1, Seq(2), Some(algo.IntType(8))))
    HDLProject("mv", tree).compileHDL //.simulateStandalone
  }

  @Test
  def testMVMemInput(): Unit = {
    Counter.resetAll()
    val tree = MVMul(Input("matrix", algo.IntType(8), 2, 2), Input("vector", algo.IntType(16), 2))
    HDLProject("mv", tree).compileHDL
  }

  @Test
  def testMM(): Unit = {
    Counter.resetAll()
    val tree = MMMul(CounterInteger(1, 1, Seq(2, 2), None), CounterInteger(11, 1, Seq(2, 2), None))
    HDLProject("mm", tree).compileHDL //simulateStandalone
  }

  @Test
  def testJoinOrderedStream(): Unit = {
    Counter.resetAll()
    val tree = Join(Repeat(Repeat(CounterInteger(2, 2, 10), 10), 4))
    HDLProject("join", tree).compileHDL
  }

  @Test
  def testSlide(): Unit = {
    Counter.resetAll()
    val tree = Slide(CounterInteger(0, 1, 10), 3)

    HDLProject("slide", tree).compileHDL
  }

  @Test
  def testBuffer(): Unit = {
    Counter.resetAll()
    val tree = Buffer(Input("input", algo.IntType(3), 4))
    HDLProject("buffer", tree).simulateWithHostRam(
      Predef.Map(
        "input" -> Seq(1, 2, 3, 4),
        "result" -> Seq(1, 2, 3, 4),
      )
    ).assertCorrect.print()
  }

  @Test
  def testFold2D(): Unit = {
    val input = Split(CounterInteger(0, 1, 64), 8)
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val tree =
      Fold(
        {
          val p1 = ParamDef()
          val p2 = ParamDef()
          AlgoLambda(Seq(p1, p2),
            algo.Map(
              Add.asFunction(),
              Zip(Tuple(ParamUse(p1), ParamUse(p2)))
            )
          )
        },
        inputtc
      )

    HDLProject("fold2D", tree).simulateStandalone()
  }

  @Test
  def testInputClash(): Unit = {
    val tree = Map(Zip.asFunction(), Zip(Tuple(
        Input("matrix", 64, 64, algo.IntType(8)),
        Input("matrix", 64, 64, algo.IntType(8))
    )))

    HDLProject("inputs", tree).simulateStandalone()
  }

  @Test
  def testIfelse(): Unit = {
    val input = CounterInteger(0, 1, 8)
    val cond = Join(Repeat(CounterInteger(0, 1, 2), 4))
    val inTuple = Zip(Tuple(input, cond))
    val f1 = {
      val p = ParamDef(input.t.asInstanceOf[SeqTypeT].et)
      AlgoLambda(p, Add(Tuple(ParamUse(p), ConstantInteger(1, Some(algo.IntType(3))))))
    }
    val f2 = {
      val p = ParamDef(input.t.asInstanceOf[SeqTypeT].et)
      AlgoLambda(p, Add(Tuple(ParamUse(p), ConstantInteger(2, Some(algo.IntType(3))))))
    }
    val tree = algo.Map(
      {
        val p = ParamDef()
        AlgoLambda(
          p,
          Ifelse(f1, f2, ParamUse(p))
        )
      }, inTuple
    )

    HDLProject("inputs", tree).simulateStandalone()
  }

  def testMax(): Unit = {
    Counter.resetAll()
    val tree = Max(Tuple(ConstantInteger(1, Some(algo.IntType(8))), ConstantInteger(3, Some(algo.IntType(8)))))
    HDLProject("max", tree).compileHDL //simulateStandalone
  }

  @Test
  def testFoldMax(): Unit = {
    val input = CounterInteger(0, 1, 8)
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val tree =
      Fold(
        Max2.asFunction(),
        inputtc
      )

    HDLProject("foldMax", tree).simulateStandalone()
  }

  @Test
  def testFoldMax2D(): Unit = {
    val input = Split(CounterInteger(0, 1, 64), 8)
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val tree =
      Fold(
        {
          val p1 = ParamDef()
          val p2 = ParamDef()
          AlgoLambda(Seq(p1, p2),
            algo.Map(
              Max.asFunction(),
              Zip(Tuple(ParamUse(p1), ParamUse(p2)))
            )
          )
        },
        inputtc
      )

    HDLProject("foldMax2D", tree).simulateStandalone()
  }

  def testFold2DSigned(): Unit = {
    val input = Split(algo.Map(Signed.asFunction(), CounterInteger(0, 1, 64)), 8)
    //val input = Split(CounterInteger(0, 1, 64), 8)
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val tree =
      Fold(
        {
          val p1 = ParamDef()
          val p2 = ParamDef()
          AlgoLambda(Seq(p1, p2),
            algo.Map(
              Add.asFunction(),
              Zip(Tuple(ParamUse(p1), ParamUse(p2)))
            )
          )
        },
        inputtc
      )

    HDLProject("fold2DSigned", tree).simulateStandalone()
  }

  @Test
  def testSignedSum(): Unit = {
    Counter.resetAll()

    val tree = TypeChecker.check(Fold(Add2.asFunction(), TypeChecker.check(algo.Map(Signed.asFunction(), CounterInteger(0, 1, 8)))))

    HDLProject("signedSum", tree).simulateStandalone()
  }

  @Test
  def testSigned(): Unit = {
    Counter.resetAll()
    val tree = Input("input", 4, 1, algo.SignedIntType(4))
    HDLProject("SignedInput", tree).simulateWithHostRam(
      Predef.Map(
        "input" -> Seq(Seq(-1, -2, 3, 4)),
        "result" -> Seq(Seq(-1, -2, 3, 4)),
      )
    ).assertCorrect.print()
  }

  @Test
  def testReLU(): Unit = {
    Counter.resetAll()
    val tree = ReLU(ConstantInteger(10))
    HDLProject("max", tree).compileHDL //simulateStandalone()
  }
}
