package backend.hdl.arch

import algo.{Concat, DotProduct}
import backend.hdl._
import backend.hdl.arch.mem.{BankedBufferRWStream, BufferStreamInBlockRam, MemFunctionsCompiler}
import backend.hdl.arch.rewrite.transpose.TransposeConversionRules
import backend.hdl.arch.rewrite.{FixTimingRules, MoveUpJoinStreamRules, MoveUpVectorToStreamRules, RemoveSelectRules}
import core._
import core.compile.CompilerPhase
import core.rewrite.{RewriteAll, RewriteStep}
import org.junit.Test

class CompileTest {

  @Test
  def testId(): Unit = {
    Counter.resetAll()
    val tree = Sink(Id(ConstantValue(4)))

    HDLProject("idconst", tree).compileHDL
  }

  @Test
  def testMapZip(): Unit = {
    Counter.resetAll()
    def zip(input1: Expr): LambdaT = {
      val stvp = ParamDef()
      ArchLambda(stvp, ZipOrderedStream(Tuple(input1, ParamUse(stvp))))
    }
    val tree = Sink(
      MapOrderedStream(
        zip(CounterInteger(0, 1, Seq(3, 4), Some(IntType(8)))),
        CounterInteger(0, 1, Seq(3, 4, 5), Some(IntType(8)))
      )
    )

    HDLProject("mapzip", tree).compileHDL
  }

  @Test
  def reduce(): Unit = {
    Counter.resetAll()
    val initValue = ConstantValue(0, Some(IntType(8)))
    val input = CounterInteger(0, 1, Seq(3), Some(IntType(8)))

    val tree = ReduceOrderedStream(
      {
        val accParam = ParamDef()
        val addParam = ParamDef()
        ArchLambdas(
          Seq(accParam, addParam),
          AddInt(Tuple(
            ParamUse(accParam),
            ParamUse(addParam)
          ))
        )
      },
      initValue,
      input
    )

    HDLProject("reduce", tree).compileHDL
  }

  @Test
  def testUnorderedStream(): Unit = {
    Counter.resetAll()
    val tree = OrderedStreamToUnorderedStream(
      CounterInteger(0, 1, Seq(3, 4), Some(IntType(8)))
    )

    HDLProject("uos", tree).compileHDL
  }

  @Test
  def testMapVector(): Unit = {
    Counter.resetAll()
    val incF = {
      val in = ParamDef()
      ArchLambda(
        in,
        AddInt(Tuple(ParamUse(in), ConstantValue(1)))
      )
    }
    val tree = MapVector(incF, OrderedStreamToVector(CounterInteger(0, 1, Seq(4), Some(IntType(8)))))

    HDLProject("mapvec", tree).compileHDL //simulateStandalone
  }

  @Test
  def testFoldVector(): Unit = {
    Counter.resetAll()
    val tree = FoldVector(AddInt.asTypeFunction(), OrderedStreamToVector(CounterInteger(0, 1, Seq(32), Some(IntType(32)))))

    HDLProject("foldvector", tree).compileHDL //simulateStandalone
  }

  @Test
  def testFoldVectorNonPow2(): Unit = {
    Counter.resetAll()
    val tree = FoldVector(AddInt.asTypeFunction(), OrderedStreamToVector(CounterInteger(0, 1, Seq(25), Some(IntType(32)))))

    HDLProject("foldvectorNonPow2", tree, CompilerPhase.first(), Seq(
      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteAll(), FixTimingRules.get())),
    )).simulateStandalone
  }

  @Test
  def testIncreaseIntegerBitWidth(): Unit = {
    Counter.resetAll()
    val tree = TruncInteger(ConstantValue(9, Some(IntType(10))), 15)

    HDLProject("IncInt", tree, CompilerPhase.first(), Seq(
      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteAll(), FixTimingRules.get())),
    )).compileHDL //simulateStandalone
  }

  @Test
  def testPermuteVector(): Unit = {
    Counter.resetAll()
    val inputLen = 9
    val input = CounterInteger(1, 1, 1, Seq(inputLen), None)
    val v = ArithTypeVar()

    val test = PermuteVector(OrderedStreamToVector(input), ArithLambdaType(v, (v+1)%9))
    HDLProject("permute", test).simulateStandalone()
  }

  @Test
  def testSlideVector(): Unit = {
    Counter.resetAll()
    val inputLen = 9
    val input = CounterInteger(1, 1, 1, Seq(inputLen), None)
    val test = SlideVector(OrderedStreamToVector(input), 3, 2)

    HDLProject("permute", test).simulateStandalone()
  }

  @Test
  def testMapRegisterBuffer(): Unit = {
    Counter.resetAll()
    val vectorBuffer: Expr = {
      val stvp = ParamDef()
      ArchLambda(stvp, VectorToOrderedStream(OrderedStreamToVector(ParamUse(stvp))))
    }
    val tree = Sink(
      MapOrderedStream(
        vectorBuffer,
        CounterInteger(0, 1, Seq(8, 4))
      )
    )

    HDLProject("mapbuffer", tree).compileHDL
  }

  @Test
  def testSharedFunction(): Unit = {
    Counter.resetAll()

    val idParam = ParamDef(IntType(10))
    val func = ArchLambda(idParam, Id(ParamUse(idParam)))
    val funcParam = ParamDef(func.t)
    val tree = Sink(
      Let(
        funcParam,
        FunctionCall(
          ParamUse(funcParam),
          FunctionCall(
            ParamUse(funcParam),
            ConstantValue(3, Some(IntType(10)))
          )
        ),
        func
      )
    )

    HDLProject("sharedfunc", tree).compileHDL //simulateStandalone
  }

  @Test
  def testAlternate(): Unit = {
    Counter.resetAll()
    val f1 = {
      val p = ParamDef()
      ArchLambda(p, AddInt(Tuple(ParamUse(p), ConstantValue(1))))
    }
    val f2 = {
      val p = ParamDef()
      ArchLambda(p, AddInt(Tuple(ParamUse(p), ConstantValue(2))))
    }
    val tree = Alternate(f1, f2, ConstantValue(10))

    HDLProject("alternate", tree).compileHDL //simulateStandalone
  }

  @Test
  def testBRam(): Unit = {
    Counter.resetAll()
    val tree = {
      TestBench(
        Sink(
          BufferStreamInBlockRam(
//            CounterInteger(4, 1, Seq(4))
            CounterInteger(0, 1, Seq(8, 4))
          )
        )
      )
    }

    HDLProject("bram", tree).compileHDL
  }

  @Test
  def testMapVectorBuffer(): Unit = {
    Counter.resetAll()
    val vectorBuffer: Expr = {
      val stvp = ParamDef()
      ArchLambda(stvp, UnorderedStreamToOrderedStream(OrderedStreamToUnorderedStream(ParamUse(stvp))))
    }
    val tree = Sink(
      MapOrderedStream(
        vectorBuffer,
        CounterInteger(0, 1, Seq(8, 4))
      )
    )

    HDLProject("mapbuffer", tree).compileHDL
  }

  @Test
  def testUntuple(): Unit = {
    val input = algo.Tuple(algo.CounterInteger(0, 1, 4), algo.CounterInteger(0, 1, 4))
    val param = ParamDef()
    val tree = Let(
      param,
      Concat(algo.Select(ParamUse(param), 0), algo.Select(ParamUse(param), 1)),
      input
    )
    HDLProject("untuple", tree).writeHDLFiles()
  }

  @Test
  def testFold2DStream(): Unit = {
    val tree =
      FoldOrderedStream(
        {
          val p1 = ParamDef()
          val p2 = ParamDef()
          ArchLambdas(Seq(p1, p2),
            MapSimpleOrderedStream(
              AddInt.asFunction(),
              ZipOrderedStream(Tuple(ParamUse(p1), ParamUse(p2)))
            )
          )
        },
        CounterInteger(0, 1, Seq(8, 8))
      )

    HDLProject("fold2D", tree).simulateStandalone()
  }

  @Test
  def testFold3DStream(): Unit = {
    val input = CounterInteger(0, 1, 1, Seq(8, 8, 8), None)
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val tree = MapSimpleOrderedStream(
      {
        val param = ParamDef(inputtc.t.asInstanceOf[OrderedStreamTypeT].et)
        ArchLambda(
          param,
          FoldOrderedStream(
            {
              val p1 = ParamDef()
              val p2 = ParamDef()
              ArchLambdas(Seq(p1, p2),
                MapSimpleOrderedStream(
                  AddInt.asFunction(),
                  ZipOrderedStream(Tuple(ParamUse(p1), ParamUse(p2)))
                )
              )
            },
            ParamUse(param)
          )
        )
      }, inputtc
    )
    HDLProject("fold3D", tree).simulateStandalone()
  }

  @Test
  def testConcatStreamDimsInnerMost(): Unit = {
    val input1 = CounterInteger(0, 1, 1, Seq(2, 8, 2), Some(IntType(8)))
    val input2 = CounterInteger(0, 1, 1, Seq(4, 8, 2), Some(IntType(8)))
    val tree = ConcatOrderedStream(input1, input2, 2)
    HDLProject("concatstm_innermost", tree).simulateStandalone()
  }

  @Test
  def testConcatStreamDimsMid(): Unit = {
    val input1 = CounterInteger(0, 1, 1, Seq(2, 8, 2), Some(IntType(8)))
    val input2 = CounterInteger(0, 1, 1, Seq(2, 8, 2), Some(IntType(8)))
    val tree = ConcatOrderedStream(input1, input2, 1)
    HDLProject("concatstm_mid", tree).simulateStandalone()
  }

  @Test
  def testConcatStreamDimsOuterMost(): Unit = {
    val input1 = CounterInteger(0, 1, 1, Seq(2, 8, 2), Some(IntType(8)))
    val input2 = CounterInteger(0, 1, 1, Seq(2, 8, 2), Some(IntType(8)))
    val tree = ConcatOrderedStream(input1, input2, 0)
    HDLProject("concatstm_outermost", tree).simulateStandalone()
  }

  @Test
  def testRotateVector(): Unit = {
    val input = CounterInteger(0, 1, 1, Seq(8), Some(IntType(8)))
    val tree = RotateVector(Tuple(OrderedStreamToVector(input), ConstantValue(2)), 0)
    HDLProject("rotate", tree).simulateStandalone()
  }

  @Test
  def testBankedRamWriteAllBanks(): Unit = {
    val numBanks = 4
    val input = CounterInteger(0, 1, 1, Seq(8,8), Some(IntType(8)))
    val inputVec = MapOrderedStream(OrderedStreamToVector.asFunction(), input)
    val wAddr = MapOrderedStream(VectorGenerator.asFunction(Seq(None), Seq(numBanks)), CounterInteger(0, 1, 1, Seq(8), None))
    val wEn = Repeat(VectorGenerator(ConstantValue(1, Some(LogicType())), numBanks), 8)
    val rAddr = MapOrderedStream(VectorGenerator.asFunction(Seq(None), Seq(numBanks)), CounterInteger(0, 1, 1, Seq(8), None))
    val tree = BankedBufferRWStream(inputVec, wAddr, wEn, rAddr, numBanks, 8)
    HDLProject("bankedRam", tree).simulateStandalone()
  }

  @Test
  def testBankedRamWriteSingleBank(): Unit = {
    val numBanks = 4
    val numOfAddresses = 2
    val totalElements = numBanks * numOfAddresses
    val input = CounterInteger(0, 1, 1, Seq(totalElements), Some(IntType(8)))
    val inputVec = MapOrderedStream(VectorGenerator.asFunction(Seq(None), Seq(numBanks)), input)
    val wAddr = MapOrderedStream(VectorGenerator.asFunction(Seq(None), Seq(numBanks)),
      JoinOrderedStream(MapOrderedStream(Repeat.asFunction(Seq(None), Seq(numBanks)), CounterInteger(0, 1, 1, Seq(numOfAddresses), None))))

    // Calculate write enable signals. Only one bank is accessed at a time
    val enableBase = ConcatVector(Tuple(VectorGenerator(ConstantValue(1, Some(LogicType())), 1), VectorGenerator(ConstantValue(0, Some(LogicType())), numBanks - 1)))
    val enableBaseRepeat = Repeat(enableBase, totalElements)
    val enableBaseRotateDistance = JoinOrderedStream(Repeat(CounterInteger(0, 1, 1, Seq(numBanks), None), numOfAddresses))
    val wEn = MapOrderedStream(RotateVector.asFunction(Seq(None), Seq(1)), ZipOrderedStream(Tuple(enableBaseRepeat, enableBaseRotateDistance)))

    val rAddr = MapOrderedStream(VectorGenerator.asFunction(Seq(None), Seq(numBanks)), CounterInteger(0, 1, 1, Seq(numOfAddresses), None))
    val tree = BankedBufferRWStream(inputVec, wAddr, wEn, rAddr, numBanks)
    HDLProject("bankedRam", tree).simulateStandalone()
  }

  @Test
  def testTest(): Unit = {
    val input1 = CounterInteger(0, 1, Seq(8,8), Some(IntType(8)))
    val input2 = CounterInteger(0, 1, Seq(8,8), Some(IntType(8)))
    val input3 = CounterInteger(0, 1, Seq(8), Some(IntType(8)))
    val tupInput = ZipOrderedStream(Tuple(ZipOrderedStream(Tuple(input1, input2)), input3))
    val tupInputtc = TypeChecker.checkIfUnknown(tupInput, tupInput.t)
    val tree = TypeChecker.check(MapOrderedStream(
      {
        val param = ParamDef(tupInputtc.t.asInstanceOf[OrderedStreamTypeT].et)
        ArchLambda(
          param,
          Tuple(Select(Select(ParamUse(param), 0), 0), Repeat(Select(ParamUse(param), 1), 8))
        )
      }, tupInputtc
    ))

    val treetc = RewriteStep(RewriteAll(), RemoveSelectRules.get()).apply(tree)
    assert(!treetc.hasUnknownType)
    assert(tree.t == treetc.t)
  }

  @Test
  def testIdStmFunc(): Unit = {
    Counter.resetAll()

    val input = CounterInteger(1, 1, Seq(8), Some(IntType(8)))
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val bufParam = ParamDef(inputtc.t)
    val func = ArchLambda(bufParam, Id(ParamUse(bufParam)))
    val functc = TypeChecker.check(func)
    val funcParam = ParamDef(functc.t)

    val tree = Sink(
      Let(
        funcParam,
        FunctionCall(
          ParamUse(funcParam),
          // The buffer is required to store the whole stream or the arbiter will get stuck.
          BufferStreamInBlockRam(FunctionCall(
            ParamUse(funcParam),
            inputtc
          ))
        ),
        functc
      )
    )

    HDLProject("IdStmFunc", tree).simulateStandalone()
  }

  @Test
  def testIdStm2DFunc(): Unit = {
    Counter.resetAll()

    val input = CounterInteger(1, 1, Seq(5,5), Some(IntType(8)))
    val inputtc = TypeChecker.checkIfUnknown(input, input.t)
    val bufParam = ParamDef(inputtc.t)
    val func = ArchLambda(bufParam, Id(ParamUse(bufParam)))
    val functc = TypeChecker.check(func)
    val funcParam = ParamDef(functc.t)

    val tree = Sink(
      Let(
        funcParam,
        FunctionCall(
          ParamUse(funcParam),
          // The buffer is required to store the whole stream or the arbiter will get stuck.
          SplitOrderedStream(BufferStreamInBlockRam(JoinStream(FunctionCall(
            ParamUse(funcParam),
            inputtc
          ))), 5)
        ),
        functc
      )
    )

    HDLProject("IdStm2DFunc", tree).simulateStandalone()
  }

  @Test
  def testRemoveTransposeND(): Unit = {
    val input = CounterInteger(0, 1, Seq(2, 3, 4, 5), Some(IntType(32)))
    val tree = TransposeNDOrderedStream(TransposeNDOrderedStream(input, Seq(0, 2, 1, 3)), Seq(0, 2, 1, 3))

    HDLProject("noTranspose", tree, CompilerPhase.first(), Seq(
      (ArchCompiler.phaseAfter, RewriteStep(RewriteAll(), TransposeConversionRules.get())),
    )).compileHDL
  }

  def testMoveUpJoin_skipMapNDSplit(): Unit = {
    val input = Id(CounterInteger(0, 1, Seq(9, 6, 5, 4, 3, 2), Some(IntType(32))))
    val joinedInput = TypeChecker.check(JoinOrderedStream(input))
    val tree = MapOrderedStream(4, SplitOrderedStream.asFunction(Seq(None), Seq(3)), joinedInput)


    HDLProject("moveUpJoin_skipMapNDSplit", tree, CompilerPhase.first(), Seq(
      (ArchCompiler.phaseAfter, RewriteStep(RewriteAll(), Seq(MoveUpJoinStreamRules.skipMapNDSplit))),
    )).compileHDL
  }

  @Test
  def testMoveUpVecToStm_mapNDFission(): Unit = {
    val input = TypeChecker.check(Id(CounterInteger(0, 1, Seq(6, 5, 4, 3, 2), Some(IntType(32)))))
    val tree = MapOrderedStream(4,
      {
        val param = ParamDef()
        ArchLambda(
          param,
          MapOrderedStream(VectorToOrderedStream.asFunction(), SlideOrderedStream(ParamUse(param), 3))
        )
      }, input
    )

    HDLProject("moveUpVecToStm_mapNDFission", tree, CompilerPhase.first(), Seq(
      (ArchCompiler.phaseAfter, RewriteStep(RewriteAll(), MoveUpVectorToStreamRules.get())),
    )).compileHDL
  }

  @Test
  def testMapBuffer(): Unit = {
    Counter.resetAll()
    val input = CounterInteger(0, 1, Seq(5, 5), None)
    val test = MapOrderedStream(
      {
        val param = ParamDef(input.t.asInstanceOf[OrderedStreamTypeT].et)
        ArchLambda(
          param,
          BufferStreamInBlockRam(ParamUse(param))
        )
      }, input
    )
    HDLProject("mapBuf", test).simulateStandalone()
  }

  @Test
  def testTransposeIndexStream(): Unit = {
    Counter.resetAll()
    val tree = IndexStream(JoinOrderedStream(JoinOrderedStream(TransposeNDOrderedStream(Id(CounterInteger(0, 1, Seq(3,4,5))), Seq(2, 0, 1)))))

    HDLProject("transposeIndex", tree).compileHDL
  }

  @Test
  def testSubtraction(): Unit = {
    Counter.resetAll()

    val tree = Unsigned(SubInt(Tuple(Signed(ConstantValue(10)), Signed(ConstantValue(3)))))

    HDLProject("subtraction", tree).compileHDL //simulateStandalone()
  }

  @Test
  def testReLU(): Unit = {
    Counter.resetAll()

    val tree = Unsigned(ReLUInt(SubInt(Tuple(Signed(ConstantValue(3)), Signed(ConstantValue(10))))))

    HDLProject("relu", tree).compileHDL //.simulateStandalone()
  }

  @Test
  def testSignedConst(): Unit = {
    Counter.resetAll()

    val tree = ConstantValue(-1, Some(SignedIntType(8)))

    HDLProject("signedConst", tree).simulateStandalone()
  }

  @Test
  def testSignedSum(): Unit = {
    Counter.resetAll()

    val tree = FoldOrderedStream(AddInt2.asFunction(), TypeChecker.check(MapOrderedStream(Signed.asFunction(), CounterInteger(0, 1, Seq(8)))))

    HDLProject("signedSum", tree).simulateStandalone()
  }

  @Test
  def testFold2DStreamSigned(): Unit = {
    val tree =
      FoldOrderedStream(
        {
          val p1 = ParamDef()
          val p2 = ParamDef()
          ArchLambdas(Seq(p1, p2),
            MapOrderedStream(
              AddInt.asFunction(),
              ZipOrderedStream(Tuple(ParamUse(p1), ParamUse(p2)))
            )
          )
        },
        TypeChecker.check(MapOrderedStream(2, Signed.asFunction(), CounterInteger(0, 1, Seq(8, 8))))
      )

    HDLProject("fold2DSigned", tree).simulateStandalone()
  }

  @Test
  def testFoldVectorSigned(): Unit = {
    Counter.resetAll()
    val tree = FoldVector(AddInt.asTypeFunction(), OrderedStreamToVector(MapOrderedStream(Signed.asFunction(), CounterInteger(0, 1, Seq(32), Some(IntType(32))))))

    HDLProject("foldvectorsigned", tree).simulateStandalone()
  }
}
