package cGen

import cGen.eqsat.Extractor.DefaultSmallest
import cGen.eqsat._
import cGen.idioms.{BlasExtractor, BlasIdiomRules, ExtractionBasedSemanticRules, LowerFunctionalPrimitives}
import core._
import org.junit.Test

import java.time.Duration

class BoundedEqualitySaturationTest {
  def areEquivalent[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass])(eClass: EClass, pattern: Expr): Boolean =
    Pattern(pattern).searchEClass(graph)(eClass, HierarchicalStopwatch.disabled).nonEmpty

  val simplificationRules = {
    val allRules = ExtractionBasedSemanticRules()
    allRules.inliningRule +: allRules.simplificationRules
  }

  val expansionRules = ExtractionBasedSemanticRules().expansionRules

  val lowLevelRules = simplificationRules ++ expansionRules

  val bimodalStrategy = BimodalSaturationStrategy(
    SimpleSaturationStrategy(expansionRules),
    SimpleSaturationStrategy(simplificationRules))

  val extractor = BlasExtractor.analysis

  val instructionMatchingRules = BlasIdiomRules(DefaultSmallest).rules

  @Test
  def canRewriteSumAsDotProduct(): Unit = {
    // As our starting point, we'll create a sum expression.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(t(), 3)))
    var sum = TypeChecker.check(Lib.VectorSum(a))
    sum = TypeChecker.check(DeBruijnTransform.forward(LowerFunctionalPrimitives(true)(sum).asInstanceOf[Expr]))

    // Build a graph and populate it with the starting point expression.
    val stats = new PEGStatisticsObserver()
    val graph = new Egg2022EGraph(stats)
    val sumClass = graph.add(sum)
    graph.rebuild()

    // Explore how we can transform our initial expression.
    graph.saturate(lowLevelRules, _ >= 3)

    // Compose an expression corresponding to the dot product we'd like to find.
    //    val zeroVec = TypeChecker.check(constantVector(3, Lib.one))
    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(Lib.DotProd(ExprVar(a.t), a)))
        .asInstanceOf[Expr])

    // Check that we have found our vanilla vector sum to be equivalent to the dot product.
    assert(areEquivalent(graph)(sumClass, desired))

    println(stats)
  }

  /**
   * Metrics for a single saturation operation.
   * @param numIterations The number of iterations performed by the saturation process.
   * @param numClasses The number of classes at the end of the saturation process.
   * @param numNodes The number of nodes at the end of the saturation process.
   * @param duration The wall clock time required for the saturation process.
   */
  private case class SaturationMetrics(numIterations: Int, numClasses: Int, numNodes: Int, duration: Duration)

  private def measureSaturationMetrics(maxIterations: Int,
                                       createGraph: PEGObserver => Egg2022EGraph,
                                       postSaturation: (Int, Egg2022EGraph) => Unit = (_, _) => (),
                                       useBimodalSaturation: Boolean = false) = {
    for (numIterations <- Seq.range(0, maxIterations + 1)) yield {
      // Build a graph and populate it with the starting point expression.
      val stats = new PEGStatisticsObserver()
      val graph = createGraph(stats)

      // Explore how we can transform our initial expression.
      val stopwatch = HierarchicalStopwatch.start("")
      if (useBimodalSaturation) {
        graph.saturate(
          BimodalSaturationStrategy(
            SimpleSaturationStrategy(expansionRules),
            SimpleSaturationStrategy(simplificationRules)),
          _ >= numIterations)
      } else {
        graph.saturate(lowLevelRules, _ >= numIterations)
        graph.saturate(simplificationRules)
      }
      postSaturation(numIterations, graph)
      SaturationMetrics(
          numIterations,
          graph.classes.size,
          graph.classes.flatMap(c => c.nodes).size,
          stopwatch.complete().duration)
    }
  }

  private def saturationMetricsByConfig(maxIterations: Int,
                                        configs: Map[String, PEGObserver => Egg2022EGraph],
                                        postSaturation: (Int, Egg2022EGraph) => Unit = (_, _) => (),
                                        useBimodalSaturation: Boolean = false) = {

    configs.map(t => (t._1, measureSaturationMetrics(maxIterations, t._2, postSaturation, useBimodalSaturation)))
  }

  private def configMetricsToCsv(metrics: Map[String, Seq[SaturationMetrics]]) = {
    val sortedKeys = metrics.keys.toSeq.sorted
    val headerData = Seq("Iterations") ++ sortedKeys.flatMap(k => Seq(s"${k}Nodes", s"${k}Classes", s"${k}Ms"))
    val rowCount = metrics.values.map(_.size).max
    val rowData = Seq.range(0, rowCount).map(i => Seq(i) ++ sortedKeys.flatMap(k => {
      if (metrics(k).size <= i) {
        Seq("", "", "")
      } else {
        val data = metrics(k)(i)
        Seq(data.numNodes.toString, data.numClasses.toString, data.duration.toMillis.toString)
      }
    }))
    val allRowData = Seq(headerData) ++ rowData
    allRowData.map(row => row.mkString(",")).mkString("\n")
  }

  private def configFromExpr(expr: Expr,
                             useAnalysisExpressionExtractor: Boolean = false,
                             applyMatchesOnce: Boolean = false,
                             useBatchSearch: Boolean = true) = {
    (observer: PEGObserver) => {
      val analysesBuilder = new AnalysesBuilder()
      if (useAnalysisExpressionExtractor) {
        analysesBuilder.add(ExtractionAnalysis.smallestExpressionExtractor)
      }
      analysesBuilder.add(extractor)
      val analyses = analysesBuilder.toAnalyses

      // Build a graph and populate it with the starting point expression.
      val graph = new Egg2022EGraph(
        observer,
        analyses = analyses,
        applyMatchesOnce = applyMatchesOnce,
        useBatchSearch = useBatchSearch)
      graph.add(expr)
      graph.rebuild()

      graph
    }
  }

  @Test
  def timeSumSaturation(): Unit = {
    // As our starting point, we'll create a sum expression.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(t(), 3)))
    var sum = TypeChecker.check(Lib.VectorSum(a))
    sum = TypeChecker.check(LowerFunctionalPrimitives(true)(sum).asInstanceOf[Expr])

    val configs = scala.Predef.Map(
      "Baseline" -> configFromExpr(sum),
      "RememberMatches" -> configFromExpr(sum, applyMatchesOnce = true),
      "AnalysisExprExtraction" -> configFromExpr(sum, useAnalysisExpressionExtractor = true),
      "AllDressed" -> configFromExpr(sum, useAnalysisExpressionExtractor = true, applyMatchesOnce = true)
    )

    val extendedConfigs = configs ++ scala.Predef.Map("NoBatchSearch" -> configFromExpr(sum, useBatchSearch = false))

    println(configMetricsToCsv(
      saturationMetricsByConfig(3, extendedConfigs) ++ saturationMetricsByConfig(5, configs)))
  }

  @Test
  def timeMatrixMulSaturation(): Unit = {
    // As our starting point, we'll create a matrix multiplication expression.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(ArrayType(t(), 3), 3)))
    val b = ParamUse(ParamDef(ArrayType(ArrayType(t(), 3), 3)))
    var product = TypeChecker.check(Lib.MatrixMul(a, b))
    product = TypeChecker.check(LowerFunctionalPrimitives(true)(product).asInstanceOf[Expr])

    val configs = scala.Predef.Map(
      "" -> configFromExpr(product, useAnalysisExpressionExtractor = true, applyMatchesOnce = true)
    )

    val maxIters = 4
    println(configMetricsToCsv(saturationMetricsByConfig(maxIters, configs, (iters, graph) => {
      if (iters == maxIters) {
        // Compose an expression corresponding to the GEMM call we'd like to find.
        val eqSatTest = new EqualitySaturationTest()
        val gemmPattern = {
          val alpha = ExprVar(t())
          val beta = ExprVar(t())
          val n = 3
          val a = ExprVar(ArrayType(ArrayType(t(), n), n))
          val b = ExprVar(a.t)
          val c = ExprVar(a.t)
          val mm = DeBruijnTransform.forward(
            LowerFunctionalPrimitives(true)(
              TypeChecker.check(eqSatTest.gemm(alpha, a, b, beta, c)))
              .asInstanceOf[Expr])
          println(ExprToCode(mm))
          mm
        }

        // Check that we have found our vanilla matrix multiplication to be equivalent to the GEMM call.
        assert(areEquivalent(graph)(graph.add(product), gemmPattern))
      }
    })))
  }

  private def createConvolutionKernel(flattenDimensions: Boolean = false) = {
    implicit val t: () => CGenDataTypeT = DoubleType

    val inChs = 3
    val inW = 1024
    val inH = 1024
    val kerW = 3
    val kerH = 3
    val outW = 1022
    val outChs = 3

    val in = ParamUse(ParamDef(ArrayType(ArrayType(ArrayType(t(), inChs), inW), inH)))
    val ks = ParamUse(ParamDef(ArrayType(ArrayType(ArrayType(ArrayType(t(), inChs), kerW), kerH), outChs)))

    def joinND2() = {
      val tensor = ParamDef(ArrayType(ArrayType(ArrayType(CGenDataTypeVar(), ArithTypeVar()), ArithTypeVar()), ArithTypeVar()))
      CLambda(tensor, Join(Join(ParamUse(tensor))))
    }

    def slideND2(windowHeight: ArithTypeT, windowWidth: ArithTypeT, matrix: Expr): Expr = {
      Map({
        val window = ParamDef(ArrayType(ArrayType(CGenDataTypeVar(), ArithTypeVar()), ArithTypeVar()))
        CLambda(window, Lib.transpose(ParamUse(window)))
      }, Slide(Map({
        val vec = ParamDef(matrix.t.asInstanceOf[ArrayTypeT].et)
        CLambda(vec, Slide(ParamUse(vec), windowHeight))
      }, matrix), windowWidth))
    }

    val inner = Map({
      val slideWin = ParamDef(ArrayType(t(), ArithTypeVar()))
      CLambda(slideWin, Map({
        val k = ParamDef(ArrayType(t(), ArithTypeVar()))
        CLambda(k, Reduce(Lib.add2, Constant(0, t()), Map(Lib.mulTup, Zip(Tuple(ParamUse(slideWin), ParamUse(k)), EagerType()))))
      }, Map(joinND2(), ks)))
    }, Map(joinND2(), Join(slideND2(kerH, kerW, in))))

    if (flattenDimensions) {
      inner
    } else {
      Split(inner, outW)
    }
  }

  private def findMistypedExpr(expr: Expr): Option[Expr] = {
    try {
      TypeChecker.check(expr)
      None
    }
    catch {
      case _: Throwable =>
        // See if any of the expression's children are mistyped.
        val children = expr match {
          case cGenExpr: CGenExpr => cGenExpr.args
          case _ => expr.children.collect({ case child: Expr => child })
        }
        for (child <- children) {
            findMistypedExpr(child) match {
              case Some(mistypedDescendant) => return Some(mistypedDescendant)
              case None =>
            }
        }
        Some(expr)
    }
  }

  @Test
  def timeConvolutionSaturation(): Unit = {
    // As our starting point, we'll create a convolution expression.
    var kernel: Expr = createConvolutionKernel(flattenDimensions = true)
    kernel = TypeChecker.check(kernel)
    kernel = TypeChecker.check(LowerFunctionalPrimitives(false)(kernel).asInstanceOf[Expr])
    println(kernel)

    val configs = scala.Predef.Map(
      "" -> configFromExpr(kernel, useAnalysisExpressionExtractor = true, applyMatchesOnce = true)
    )

    val maxIters = 4
    println(configMetricsToCsv(saturationMetricsByConfig(maxIters, configs, (iters, graph) => {
      if (iters == maxIters) {
        // Compose an expression corresponding to the GEMM call we'd like to find.
        val eqSatTest = new EqualitySaturationTest()
        val gemmPattern = {
          implicit val t = DoubleType
          val alpha = ExprVar(t())
          val beta = ExprVar(t())
          val a = ExprVar(ArrayType(ArrayType(t(), ArithTypeVar()), ArithTypeVar()))
          val b = ExprVar(ArrayType(ArrayType(t(), ArithTypeVar()), ArithTypeVar()))
          val c = ExprVar(ArrayType(ArrayType(t(), ArithTypeVar()), ArithTypeVar()))
          val mm = DeBruijnTransform.forward(
            LowerFunctionalPrimitives(true)(
              TypeChecker.check(eqSatTest.gemm(alpha, a, b, beta, c)))
              .asInstanceOf[Expr])
          println(ExprToCode(mm))
          mm
        }

        // Check that we have found our vanilla matrix multiplication to be equivalent to the GEMM call.
        graph.saturate(instructionMatchingRules, _ >= 1)
        val kernelClass = graph.add(kernel)
        graph.rebuild()
        assert(areEquivalent(graph)(kernelClass, gemmPattern))
        println(kernelClass.analysisResult(extractor).get._1)
        println(ExprToText(kernelClass.analysisResult(extractor).get._1))
      }
    }, useBimodalSaturation = false)))
  }

  @Test
  def testMulOneRules() = {
    val graph = new Egg2022EGraph(PEGDotGraph)
    val eClass = graph.add(DeBruijnParamUse(0, IntType(32)))
    graph.rebuild()
    graph.saturate(lowLevelRules, _ >= 4)
    assert(areEquivalent(graph)(eClass, Constant(1.0, IntType(32))))
  }
}
