package cGen

import algo.{AlgoFunType, AlgoFunTypeVar}
import cGen.Lib.{VectorAdd, tt}
import cGen.eqsat.{AdditivePattern, ConditionalSearcher, DeBruijnForall, DeBruijnLambda, DeBruijnParamUse, DeBruijnShift, DeBruijnShiftExpr, DeBruijnTransform, EClassT, EGraphT, ENodeT, GeneratedApplier, HierarchicalStopwatch, ExtractionBasedInliner, PEGDotGraph, PEGStatisticsObserver, Pattern, Egg2022EGraph, Rewrite, Searcher, StructuralArrayEquality, StructuralFunctionEquality, StructuralTupleEquality}
import cGen.idioms.LowerFunctionalPrimitives
import cGen.primitive.{Fst, Idx, Snd}
import core.{AnyTypeVar, ArithTypeT, ArithTypeVar, Expr, ExprVar, FunctionCall, ParamDef, ParamUse, Type, TypeChecker}
import org.junit.{Ignore, Test}

@Ignore
class EqualitySaturationTest {
  private val sumTreeType = IntType(32)
  private val sumTreeParam = ParamDef(sumTreeType)
  private val sumTree = BinaryOp(ParamUse(sumTreeParam), Constant(0.0, sumTreeType), AlgoOp.Add)

  private def multiplyVectorByScalar(vector: Expr, scalar: Expr) = {
    val checkedVector = TypeChecker.check(vector)
    val element = ParamDef(checkedVector.t.asInstanceOf[ArrayType].et)
    Map(
      CLambda(element,
        BinaryOp(scalar, ParamUse(element), AlgoOp.Mul)),
      checkedVector)
  }

  private def multiplyMatrixByScalar(matrix: Expr, scalar: Expr) = {
    val row = ParamDef(matrix.t.asInstanceOf[ArrayType].et)
    val element = ParamDef(Lib.elementType(row))
    Map(
      CLambda(row,
        Map(
          CLambda(element,
            BinaryOp(scalar, ParamUse(element), AlgoOp.Mul)),
          ParamUse(row))),
      matrix)
  }

  private def strided(xs: Expr, stride: ArithTypeT) = {
    val i = ParamDef(IntType(32))
    Build(
      CLambda(i, Idx(xs, BinaryOp(Lib.num(stride.ae.evalLong)(() => IntType(32)), ParamUse(i), AlgoOp.Mul), EagerType())),
      xs.t.asInstanceOf[ArrayTypeT].len.ae / stride.ae)
  }

  private def axpy(alpha: Expr, x: Expr, y: Expr, strideX: ArithTypeT, strideY: ArithTypeT)(implicit t: () => Type) = {
    VectorAdd(multiplyVectorByScalar(strided(x, strideX), alpha), strided(y, strideY))
  }

  private def addMatrices(m1: Expr, m2: Expr)(implicit t: () => Type = () => Lib.elementType(m1)) = {
    // A simple matrix addition without CFC expressions.
    val in = ParamDef(TupleType(tt(m1).asInstanceOf[ArrayTypeT].et, tt(m2).asInstanceOf[ArrayTypeT].et))
    Map(CLambda(in, VectorAdd(Fst(ParamUse(in), EagerType()), Snd(ParamUse(in), EagerType()))), Zip(Tuple(m1, m2)))
  }

  private def addTensors(m1: Expr, m2: Expr)(implicit t: () => Type = () => Lib.elementType(m1)) = {
    val in = ParamDef(TupleType(tt(m1).asInstanceOf[ArrayTypeT].et, tt(m2).asInstanceOf[ArrayTypeT].et))
    TypeChecker.check(Map(CLambda(in, addMatrices(Fst(ParamUse(in), EagerType()), Snd(ParamUse(in), EagerType()))), Zip(Tuple(m1, m2))))
  }

  def gemm(a: Expr, A: Expr, B: Expr, b: Expr, C: Expr)(implicit t: () => Type) = {
    addMatrices(Lib.MatrixMul(multiplyMatrixByScalar(A, a), B), multiplyMatrixByScalar(C, b))
  }

  private def constantVector(size: ArithTypeT, constant: Expr) = {
    val param = ParamDef(IntType(32))
    Build(CLambda(param, constant), size)
  }

  def constantMatrix(n: ArithTypeT, m: ArithTypeT, constant: Expr) = {
    val param = ParamDef(IntType(32))
    Build(CLambda(param, constantVector(m, constant)), n)
  }

  private def constantTensor(n: ArithTypeT, m: ArithTypeT, k: ArithTypeT, constant: Expr) = {
    val param = ParamDef(IntType(32))
    Build(CLambda(param, constantMatrix(m, k, constant)), n)
  }

  def areEquivalent[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass])(eClass: EClass, pattern: Expr): Boolean =
    Pattern(pattern).searchEClass(graph)(eClass, HierarchicalStopwatch.disabled).nonEmpty

  @Test
  def sumCanBeAddedToGraph(): Unit = {
    val graph = new Egg2022EGraph()
    val eClass = graph.add(sumTree)
    graph.rebuild()

    assert(areEquivalent(graph)(eClass, sumTree))
  }

  @Test
  def deBruijnLambdaCanBeAddedToGraph(): Unit = {
    val graph = new Egg2022EGraph()
    val lambda = DeBruijnLambda.from(CLambda(sumTreeParam, sumTree))
    val eClass = graph.add(lambda)
    graph.rebuild()

    assert(areEquivalent(graph)(eClass, lambda))
  }

  @Test
  def algorithmicIdentitiesCanBeApplied(): Unit = {
    // Build a graph from an expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(sumTree)
    graph.rebuild()

    // Check that we have added the expression. Ensure that the expression appears syntactically different from the
    // parameter it contains.
    assert(areEquivalent(graph)(eClass, sumTree))
    assert(!areEquivalent(graph)(eClass, ParamUse(sumTreeParam)))

    // Define a rewrite rule.
    val tv = IntTypeVar()
    val exprVar = ExprVar(tv)
    val rule = Rewrite("AddZeroIsIdentity", BinaryOp(exprVar, Constant(0.0, tv), AlgoOp.Add), exprVar)

    // Apply the rewrite rule to the graph until we reach a fixpoint.
    graph.saturate(Seq(rule))

    // Check that the expression is equivalent to the parameter it contains.
    assert(areEquivalent(graph)(eClass, sumTree))
    assert(areEquivalent(graph)(eClass, ParamUse(sumTreeParam)))
  }

  //noinspection AccessorLikeMethodIsUnit
  @Test
  def canReproduceEggExample(): Unit = {
    // This test checks that we can reproduce the example presented in "Fast and Extensible Equality Saturation with
    // egg," by Max Willsey on April 6, 2021. https://blog.sigplan.org/2021/04/06/equality-saturation-with-egg/

    // Synthesize the starting point expression: `(a * 2) / 2`, where `a` is a double.
    val t = DoubleType()
    val param = ParamDef(t)
    val expr = TypeChecker.check(
      BinaryOp(BinaryOp(ParamUse(param), Constant(2.0, t), AlgoOp.Mul), Constant(2.0, t), AlgoOp.Div))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(expr)
    graph.rebuild()

    // Check that we have added the expression. Do a quick sanity check.
    assert(areEquivalent(graph)(eClass, expr))
    assert(!areEquivalent(graph)(eClass, ParamUse(param)))

    // Define rewrite rules.
    val exprVar = ExprVar(t)
    val exprVar2 = ExprVar(t)
    val exprVar3 = ExprVar(t)
    val rules = Seq(
      // x * 1 -> x
      Rewrite("OneIsMulIdentityElem", BinaryOp(exprVar, Constant(1.0, t), AlgoOp.Mul), exprVar),

      // (x * y) / z -> x * (y / z)
      Rewrite(
        "ReassociateMulDiv",
        BinaryOp(BinaryOp(exprVar, exprVar2, AlgoOp.Mul), exprVar3, AlgoOp.Div),
        BinaryOp(exprVar, BinaryOp(exprVar2, exprVar3, AlgoOp.Div), AlgoOp.Mul)),

      // x / x -> 1
      Rewrite(
        "DivIdentity",
        BinaryOp(exprVar, exprVar, AlgoOp.Div),
        Constant(1.0, t))
    )

    // Apply the rewrite rules to the graph until we reach a fixpoint.
    graph.saturate(rules)

    // Check that the expression is equivalent to the parameter it contains.
    assert(areEquivalent(graph)(eClass, expr))
    assert(areEquivalent(graph)(eClass, ParamUse(param)))
  }

  //noinspection AccessorLikeMethodIsUnit
  @Test
  def canRecognizeMatMul(): Unit = {
    // This test checks that we can recognize a matrix multiplication and rewrite it as a well-known function call.

    // As our starting point, we'll create a simple matrix multiplication expression.
    val t = DoubleType()
    val m1 = ParamDef(ArrayType(ArrayType(t, 3), 3))
    val m2 = ParamDef(ArrayType(ArrayType(t, 4), 3))
    var tree: Expr = CLambda(m1, CLambda(m2, Lib.MatrixMul(ParamUse(m1), ParamUse(m2))(isTransposeLazy = true, DoubleType)))
    tree = TypeChecker.check(tree)

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(tree)
    graph.rebuild()

    // Define rewrite rules.
    val rules = Seq(
      // M x N -> matrix_mul(M, N)
      {
        val scalarType = DoubleType()
        val n = ArithTypeVar()
        val m = ArithTypeVar()
        val k = ArithTypeVar()
        val l = ArithTypeVar()
        val lhs = ExprVar(ArrayType(ArrayType(scalarType, n), m))
        val rhs = ExprVar(ArrayType(ArrayType(scalarType, k), l))
        Rewrite(
          "DoubleMatrixMul",
          Lib.MatrixMul(lhs, rhs)(isTransposeLazy = true, DoubleType),
          Extern.MatrixMul(lhs, rhs)
        )
      }
    )

    graph.saturate(rules)

    assert(areEquivalent(graph)(
      eClass,
      TypeChecker.check(CLambda(m1, CLambda(m2, Extern.MatrixMul(ParamUse(m1), ParamUse(m2)))))))
  }

  @Test
  def injectConstantInMatMul(): Unit = {
    // This test checks that we can introduce a constant multiplication into a matrix multiplication. This takes us part
    // of the way to recognizing that matrix multiplications are a special form of gemm.

    // As our starting point, we'll create a simple matrix multiplication expression.
    val t = DoubleType()
    val m1 = ParamDef(ArrayType(ArrayType(t, 3), 3))
    val m2 = ParamDef(ArrayType(ArrayType(t, 4), 3))
    var tree: Expr = CLambda(m1, CLambda(m2, Lib.MatrixMul(ParamUse(m1), ParamUse(m2))(isTransposeLazy = true, DoubleType)))
    tree = TypeChecker.check(tree)

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(tree)
    val paramEClass = graph.add(ParamUse(m1))
    graph.rebuild()

    // Define rewrite rules.
    val rules = Seq(
      // x -> 1 * x
      {
        val exprVar = ExprVar(t)
        Rewrite("OneIsMulIdentityElem", exprVar, BinaryOp(Constant(1.0, t), exprVar, AlgoOp.Mul))
      },

      // xs -> map (\x. x) xs
      {
        val tVar = CGenDataTypeVar()
        val len = ArithTypeVar()
        val exprVar = ExprVar(ArrayType(tVar, len))
        val param = ParamDef(tVar)
        Rewrite("IntroduceMapId", exprVar, Map(CLambda(param, ParamUse(param)), exprVar))
      }
    )

    graph.saturate(rules)

    assert(areEquivalent(graph)(paramEClass, multiplyMatrixByScalar(ParamUse(m1), Constant(1.0, t))))

    assert(areEquivalent(graph)
          (eClass, TypeChecker.check(CLambda(m1, CLambda(m2,
            Lib.MatrixMul(multiplyMatrixByScalar(ParamUse(m1), Constant(1.0, t)), ParamUse(m2))(isTransposeLazy = true, DoubleType))))))
  }

  private val matrixAddRules = Seq(
    // x -> x + 0
    {
      val t = DoubleType()
      val exprVar = ExprVar(t)
      Rewrite("IntroduceAddZero", exprVar, BinaryOp(exprVar, Constant(0.0, t), AlgoOp.Add))
    },

    // x -> 1 * x
    {
      val t = DoubleType()
      val exprVar = ExprVar(t)
      Rewrite("IntroduceMulByOne", exprVar, BinaryOp(Constant(1.0, t), exprVar, AlgoOp.Mul))
    },

    // xs -> map (\x. x) xs
    {
      val tVar = CGenDataTypeVar()
      val len = ArithTypeVar()
      val exprVar = ExprVar(ArrayType(tVar, len))
      val param = ParamDef(tVar)
      Rewrite("IntroduceMapId", exprVar, Map(CLambda(param, ParamUse(param)), exprVar))
    },

    // map (\x. f x + g x) xs -> map (\p. p._1 + p._2) (zip (map (\x. f x) xs) (map (\x. g x) xs))
    // where x is not a tuple
    {
      val t = CGenDataTypeVar()
      val tResult = CGenDataTypeVar()
      val len = ArithTypeVar()
      val xs = ExprVar(ArrayType(t, len))
      val fx = ExprVar(tResult)
      val gx = ExprVar(tResult)
      val x = ParamDef(t)
      val p = ParamDef(TupleType(tResult, tResult))
      val from = Map(CLambda(x, BinaryOp(fx, gx, AlgoOp.Add)), xs)
      val to = Map(
        CLambda(p, BinaryOp(Fst(ParamUse(p), EagerType()), Snd(ParamUse(p), EagerType()), AlgoOp.Add)),
        Zip(Tuple(Map(CLambda(x, fx), xs), Map(CLambda(x, gx), xs))))
      Rewrite(
        "SplitMap",
        ConditionalSearcher(Pattern(from), m => !m.instantiate(t).isInstanceOf[NamedTupleStructTypeT]),
        Pattern(to))
    },

    // map (\x. y) xs -> build (\i. y) (len xs)
    // where y does not reference x
    {
      val t = CGenDataTypeVar()
      val len = ArithTypeVar()
      val x = ParamDef(t)
      val y = ExprVar(AnyTypeVar())
      val xs = ExprVar(ArrayType(t, len))
      val from = Map(CLambda(x, y), xs)
      val to = Build(CLambda(ParamDef(IntType(32)), y), len)
      Rewrite(
        "ConstMapToBuild",
        ConditionalSearcher(
          Pattern(from),
          m => m.instantiate(y).hasExprSatisfying(z => z.params.forall(_.id != m.instantiate(x).id))),
        Pattern(to))
    },

    // Tuple(Select(x, 0), Select(x, 1)) -> x
    {
      val x = ExprVar(AnyTypeVar())
      Rewrite("SimplifyTupleFromSelect", Tuple(Fst(x, EagerType()), Snd(x, EagerType())), x)
    },

    // x -> Tuple(Select(x, 0), Select(x, 1))
    {
      val x = ExprVar(TupleType(CGenDataTypeVar(), CGenDataTypeVar()))
      Rewrite("ExpandTupleSelect", x, Tuple(Fst(x, EagerType()), Snd(x, EagerType())))
    },

    // Select(Tuple(x, y), 0) -> x
    {
      val x = ExprVar(CGenDataTypeVar())
      val y = ExprVar(CGenDataTypeVar())
      Rewrite("SimplifyTupleSelect0", Fst(Tuple(x, y), EagerType()), x)
    },

    // Select(Tuple(x, y), 1) -> y
    {
      val x = ExprVar(CGenDataTypeVar())
      val y = ExprVar(CGenDataTypeVar())
      Rewrite("SimplifyTupleSelect1", Snd(Tuple(x, y), EagerType()), y)
    },

    // Map(CLambda(xs, Map(f, Zip(Tuple(xs, ys))), m) ->
    //   Map(CLambda(in, Map(f, Zip(in))), Zip(Tuple(m, Map(CLambda(xs, ys), m)))
    {
      val t = CGenDataTypeVar()
      val rows = ArithTypeVar()
      val columns = ArithTypeVar()
      val xsType = ArrayType(t, rows)
      val ysType = CGenDataTypeVar()
      val m = ExprVar(ArrayType(xsType, columns))
      val xs = ParamDef(xsType)
      val ys = ExprVar(ysType)
      val in = ParamDef(TupleType(xsType, ysType))
      val f = ExprVar(AnyTypeVar())
      val from = Map(CLambda(xs, Map(f, Zip(Tuple(ParamUse(xs), ys)), EagerType())), m)
      val to = Map(CLambda(in, Map(f, Zip(ParamUse(in)), EagerType())), Zip(Tuple(m, Map(CLambda(xs, ys), m, EagerType()))))
      Rewrite("HoistTupleFromMap", from, to)
    }
  )

  @Test
  def canIntroduceVectorAdd(): Unit = {
    // As our starting point, we'll create a vector variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(ArrayType(t, 3)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(v)
    graph.rebuild()

    graph.saturate(matrixAddRules)

    assert(areEquivalent(graph)(
      eClass,
      Map(Lib.addTup(DoubleType), Zip(Tuple(v, constantVector(Lib.length(v), Lib.num(0)(DoubleType)))))))

    assert(areEquivalent(graph)(
      eClass,
      VectorAdd(v, constantVector(Lib.length(v), Lib.num(0)(DoubleType)))(DoubleType)))
  }

  @Test
  def tuplesDoNotExpandForever(): Unit = {
    // As our starting point, we'll create a matrix variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(TupleType(t, t)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(v)
    graph.rebuild()

    graph.saturate(matrixAddRules)

    // If this test terminates, then we win. :)
  }

  @Test
  def canIntroduceMatrixAdd(): Unit = {
    // As our starting point, we'll create a matrix variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(ArrayType(ArrayType(t, 3), 3)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(v)
    graph.rebuild()

    graph.saturate(matrixAddRules)

    // First check that we can add a zero vector to each row of the input matrix.
    assert(areEquivalent(graph)(
      eClass,
      {
        val p = ParamDef(ArrayType(t, 3))
        Map(
          CLambda(p,
            VectorAdd(
              ParamUse(p),
              constantVector(Lib.length(p), Lib.num(0)(DoubleType)))(DoubleType)),
          v)
      }))

    // We now want to check that we can add a zero matrix to the entire input matrix. For this to work, the equality
    // saturation engine will have to make the following transformation:
    //
    // Map(CLambda(xs, VectorAdd(xs, 0)), m)
    // = Map(CLambda(xs, Map(addTup, Zip(Tuple(xs, 0))), m)
    // = Map(CLambda(in, Map(addTup, Zip(in)), Zip(Tuple(m, 0)))
    // = Map(CLambda(in, Map(addTup, Zip(Tuple(Select(ParamUse(in), 0), Select(ParamUse(in), 1)))))), Zip(Tuple(m, 0)))
    // = Map(CLambda(in, VectorAdd(Select(ParamUse(in), 0), Select(ParamUse(in), 1))), Zip(Tuple(m, 0)))

    // Check that we can add a zero matrix to the input.
    assert(areEquivalent(graph)(
      eClass,
      addMatrices(v, constantMatrix(3, 3, Lib.num(0)(DoubleType)))))
  }

  @Test
  def canIntroduceTensorAdd(): Unit = {
    // As our starting point, we'll create a 3D tensor variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(ArrayType(ArrayType(ArrayType(t, 3), 3), 3)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(v)
    graph.rebuild()

    graph.saturate(matrixAddRules)

    // First check that we can inject vector addition where the rhs is a zero vector.
    assert(areEquivalent(graph)(
      eClass,
      {
        val p1 = ParamDef(ArrayType(t, 3))
        val p2 = ParamDef(ArrayType(ArrayType(t, 3), 3))
        Map(
          CLambda(p2,
            Map(
              CLambda(p1,
                VectorAdd(
                  ParamUse(p1),
                  constantVector(Lib.length(p1), Lib.num(0)(DoubleType)))(DoubleType)),
              ParamUse(p2))),
          v)
      }))

    // Next, check that we can inject matrix addition where the rhs is a zero matrix.
    assert(areEquivalent(graph)(
      eClass,
      {
        val p = ParamDef(ArrayType(ArrayType(t, 3), 3))
        Map(
          CLambda(p, addMatrices(ParamUse(p), constantMatrix(3, 3, Lib.num(0)(DoubleType)))),
          v)
      }))

    // Finally, check that we can inject tensor addition where the rhs is a zero tensor.
    assert(areEquivalent(graph)(
      eClass,
      addTensors(v, constantTensor(3, 3, 3, Lib.num(0)(DoubleType)))))
  }

  @Test
  def canRewriteMatrixMulAsGemm(): Unit = {
    // As our starting point, we'll create two matrix variables.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(ArrayType(t(), 3), 3)))
    val b = ParamUse(ParamDef(a.t))
    val product = TypeChecker.check(Lib.MatrixMul(a, b))

    // Build a graph and populate it with the starting point expression.
    val stats = new PEGStatisticsObserver()
    val graph = new Egg2022EGraph(stats)
    val productClass = graph.add(product)
    graph.rebuild()

    graph.saturate(matrixAddRules)

    assert(areEquivalent(graph)(
      productClass,
      TypeChecker.check(
        Lib.MatrixMul(multiplyMatrixByScalar(a, Lib.one), b))))

    assert(areEquivalent(graph)(
      productClass,
      addMatrices(product, constantMatrix(3, 3, Lib.zero))))

    assert(areEquivalent(graph)(
      productClass,
      addMatrices(product, multiplyMatrixByScalar(TypeChecker.check(constantMatrix(3, 3, Lib.zero)), Lib.one))))

    assert(areEquivalent(graph)(
      productClass,
      addMatrices(
        multiplyMatrixByScalar(product, Lib.one),
        multiplyMatrixByScalar(TypeChecker.check(constantMatrix(3, 3, Lib.zero)), Lib.one))))

    assert(areEquivalent(graph)(
      productClass,
      addMatrices(
        Lib.MatrixMul(multiplyMatrixByScalar(a, Lib.one), b),
        multiplyMatrixByScalar(TypeChecker.check(constantMatrix(3, 3, Lib.zero)), Lib.one))))

    assert(areEquivalent(graph)(
      productClass,
      gemm(Lib.one, a, b, Lib.one, TypeChecker.check(constantMatrix(3, 3, Lib.zero)))))

    println(stats)
  }

  val lowLevelRules = {
    def doNotMatchSpecial(searcher: Searcher) = {
      ConditionalSearcher(searcher, matchInfo => {
        !matchInfo.where.nodes.map(_.expr).exists({
          case _: DeBruijnForall => true
          case DeBruijnShift(_, _) => true
          case _ => false
        })
      })
    }

    Seq(
      // ====
      // Axioms, left to right.
      // (build f N)[i] -> f i
      {
        val f = ExprVar(AnyTypeVar())
        val i = ExprVar(IntTypeVar())
        Rewrite("EliminateBuild", Idx(Build(f, ArithTypeVar(), EagerType()), i, EagerType()), FunctionCall(f, i))
      },

      // (a, b)._1 -> a
      {
        val a = ExprVar(CGenDataTypeVar())
        val b = ExprVar(CGenDataTypeVar())
        Rewrite("EliminateTuple1", Fst(Tuple(a, b), EagerType()), a)
      },

      // (a, b)._2 -> b
      {
        val a = ExprVar(CGenDataTypeVar())
        val b = ExprVar(CGenDataTypeVar())
        Rewrite("EliminateTuple2", Snd(Tuple(a, b), EagerType()), b)
      },

      // (\x. e) y -> e[y/x]
      {
        val t = AnyTypeVar()
        val body = ExprVar(AnyTypeVar())
        val arg = ExprVar(t)
        Rewrite("InlineLambda", Pattern(FunctionCall(DeBruijnLambda(t, body), arg)), ExtractionBasedInliner(true))
      },

      // shift(x) -> x
      // if the shift is a nop
      // *and*
      // shift(param i) -> param (i + 1)
      // *and*
      // shift(\ expr) -> \ shifted-expr
      {
        val x = ExprVar()
        Rewrite("ElideShift", Pattern(DeBruijnShift(x)), GeneratedApplier((matchInfo, _) => {
          matchInfo.where.toSmallestExpr match {
            case shift: DeBruijnShiftExpr if shift.isNop => Pattern(x)
            case DeBruijnShift(param: DeBruijnParamUse, _) => Pattern(DeBruijnParamUse(param.index + 1, param.t))
//            case DeBruijnShift(lam: DeBruijnLambda, _) => Pattern(DeBruijnParamUse.incrementUnbound(lam))
            case _ => Pattern(DeBruijnShift(x))
          }
        }))
      },
      // ====

      // ====
      // Rules derived from axioms, right to left.
      // xs -> build (\i. xs[i]) (length xs)
      // aka xs -> map (\x. x) xs
      {
        val elementType = CGenDataTypeVar()
        val len = ArithTypeVar()
        val i32 = IntType(32)
        val xs = ExprVar(ArrayType(elementType, len))
        Rewrite(
          "IntroduceMapId",
          doNotMatchSpecial(Pattern(xs)),
          Pattern(
            Build(
              DeBruijnLambda(i32, Idx(DeBruijnShift(xs), DeBruijnParamUse(0, i32), EagerType())),
              len),
            mayRename = _ => false))
      },
      // ====

      // ====
      // Equality definitions, quantifier introduction
      // xs -> xs, create forall i. shift(xs)[i],
      // if xs is not a shift
      {
        val t = CGenDataTypeVar()
        val xs = ExprVar(ArrayTypeVar(t))
        Rewrite(
          "IntroduceArrayForall",
          doNotMatchSpecial(Pattern(xs)),
          AdditivePattern({
            val indexType = IntType(32)
            DeBruijnForall(indexType, Idx(DeBruijnShift(xs), DeBruijnParamUse(0, indexType), EagerType()))
          }))
      },

      // t -> t, create t._1
      {
        val t = ExprVar(TupleTypeVar(CGenDataTypeVar(), CGenDataTypeVar()))
        Rewrite(
          "IntroduceTupleFst",
          Pattern(t),
          AdditivePattern(Fst(t, EagerType())))
      },

      // t -> t, create t._2
      {
        val t = ExprVar(TupleTypeVar(CGenDataTypeVar(), CGenDataTypeVar()))
        Rewrite(
          "IntroduceTupleSnd",
          Pattern(t),
          AdditivePattern(Snd(t, EagerType())))
      },

      // f -> f, create forall x. shift(f) x,
      // if f is a function and f is not a forall quantifier or a shift
      {
        val paramType = AnyTypeVar()
        val f = ExprVar(AlgoFunTypeVar(paramType))
        Rewrite(
          "IntroduceFunctionForall",
          doNotMatchSpecial(Pattern(f)),
          AdditivePattern(DeBruijnForall(paramType, FunctionCall(DeBruijnShift(f), DeBruijnParamUse(0, paramType)))))
      },

      // ====

      // ====
      // Equality definitions, structural equality
      // xs[x_0] = ys[x_0] => xs = ys
      {
        Rewrite(
          "StructuralArrayEquality",
          StructuralArrayEquality.pattern,
          StructuralArrayEquality)
      },

      // t._1 = u._1 and t._2 = u._2 => t = u
      {
        val t = ExprVar(TupleTypeVar(CGenDataTypeVar(), CGenDataTypeVar()))
        Rewrite(
          "StructuralTupleEquality",
          Pattern(Fst(t, EagerType())),
          StructuralTupleEquality)
      },

      // f x_0 = g x_0 => f = g
      {
        Rewrite(
          "StructuralFunctionEquality",
          StructuralFunctionEquality.pattern,
          StructuralFunctionEquality)
      },
      // ====

      // ====
      // Arithmetic axioms
      // x -> x + 0
      {
        val t = DoubleType()
        val exprVar = ExprVar(t)
        Rewrite("IntroduceAddZero", exprVar, BinaryOp(exprVar, Constant(0.0, t), AlgoOp.Add))
      },

      // x -> 1 * x
      {
        val t = DoubleType()
        val exprVar = ExprVar(t)
        Rewrite("IntroduceMulOne", exprVar, BinaryOp(Constant(1, t), exprVar, AlgoOp.Mul))
      },

      // x -> 1 * x
      {
        val t = IntTypeVar()
        val exprVar = ExprVar(t)
        Rewrite("IntroduceMulOne", exprVar, BinaryOp(Constant(1, t), exprVar, AlgoOp.Mul))
      }
      // ====
    )
  }

  @Test
  def canSimplifyAccessBuild(): Unit = {
    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val p = ParamDef(IntType(32))
    val lambda = CLambda(p, ParamUse(p))
    val const = Constant(5, p.t)
    val eClass = graph.add(DeBruijnTransform.forward(Idx(Build(lambda, 10), const, EagerType())))
    graph.rebuild()

    graph.saturate(lowLevelRules)

    assert(areEquivalent(graph)(eClass, FunctionCall(DeBruijnTransform.forward(lambda), const)))
    assert(areEquivalent(graph)(eClass, const))
  }

  @Test
  def canSimplifyMapId(): Unit = {
    // Build a graph and populate it with `map (\x. x) xs`.
    val graph = new Egg2022EGraph()
    val t = DoubleType()
    val xs = ParamUse(ParamDef(ArrayType(t, 20)))
    val p = ParamDef(t)
    val idLambda = CLambda(p, ParamUse(p))
    val mapClass = graph.add(DeBruijnTransform.forward(LowerFunctionalPrimitives(true)(Map(idLambda, xs)).asInstanceOf[Expr]))
    graph.rebuild()

    // Saturate.
    graph.saturate(lowLevelRules)

    // Check that we were able to show that `map (\x. x) xs = xs`.
    assert(areEquivalent(graph)(mapClass, xs))
  }

  @Test
  def canSimplifyFunctionWrapper(): Unit = {
    // Build a graph and populate it with `(\x. f x)`.
    val graph = new Egg2022EGraph()
    val t = DoubleType()
    val f = ParamUse(ParamDef(AlgoFunType(t, t)))
    val p = ParamDef(t)
    val wrapperLambda = CLambda(p, FunctionCall(f, ParamUse(p)))
    val wrapperLambdaClass = graph.add(DeBruijnTransform.forward(LowerFunctionalPrimitives(true)(wrapperLambda).asInstanceOf[Expr]))
    graph.rebuild()

    // Saturate.
    graph.saturate(lowLevelRules)

    // Check that we were able to show that `(\x. f x) = f`.
    assert(areEquivalent(graph)(wrapperLambdaClass, f))
  }

  @Test
  def canSimplifyTupleRecreation(): Unit = {
    // Build a graph and populate it with `(tup._1, tup._2)`.
    val graph = new Egg2022EGraph()
    val t = DoubleType()
    val tup = ParamUse(ParamDef(TupleType(t, t)))
    val recreatedTuple = Tuple(Fst(tup, EagerType()), Snd(tup, EagerType()))
    val recreatedTupleClass = graph.add(DeBruijnTransform.forward(recreatedTuple))
    graph.rebuild()

    // Saturate.
    graph.saturate(lowLevelRules)

    // Check that we were able to show that `(tup._1, tup._2) = tup`.
    assert(areEquivalent(graph)(recreatedTupleClass, tup))
  }

  @Test
  def canIntroduceVectorAddDeBruijn(): Unit = {
    // As our starting point, we'll create a vector variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(ArrayType(t, 3)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(DeBruijnTransform.forward(v))
    graph.rebuild()

    graph.saturate(lowLevelRules)

    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(VectorAdd(v, constantVector(Lib.length(v), Lib.num(0)(DoubleType)))(DoubleType)))
        .asInstanceOf[Expr])

    assert(areEquivalent(graph)(eClass, desired))
  }

  @Test
  def canRecognizeAXPYDeBruijn(): Unit = {
    // As our starting point, we'll create two vector variables.
    implicit val t = DoubleType
    val x = ParamUse(ParamDef(ArrayType(t(), 200)))
    val y = ParamUse(ParamDef(ArrayType(t(), 200)))

    // Build a graph and populate it with the starting point expression: x + y.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(DeBruijnTransform.forward(LowerFunctionalPrimitives(true)(TypeChecker.check(VectorAdd(x, y))).asInstanceOf[Expr]))
    graph.rebuild()

    graph.saturate(lowLevelRules)

    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(axpy(Lib.num(1.0), x, y, 1, 1)))
        .asInstanceOf[Expr])

    assert(areEquivalent(graph)(eClass, desired))
  }

  @Test
  def canIntroduceMatrixAddDeBruijn(): Unit = {
    // As our starting point, we'll create a matrix variable.
    val t = DoubleType()
    val v = ParamUse(ParamDef(ArrayType(ArrayType(t, 3), 3)))

    // Build a graph and populate it with the starting point expression.
    val graph = new Egg2022EGraph()
    val eClass = graph.add(DeBruijnTransform.forward(v))
    graph.rebuild()

    graph.saturate(lowLevelRules)

    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(addMatrices(v, constantMatrix(3, 3, Lib.num(0)(DoubleType)))))
        .asInstanceOf[Expr])

    assert(areEquivalent(graph)(eClass, desired))
  }

  @Test
  def canRewriteMatrixMulAsGemmDeBruijn(): Unit = {
    // As our starting point, we'll create a matrix multiplication expression.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(ArrayType(t(), 3), 3)))
    val b = ParamUse(ParamDef(a.t))
    var product = TypeChecker.check(Lib.MatrixMul(a, b))
    product = TypeChecker.check(LowerFunctionalPrimitives(true)(product).asInstanceOf[Expr])

    // Build a graph and populate it with the starting point expression.
    val stats = new PEGStatisticsObserver()
    val graph = new Egg2022EGraph(stats)
    val productClass = graph.add(DeBruijnTransform.forward(product))
    graph.rebuild()

    // Explore how we can transform our initial expression.
    graph.saturate(lowLevelRules)

    // Compose an expression corresponding to the GEMM call we'd like to find.
    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(gemm(Lib.one, a, b, Lib.one, TypeChecker.check(constantMatrix(3, 3, Lib.zero)))))
        .asInstanceOf[Expr])

    // Check that we have found our vanilla matrix multiplication to be equivalent to the GEMM call.
    assert(areEquivalent(graph)(productClass, desired))

    println(stats)
  }

  // 0 -> (\i. 0) j -> (build (\i. 0) N)[j] -> (\k. (build (\i. 0) N) [j]) l -> (build (\k. (build (\i. 0) N)[j]) M)[l]
  //   -> (build (\k. (build (\i. 0) N)) M)[l][j]

  // lemma: (build (\i. xs[j]) N)[k] = (build (\i. xs) N)[k][j]
  //
  // (build (\i. xs[j]) N)[k] = (\i. xs[j]) k
  //
  // (build (\i. xs) N)[k][j] = ((\i. xs) k)[j]

  @Test
  def canMatchTypeVar(): Unit = {
    // Check that rewrite rules can reify type variables.
    implicit val t: () => CGenDataTypeT = DoubleType
    val expr = LowerFunctionalPrimitives(true)(TypeChecker.check(constantMatrix(ArithTypeVar(), ArithTypeVar(), Lib.zero)))
      .asInstanceOf[Expr]
    val graph = new Egg2022EGraph()
    val exprClass = graph.add(expr)
    graph.rebuild()

    val desired = LowerFunctionalPrimitives(true)(TypeChecker.check(constantMatrix(3, 3, Lib.zero))).asInstanceOf[Expr]

    assert(areEquivalent(graph)(exprClass, desired))
  }

  @Test
  def canRewriteSumAsDotProduct(): Unit = {
    // As our starting point, we'll create a sum expression.
    implicit val t: () => CGenDataTypeT = DoubleType
    val a = ParamUse(ParamDef(ArrayType(t(), 3)))
    var sum = TypeChecker.check(Lib.VectorSum(a))
    sum = TypeChecker.check(LowerFunctionalPrimitives(true)(sum).asInstanceOf[Expr])

    // Build a graph and populate it with the starting point expression.
    val stats = new PEGStatisticsObserver()
    val graph = new Egg2022EGraph(PEGDotGraph)
    val sumClass = graph.add(DeBruijnTransform.forward(sum))
    graph.rebuild()

    // Explore how we can transform our initial expression.
    graph.saturate(lowLevelRules)

    // Compose an expression corresponding to the dot product we'd like to find.
//    val zeroVec = TypeChecker.check(constantVector(3, Lib.one))
    val desired = DeBruijnTransform.forward(
      LowerFunctionalPrimitives(true)(
        TypeChecker.check(Lib.DotProd(ExprVar(a.t), a)))
        .asInstanceOf[Expr])

    // Check that we have found our vanilla matrix multiplication to be equivalent to the dot product.
    assert(areEquivalent(graph)(sumClass, desired))

    println(stats)
  }
}
