package cGen

import cGen.Util.{FileAppender, FileWriter}
import cGen.eqsat.explain.TreeExplanation
import cGen.eqsat._
import cGen.idioms.ExtractionBasedSemanticRules
import cGen.kernels.Kernel
import cGen.primitive.{Idx, IdxS}
import core.util.Prune
import core._
import org.junit.Test

import java.time.Duration
import scala.annotation.tailrec
import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer

class PolyBenchEqualitySaturationTest {
  val kernelSize = 1024

  object ConvolutionKernel extends Kernel {
    override def name: String = "convolution"
    override def suite: String = "custom"

    override def input: Expr = createConvolutionKernel(flattenDimensions = true)
    override def oracles: Seq[Expr] = Seq()

    private def createConvolutionKernel(flattenDimensions: Boolean = false) = {
      implicit val t: () => CGenDataTypeT = DoubleType

      val inChs = 3
      val inW = 1024
      val inH = 1024
      val kerW = 3
      val kerH = 3
      val outW = 1022
      val outChs = 3

      val in = ParamUse(ParamDef(ArrayType(ArrayType(ArrayType(t(), inChs), inW), inH)))
      val ks = ParamUse(ParamDef(ArrayType(ArrayType(ArrayType(ArrayType(t(), inChs), kerW), kerH), outChs)))

      def joinND2() = {
        val tensor = ParamDef(ArrayType(ArrayType(ArrayType(CGenDataTypeVar(), ArithTypeVar()), ArithTypeVar()), ArithTypeVar()))
        CLambda(tensor, Join(Join(ParamUse(tensor))))
      }

      def slideND2(windowHeight: ArithTypeT, windowWidth: ArithTypeT, matrix: Expr): Expr = {
        Map({
          val window = ParamDef(ArrayType(ArrayType(CGenDataTypeVar(), ArithTypeVar()), ArithTypeVar()))
          CLambda(window, Lib.transpose(ParamUse(window)))
        }, Slide(Map({
          val vec = ParamDef(matrix.t.asInstanceOf[ArrayTypeT].et)
          CLambda(vec, Slide(ParamUse(vec), windowHeight))
        }, matrix), windowWidth))
      }

      val inner = Map({
        val slideWin = ParamDef(ArrayType(t(), ArithTypeVar()))
        CLambda(slideWin, Map({
          val k = ParamDef(ArrayType(t(), ArithTypeVar()))
          CLambda(k, Reduce(Lib.add2, Constant(0, t()), Map(Lib.mulTup, Zip(Tuple(ParamUse(slideWin), ParamUse(k)), EagerType()))))
        }, Map(joinND2(), ks)))
      }, Map(joinND2(), Join(slideND2(kerH, kerW, in))))

      if (flattenDimensions) {
        inner
      } else {
        Split(inner, outW)
      }
    }
  }

//  val kernels = Seq(cGen.kernels.AtaxKernel(1024, 1024))

  val kernels = Seq(
//    cGen.kernels.VectorSumKernel(256),
    cGen.kernels.AxpyKernel(256),
//    cGen.kernels.GesummvKernel(256),
//    cGen.kernels.GemverKernel(256),
//    cGen.kernels.GEMMKernel(256),
//    cGen.kernels.OneMMKernel(256),
//    cGen.kernels.DoitgenKernel(256, 256, 256),
//    cGen.kernels.MvtKernel(1024),
//    cGen.kernels.BicgKernel(1024, 1024),
//    cGen.kernels.AtaxKernel(1024, 1024),
//    cGen.kernels.Blur1DKernel(100000),
//    cGen.kernels.Jacobi1DKernel(100000),
//    cGen.kernels.Stencil2DKernel(256),
//    cGen.kernels.SlimTwoMMKernel(256),
//    cGen.kernels.TwoMMKernel(256)
  )

  private val writeKernels = false

  @Test
  def findBLAS(): Unit = {
    val rows = mutable.ArrayBuffer[Seq[Any]]()
    if (writeKernels) {
      FileWriter("./codegen/OnlyDot.scala", "")
      FileWriter("./codegen/OnlyDotAndGemv.scala", "")
      FileWriter("./codegen/All.scala", "")
    }

    for (kernel <- kernels) {
      FileWriter(s"./codegen/${kernel.name}Baseline.scala", ExprToScala.generateTest(DeBruijnLambda.toNamedRec(kernel.input).asInstanceOf[Expr], kernel.name))

//      val preferredExtractor = Extractor.DefaultSmallest
      val preferredExtractor = kernel.extractor
      println(s"Optimizing ${kernel.name}...")

      val stopwatch = HierarchicalStopwatch.start("")
      val loweredKernel = kernel.transformedInput

      Prune.findUnsolvedTypeVar(loweredKernel).foreach(smallest =>
        println(s"warning: found unsolved type var in ${ExprToText(smallest)}"))

      // Build a graph and populate it with the kernel.
      val nodeCounter = new NodesOverTimeObserver()
      val externCounter = new ExternCallInResultObserver(None, kernel.extractor)
      val stats = new PEGStatisticsObserver()
      val (graph, root) = kernel.buildGraph(
        PEGCompositeObserver(stats, nodeCounter, externCounter),
        // shouldPrintProgress = true,
        enableExplanations = false,
        checkConsistency = true)

      // Configure extern call counter root.
      externCounter.root = Some(root)

      // Explore how we can transform the kernel.
      val timer = HierarchicalStopwatch.start("")
      kernel.saturateGraphNaively(graph, rules = ExtractionBasedSemanticRules(preferredExtractor).rules)
      graph.recomputeAnalysis(kernel.extractionAnalysis)
      val duration = timer.complete().duration

      val bestExpr = kernel.extractor(root).get

      val onlyDotAnalysis = kernel.extractionAnalysis.exclude("only-dot-extractor", {
        case ExternFunctionCall("idiom_dgemv", _, _, _) => true
        case ExternFunctionCall("idiom_dgemm", _, _, _) => true
        case _ => false
      })
      graph.recomputeAnalysis(onlyDotAnalysis)

      val onlyDotAndGemvAnalysis = kernel.extractionAnalysis.exclude("only-dot-and-gemv-extractor", {
        case ExternFunctionCall("idiom_dgemm", _, _, _) => true
        case _ => false
      })
      graph.recomputeAnalysis(onlyDotAndGemvAnalysis)

      if (kernel.matchesOracle(bestExpr)) {
        println(s"Perfect result for ${kernel.name} after ${stopwatch.complete().duration.toSeconds} seconds")
        graph.explainer.foreach(explain => printReasoning(explain.explainEquivalence(kernel.transformedInput, bestExpr)))

        val oracle = kernel.tryFindOracle(bestExpr).get
        if (!kernel.matchesOracle(graph, root)) {
          println("warning: oracle not found by pattern matcher")
          println(s"Oracle:      $oracle")
          println(s"Best result: $bestExpr")
          println(s"Smallest unequal expression: ${Pattern.findDivergence(oracle, bestExpr)}.")

          kernel.relaxOracle(graph, root) match {
            case Some(superpattern) => println(s"Pattern matching both oracle and graph: ${ExprToText(superpattern)}")
            case None => println("No nontrivial pattern matches both oracle and graph")
          }
        }
      } else {
        println(s"Imperfect result for ${kernel.name}!")
        println(s"Expected: ${kernel.oracles.map(ExprToText(_)).mkString("\n -or-     ")}")
        println(s"Got:      ${ExprToText(bestExpr)}")

        if (kernel.matchesOracle(graph, root)) {
          println("note: oracle found by pattern matcher")
        } else {
          println("note: oracle not found by pattern matcher")
        }

//        println(stats)
        println(s"Original kernel: ${ExprToText(loweredKernel)}")
        println(s"Original cost: ${kernel.extractionAnalysis.totalCost(loweredKernel)}")
//        println(s"Optimized kernel cost: ${bestExpr._2}, size: ${bestExpr._3}")
        println(ExprToText(root.toMinimalExpr(ExtractionAnalysis.smallestExpressionExtractor.cost)))
        graph.explainer.foreach(explain => printReasoning(explain.explainEquivalence(kernel.transformedInput, bestExpr)))
      }

      val row = buildCsvRow(kernel, graph, root, duration)
      rows.append(row)
      println(s"CSV row: $row")
      println(buildNodeCountChart(nodeCounter, externCounter))

      val onlyDotExpr = onlyDotAnalysis.extractor(root).get
      val onlyDotAndGemvExpr = onlyDotAndGemvAnalysis.extractor(root).get

      if (writeKernels) {
        FileAppender(s"./codegen/OnlyDot.scala", ExprToScala.generateTest(DeBruijnLambda.toNamedRec(onlyDotExpr).asInstanceOf[Expr], kernel.name))
        FileAppender(s"./codegen/OnlyDotAndGemv.scala", ExprToScala.generateTest(DeBruijnLambda.toNamedRec(onlyDotAndGemvExpr).asInstanceOf[Expr], kernel.name))
        FileAppender(s"./codegen/All.scala", ExprToScala.generateTest(DeBruijnLambda.toNamedRec(bestExpr).asInstanceOf[Expr], kernel.name))

//        println(CCodeGen(DPSPass.apply(bestExpr)).ccode())
      }
    }

    println(assembleCsv(rows))
  }

  private def buildNodeCountChart(nodeCounter: NodesOverTimeObserver, externCounter: ExternCallInResultObserver): String = {
    @tailrec
    def trimName(name: String): String = {
      if (name.startsWith("d")) {
        name.substring(1)
      } else if (name.startsWith("cblas_")) {
        trimName(name.substring("cblas_".length))
      } else {
        name
      }
    }

    val lines = nodeCounter.nodesOverTime.zipWithIndex.map(t => s"    (${t._2}, ${t._1})").mkString("\n")
    val plot = s"\\addplot[color=black] coordinates {\n$lines\n};"
    val annotations = new ArrayBuffer[String]()
    var lastExternCalls: Map[String, Int] = Seq().toMap
    for ((externCalls, i) <- externCounter.externCallsOverTime.zipWithIndex) {
      if (externCalls != lastExternCalls) {
        val note = "\\scriptsize $" + externCalls
          .map(t => (trimName(t._1), t._2))
          .map(t => s"${t._2} \\times \\texttt{${t._1}}").mkString(", ") + "$"
        val arrow = s"\\draw [<-] (axis cs:$i,${nodeCounter.nodesOverTime(i)})-- +(-5pt,10pt) node[left] {$note};"
        annotations.append(arrow)

        lastExternCalls = externCalls
      }
    }
    (plot +: annotations).mkString("\n")
  }

  private def buildCsvRow[ENode <: ENodeT, EClass <: EClassT](kernel: Kernel,
                                                              graph: EGraphT[ENode, EClass],
                                                              root: EClass,
                                                              duration: Duration): Seq[Any] = {
    val bestExpr = kernel.extractor(root).get
    val externCallCounts = countExternCalls(bestExpr)
    Seq(
      kernel.name,
      kernel.suite,
      kernel.maxExpansionSteps + kernel.maxSimplificationSteps,
      graph.classes.length,
      graph.classes.map(_.nodes.length).sum,
      externCallCounts.getOrElse("idiom_ddot", 0),
      externCallCounts.getOrElse("idiom_dgemv", 0),
      externCallCounts.getOrElse("idiom_dgemm", 0),
      duration.toSeconds,
      kernel.matchesOracle(bestExpr),
      kernel.matchesOracle(graph, root)
    )
  }

  private def assembleCsv(rows: Seq[Seq[Any]]): String = {
    val header = Seq(
      "name", "suite", "steps", "classes", "nodes",
      "ddot", "dgemv", "dgemm", "time", "bestIsOracle",
      "foundOracle"
    )

    val allRows = header +: rows
    allRows.map(_.map(_.toString).mkString(",")).mkString("\n")
  }

  private def countExternCalls(expr: Expr): Map[String, Int] = {
    val count = mutable.Map[String, Int]()
    expr.visit({
      case ExternFunctionCall(name, _, _, _) => count(name) = count.getOrElse(name, 0) + 1
      case _ =>
    })
    count.toMap
  }

  private def printReasoning[L <: ENodeT](explanation: TreeExplanation[L]): Unit = {
    val flatExplanation = explanation.flatten
    println(s"Reasoning (${flatExplanation.steps.length} steps):")
    println(flatExplanation)
  }

  @Test
  def findMinSteps(): Unit = {
    for (kernel <- kernels) {
      println(s"Finding min steps for ${kernel.name}...")
      kernel.findMinSteps() match {
        case Some((expansion, simplification)) => println(s"Min steps: $expansion steps, $simplification steps")
        case None => println(
          s"Could not solve problem for ${kernel.name}. " +
            s"${kernel.maxExpansionSteps} expansion+simplification steps and " +
            s"${kernel.maxSimplificationSteps} simplification steps are not enough")
      }
    }
  }

  @Test
  def findMinNaiveSteps(): Unit = {
    for (kernel <- kernels) {
      println(s"Finding min steps for ${kernel.name}...")
      kernel.findMinNaiveSteps() match {
        case Some(steps) => println(s"Min steps: $steps steps")
        case None => println(
          s"Could not solve problem for ${kernel.name}. " +
            s"${kernel.maxExpansionSteps + kernel.maxSimplificationSteps} steps are not enough")
      }
    }
  }
}
