package cGen

import cGen.Lib._
import cGen.pass.TuningPass.{getUpperBoundRewriteNum, perfModelTuningPassWithDPS, randomTuningPassWithDPS}
import core.{Expr, ParamDef, ParamUse, TypeChecker}
import org.junit.Test

import scala.language.postfixOps

class PolyBenchTest {
  val OneMMKernelName = "oneMM"
  val OneMMWithLazyTransposeKernelName = "oneMMLazyT"
  val OneMMWorstCaseKernelName = "oneMM_worse_case"
  val TwoMMKernelName = "twoMM"
  val TwoMMWithLazyTransposeKernelName = "twoMM_all_view"
  val TwoMMWorstCaseKernelName = "twoMM_worse_case"
  val ThreeMMKernelName = "threeMM"
  val ThreeMMWithLazyTransposeKernelName = "threeMM_all_view"
  val ThreeMMWorstCaseKernelName = "threeMM_worse_case"
  val Blur1DKernelName = "blur1D"
  val Jacobi1DKernelName = "jacobi1D"
  val Jacobi1DWorstCaseKernelName = "jacobi1D_worse_case"
  val Jacobi2DKernelName = "jacobi2D"
  val Jacobi2DWorstCaseKernelName = "jacobi2D_worse_case"
  val Seidel2DKernelName = "seidel2D"
  val Jacobi2DPadSKernelName = "jacobi2DPadS"
  val JoinKernelName = "join"

  /**
   * Creates a mapping of PolyBench kernel names to kernel expressions.
   * @param size The dimensions of the data types manipulated by the kernels.
   * @return A mapping of kernel names to expressions.
   */
  def kernels(size: Int = 1024): Map[String, Expr] = {
    Predef.Map(
      OneMMKernelName -> oneMM(size, size, size)(isTransposeLazy = false),
      OneMMWithLazyTransposeKernelName -> oneMM(size, size, size)(isTransposeLazy = true),
      OneMMWorstCaseKernelName -> {
        implicit val t = () => DoubleType()
        val A = ParamDef(ArrayType(ArrayType(t(), size), size))
        val B = ParamDef(ArrayType(ArrayType(t(), size), size))
        CLambda(Seq(A, B).reverse, {
          val m1 = ParamUse(A)
          val m2 = ParamUse(B)
          MatrixMulWorseCase(m1, m2)
        })
      },
      TwoMMKernelName -> twoMM(size, size, size, size)(isTransposeLazy = false),
      TwoMMWithLazyTransposeKernelName -> twoMM(size, size, size, size)(isTransposeLazy = true),
      TwoMMWorstCaseKernelName -> {
          implicit val t = () => DoubleType()
          val A = ParamDef(ArrayType(ArrayType(t(), size), size))
          val B = ParamDef(ArrayType(ArrayType(t(), size), size))
          val C = ParamDef(ArrayType(ArrayType(t(), size), size))
          val D = ParamDef(ArrayType(ArrayType(t(), size), size))
          val alpha = ParamDef(t())
          val beta = ParamDef(t())

          CLambda(Seq(alpha, beta, A, B, C, D).reverse, {
            val prog = new Program()
            val abmm = prog << MatrixMulWorseCase(ParamUse(A), ParamUse(B))
            val tmp = prog << MatrixMulScalar(abmm, ParamUse(alpha))
            val ctmm = prog << MatrixMulWorseCase(tmp, ParamUse(C))
            prog.build(MatrixAdd(ctmm, MatrixMulScalar(ParamUse(D), ParamUse(beta))))
          })
      },
      ThreeMMKernelName -> threeMM(size, size, size, size, size)(isTransposeLazy = false),
      ThreeMMWithLazyTransposeKernelName -> threeMM(size, size, size, size, size)(isTransposeLazy = true),
      ThreeMMWorstCaseKernelName -> {
        implicit val t = () => DoubleType()
        val A = ParamDef(ArrayType(ArrayType(t(), size), size))
        val B = ParamDef(ArrayType(ArrayType(t(), size), size))
        val C = ParamDef(ArrayType(ArrayType(t(), size), size))
        val D = ParamDef(ArrayType(ArrayType(t(), size), size))

        CLambda(Seq(A, B, C, D).reverse, {
          val prog = new Program()
          val E = prog.add(MatrixMulWorseCase(ParamUse(A), ParamUse(B)))
          val F = prog.add(MatrixMulWorseCase(ParamUse(C), ParamUse(D)))
          prog.build(MatrixMulWorseCase(E, F))
        })
      },
      Blur1DKernelName -> blur1D(size),
      Jacobi1DKernelName -> jacobi1DNonIter(size, 2),
      Jacobi1DWorstCaseKernelName -> jacobi1DWorseCase(size, 2),
      Jacobi2DKernelName -> jacobi2DWithSlide2(size, 2),
      Seidel2DKernelName -> seidel2D(size, 10),
      Jacobi2DPadSKernelName -> {
        val mat = ParamDef(ArrayType(ArrayType(DoubleType(), size), size))
        CLambda(mat, map2DE(conv2(() => DoubleType()), padS2D(slideS2D(ParamUse(mat), 3))))
      },
      JoinKernelName -> {
        val mat = ParamDef(ArrayType(ArrayType(IntType(32), size), size))
        CLambda(mat, Join(MapEId(ParamUse(mat))))
      }
    )
  }

  /**
   * Compiles a kernel with a given name to C code.
   * @param name The kernel's name.
   * @param size The dimensions of the data types manipulated by the kernel.
   */
  def compileKernel(name: String, size: Int = 1024): Unit = {
    val tree = DPSPass(kernels(size)(name))
    val cCode = CCodeGen(tree, funcName = name)
    cCode.genTest(fileName = name)
  }

  def printKernelCode(name: String, size: Int = 1024) = {
    val tree = DPSPass(kernels(size)(name))
    val cCode = CCodeGen(tree, funcName = name)
    println(cCode)
  }

  def getKernelOnly(name: String, size: Int = 1024): Expr = kernels(size)(name)

  @Test
  def GenMMTest(): Unit = {
    val sizes = Seq(16, 128, 512, 1024, 2048, 4096)
    sizes.foreach(size => compileKernel(OneMMWithLazyTransposeKernelName, size))
    sizes.foreach(size => compileKernel(OneMMKernelName, size))
  }

  @Test
  def oneMMTest2(): Unit = {
    compileKernel(OneMMKernelName, 128)
  }


  @Test
  def oneMMAllViewTest2(): Unit = {
    val size = 1024;
    val tree = oneMM(size, size, size)(true)
    CCodeGen(DPSPass(tree), "oneMM_all_view").genTest()
  }

  @Test
  def oneMMRandomTuneTest(): Unit = {
    val size = 128
    val tree = oneMM(size, size, size)
    println(getUpperBoundRewriteNum(tree))
    val partrees = randomTuningPassWithDPS(tree, 100, getUpperBoundRewriteNum(tree))
    var trees = Isomorphic.deduplicate(partrees)
    CCode.genCodes(trees.map(CCodeGen.tryCCodeGen(_, "oneMM")).filter(_.isDefined).map(_.get).seq)
  }

  @Test
  def oneMMPerfTuneTest(): Unit = {
    val size = 128
    val tree = oneMM(size, size, size)
    val trees = perfModelTuningPassWithDPS(tree, 10)
    println(trees.length)
    CCode.genCodes(trees.map(CCodeGen(_, "oneMM")))
  }

  @Test
  def oneMMAutoTuneTest(): Unit = {
    val size = 1024
    val P = size
    val Q = size
    val R = size
    val tree = {
      implicit val t = () => DoubleType()
      val A = ParamDef(ArrayType(ArrayType(t(), Q), P))
      val B = ParamDef(ArrayType(ArrayType(t(), R), Q))
      CLambda(Seq(A, B).reverse, {
        val m1 = ParamUse(A)
        val m2 = ParamUse(B)
        val len = cols(TypeChecker.check(m1))
        val row1 = ParamDef(ArrayType(t().asInstanceOf[CGenDataTypeT], len))
        val row2 = ParamDef(ArrayType(t().asInstanceOf[CGenDataTypeT], len))
        Must(MapS(CLambda(row1,
          MapS(CLambda(row2,
            DotProd(ParamUse(row1), ParamUse(row2))
          ), Mtrl(transpose(m2)))), m1))
      })
    }
    var trees = AutoTune(tree)
    trees = trees.map(DPSPass(_))
//    CCodeGen.genCodes("oneMM", trees)
  }

  @Test
  def oneMMWorseCaseTest(): Unit = {
    compileKernel(OneMMWorstCaseKernelName)
  }

  @Test
  def twoMMTest(): Unit = {
    compileKernel(TwoMMKernelName)
  }

  @Test
  def twoMMAllViewTest(): Unit = {
    compileKernel(TwoMMWithLazyTransposeKernelName)
  }

  @Test
  def twoMMAutoTuneTest(): Unit = {
    val size = 1024
    val P = size;
    val Q = size;
    val R = size;
    val S = size;
    implicit val isTransposeLazy = false
    val tree = {
      implicit val t = () => DoubleType()
      val A = ParamDef(ArrayType(ArrayType(t(), Q), P))
      val B = ParamDef(ArrayType(ArrayType(t(), R), Q))
      val C = ParamDef(ArrayType(ArrayType(t(), S), R))
      val D = ParamDef(ArrayType(ArrayType(t(), S), P))
      val alpha = ParamDef(t())
      val beta = ParamDef(t())

      CLambda(Seq(alpha, beta, A, B, C, D).reverse, {
        Must(MatrixAdd(
          AutoTuneMatrixMulScalar(AutoTuneMatrixMul(AutoTuneMatrixMul(ParamUse(A), ParamUse(B)), ParamUse(C)), ParamUse(alpha)),
          AutoTuneMatrixMulScalar(ParamUse(D), ParamUse(beta)))
        )
      })
    }
    var trees = AutoTune(tree).map(DPSPass(_))
//    CCodeGen.genCodes("twoMM", trees)
  }

  @Test
  def twoMMWorseCaseTest(): Unit = {
    compileKernel(TwoMMWorstCaseKernelName)
  }

  @Test
  def threeMMTest(): Unit = {
    compileKernel(ThreeMMKernelName)
  }

  @Test
  def threeMMAutoTuneTest(): Unit = {
    val size = 1024
    val P = size;
    val Q = size;
    val R = size;
    val S = size;
    val T = size;
    implicit val isTransposeLazy = false
    val tree = {
      implicit val t = () => DoubleType()
      val A = ParamDef(ArrayType(ArrayType(t(), Q), P))
      val B = ParamDef(ArrayType(ArrayType(t(), R), Q))
      val C = ParamDef(ArrayType(ArrayType(t(), S), R))
      val D = ParamDef(ArrayType(ArrayType(t(), S), T))

      CLambda(Seq(A, B, C, D).reverse, {
        val prog = new Program()
        val E = prog.add(Must(AutoTuneMatrixMul(ParamUse(A), ParamUse(B))))
        val F = prog.add(Must(AutoTuneMatrixMul(ParamUse(C), ParamUse(D))))
        prog.build(Must(AutoTuneMatrixMul(E, F)))
      })
    }
    var trees = AutoTune(tree)
    trees = trees.map(DPSPass(_))
//    CCodeGen.genCodes("threeMM", trees)
  }

  @Test
  def threeMMAllViewTest(): Unit = {
    compileKernel(ThreeMMWithLazyTransposeKernelName)
  }

  @Test
  def threeMMWorseCaseTest(): Unit = {
    compileKernel(ThreeMMWorstCaseKernelName)
  }

  @Test
  def Blur1DTest(): Unit = {
    compileKernel(Blur1DKernelName, 1 << 17)
  }

  @Test
  def jacobi1DTuneTest(): Unit = {
    val size = 1 << 24
    val tree = getKernelOnly(Jacobi1DKernelName, size)
    val trees = perfModelTuningPassWithDPS(tree, 10)
    CCode.genCodes(trees.map(CCodeGen(_, "jacobi1D")))
  }

  @Test
  def jacobi1DRandomTuneTest(): Unit = {
    val size = 1 << 24
    val tree = getKernelOnly(Jacobi1DKernelName, size)
    val trees = randomTuningPassWithDPS(tree, 1000, getUpperBoundRewriteNum(tree))
    CCode.genCodes(trees.map(CCodeGen.tryCCodeGen(_, "jacobi1D")).filter(_.isDefined).map(_.get).seq)
  }

  @Test
  def jacobi1DTest(): Unit = {
    compileKernel(Jacobi1DKernelName, 1 << 24)
  }

  @Test
  def jacobi1DWorseCaseTest(): Unit = {
    compileKernel(Jacobi1DWorstCaseKernelName, 1 << 24)
  }

  @Test
  def jacobi2DTuneTest(): Unit = {
    val len = 1 << 13
    val step = 2
    val tree = jacobi2DWithSlide2(len, step)
    val trees = perfModelTuningPassWithDPS(tree, 10)
    CCode.genCodes(trees.map(CCodeGen(_, "jacobi2D")))
  }

  @Test
  def jacobi2DRandomTuneTest(): Unit = {
    val len = 1 << 13
    val step = 2
    val tree = jacobi2DWithSlide2(len, step)
    val trees = randomTuningPassWithDPS(tree, 100, (getUpperBoundRewriteNum(tree) * 1.5).toInt)
    CCode.genCodes(trees.map(CCodeGen.tryCCodeGen(_, "jacobi2D")).filter(_.isDefined).map(_.get).seq)
  }

  @Test
  def jacobi2DTest(): Unit = {
    compileKernel(Jacobi2DKernelName, 1 << 13)
  }

  @Test
  def jacobi2DWorseCaseTest(): Unit = {
    compileKernel(Jacobi2DWorstCaseKernelName, 1 << 12)
  }

  @Test
  def seidel2DTest(): Unit = {
    compileKernel(Seidel2DKernelName, 100)
  }

  @Test
  def jacobi2DPadSTest(): Unit = {
    printKernelCode(Jacobi2DPadSKernelName, 1 << 12)
  }

  @Test
  def joinTest(): Unit = {
    printKernelCode(JoinKernelName, 2)
  }

  @Test
  def kroneckerProductTest(): Unit = {
    val len = 1 << 6
    implicit val t = () => DoubleType()
    val mat0 = ParamDef(ArrayType(ArrayType(t(), len), len))
    val mat1 = ParamDef(ArrayType(ArrayType(t(), len), len))
    val f = CLambda(mat0, CLambda(mat1, {
      val v = ParamDef(t())
      var r = map2DE(CLambda(v, MatrixMulScalarE(ParamUse(v), ParamUse(mat1))), ParamUse(mat0))
      var vv = ParamDef(tt(r).asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
      r = Mtrl(map2DD(CLambda(vv, Transpose(ParamUse(vv), DestViewType())), r))
      //      assert(getInverse(r).isDefined)
      vv = ParamDef(tt(r).asInstanceOf[ArrayTypeT].et)
      r = MapD(CLambda(vv, Join(ParamUse(vv), DestViewType())), r)
      //      assert(getInverse(r).isDefined)
      vv = ParamDef(tt(r).asInstanceOf[ArrayTypeT].et)
      r = Mtrl(MapD(CLambda(vv, Transpose(ParamUse(vv), DestViewType())), r))
      r = Join(r, DestViewType())
      println(TypeChecker.check(r).t)
      r
    }))
  }
}
