package cGen.eqsat

import cGen.{AlgoOp, BinaryOp, Constant, IntType}
import core.{Expr, FunctionCall, TypeChecker}
import org.junit.Test

class SubgraphRewritingTest {

  def areEquivalent[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass])(eClass: EClass, pattern: Expr): Boolean =
    Pattern(pattern).searchEClass(graph)(eClass, HierarchicalStopwatch.disabled).nonEmpty

  @Test
  def canShiftTree(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))
    val root = graph.add(TypeChecker.check(BinaryOp(DeBruijnParamUse(0, t), Constant(4, t), AlgoOp.Add)))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val shifted = builder.transferTo(
      graph,
      NoReason,
      ShiftingRewriter(graph, builder, 1).rewrite(root).asInstanceOf[builder.Promise])
    assert(areEquivalent(graph)(
      shifted,
      TypeChecker.check(BinaryOp(DeBruijnParamUse(1, t), Constant(4, t), AlgoOp.Add))))
  }

  @Test
  def canShiftDAG(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))
    val root = graph.add(TypeChecker.check(BinaryOp(DeBruijnParamUse(0, t), DeBruijnParamUse(0, t), AlgoOp.Add)))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val shifted = builder.transferTo(
      graph,
      NoReason,
      ShiftingRewriter(graph, builder, 1).rewrite(root).asInstanceOf[builder.Promise])
    assert(areEquivalent(graph)(
      shifted,
      TypeChecker.check(BinaryOp(DeBruijnParamUse(1, t), DeBruijnParamUse(1, t), AlgoOp.Add))))
  }

  @Test
  def canShiftCycle(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))
    val root = graph.add(TypeChecker.check(BinaryOp(DeBruijnParamUse(0, t), DeBruijnParamUse(0, t), AlgoOp.Add)))
    graph.merge(root, graph.add(TypeChecker.check(BinaryOp(root, Constant(0, t), AlgoOp.Add))), DirectlyAdded)
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val shifted = builder.transferTo(
      graph,
      NoReason,
      ShiftingRewriter(graph, builder, 1).rewrite(root).asInstanceOf[builder.Promise])
    assert(areEquivalent(graph)(
      shifted,
      TypeChecker.check(BinaryOp(DeBruijnParamUse(1, t), DeBruijnParamUse(1, t), AlgoOp.Add))))
  }

  @Test
  def shiftCanFork(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))

    // Tricky test case: we create an e-graph that captures expressions (\ p_0) p_0. The shifted version of this
    // expression is (\ p_0) p_1. Obtaining this shifted version using e-graph rewriting is tricky to get right because
    // p_0 appears only once  in the e-graph. This will result in a cyclic reference to p_0 that implies a value
    // assignment different from the naked reference to p_0. To handle this different value, the rewriter will have to
    // fork when it sees the lambda.

    val root = graph.add(
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, DeBruijnParamUse(0, t)), DeBruijnParamUse(0, t))))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val shifted = builder.transferTo(
      graph,
      NoReason,
      ShiftingRewriter(graph, builder, 1).rewrite(root).asInstanceOf[builder.Promise])

    assert(areEquivalent(graph)(
      shifted,
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, DeBruijnParamUse(0, t)), DeBruijnParamUse(1, t)))))

    assert(!areEquivalent(graph)(
      shifted,
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, DeBruijnParamUse(1, t)), DeBruijnParamUse(1, t)))))
  }

  @Test
  def canSubstituteTree(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))
    val root = graph.add(TypeChecker.check(BinaryOp(DeBruijnParamUse(0, t), Constant(4, t), AlgoOp.Add)))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val substituted = builder.transferTo(
      graph,
      NoReason,
      SubstitutingRewriter(graph, builder, List(Constant(4, t))).rewrite(root).asInstanceOf[builder.Promise])
    assert(areEquivalent(graph)(
      substituted,
      TypeChecker.check(BinaryOp(Constant(4, t), Constant(4, t), AlgoOp.Add))))
  }

  @Test
  def canSubstituteDAG(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))
    val root = graph.add(TypeChecker.check(BinaryOp(DeBruijnParamUse(0, t), DeBruijnParamUse(0, t), AlgoOp.Add)))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val substituted = builder.transferTo(
      graph,
      NoReason,
      SubstitutingRewriter(graph, builder, List(graph.add(Constant(4, t)))).rewrite(root).asInstanceOf[builder.Promise])
    assert(areEquivalent(graph)(
      substituted,
      TypeChecker.check(BinaryOp(Constant(4, t), Constant(4, t), AlgoOp.Add))))
  }

  @Test
  def substituteCanFork(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))

    // Tricky test case: we create an e-graph that captures expressions (\ p_0) p_0. We then substitute 4 for p_0,
    // resulting in (\ p_0) 4. Obtaining this substituted expression using e-graph rewriting is tricky to get right because
    // p_0 appears only once in the e-graph. This will result in a cyclic reference to p_0 that implies a value
    // assignment different from the naked reference to p_0. To handle this different value, the rewriter will have to
    // fork when it sees the lambda.

    val root = graph.add(
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, DeBruijnParamUse(0, t)), DeBruijnParamUse(0, t))))
    val four = graph.add(Constant(4, t))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val substituted = builder.transferTo(
      graph,
      NoReason,
      SubstitutingRewriter(graph, builder, List(four)).rewrite(root).asInstanceOf[builder.Promise])

    assert(areEquivalent(graph)(
      substituted,
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, DeBruijnParamUse(0, t)), Constant(4, t)))))

    assert(!areEquivalent(graph)(
      substituted,
      TypeChecker.check(FunctionCall(DeBruijnLambda(t, Constant(4, t)), Constant(4, t)))))
  }

  @Test
  def substituteCanShift(): Unit = {
    val t = IntType(32)

    val graph = new DeferredMergingEGraph(
      new Willsey2021EGraph(analyses = new AnalysesBuilder().add(ParameterUseAnalysis).toAnalyses))

    val root = graph.add(
      TypeChecker.check(
        FunctionCall(
          DeBruijnLambda(t, BinaryOp(DeBruijnParamUse(1, t), DeBruijnParamUse(0, t), AlgoOp.Add)),
          DeBruijnParamUse(0, t))))
    val param = graph.add(DeBruijnParamUse(2, t))
    graph.rebuild()

    val builder = new SubgraphBuilder[HashConsENode, HashConsEClass]()
    val substituted = builder.transferTo(
      graph,
      NoReason,
      SubstitutingRewriter(graph, builder, List(param)).rewrite(root).asInstanceOf[builder.Promise])

    assert(areEquivalent(graph)(
      substituted,
      TypeChecker.check(
        FunctionCall(
          DeBruijnLambda(t, BinaryOp(DeBruijnParamUse(3, t), DeBruijnParamUse(0, t), AlgoOp.Add)),
          DeBruijnParamUse(2, t)))))
  }
}
